// BoxList.cpp : implementation file
//

#include "stdafx.h"
#include "BoxList.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CBoxList

CBoxList::CBoxList()
{
	m_nScrollPos = 0;
	m_nSelected = 0;
	m_MouseHover = -1;
	m_MousePress = -1;
	m_nItemPadding = 2;
	m_nButtonPadding = 2;
	m_nScrollBarWidth = 15;
	m_nButtonHeight = 15;
	m_bShowButtons = TRUE;

	btn_normal.face		= RGB(225,225,255);
	btn_normal.border	= RGB(110,180,255);
	btn_normal.highlight= RGB(170,210,255);
	btn_normal.shadow	= RGB(40,145,255);

	btn_hover.face		= RGB(255,225,225);
	btn_hover.border	= RGB(255,180,110);
	btn_hover.highlight	= RGB(255,210,170);
	btn_hover.shadow	= RGB(255,145,40);

	btn_down.face		= RGB(255,225,225);
	btn_down.border		= RGB(255,225,225);
	btn_down.highlight	= RGB(255,45,0);
	btn_down.shadow		= RGB(255,45,0);
}

CBoxList::~CBoxList()
{

}


BEGIN_MESSAGE_MAP(CBoxList, CWnd)
	//{{AFX_MSG_MAP(CBoxList)
	ON_WM_PAINT()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_TIMER()
	ON_WM_LBUTTONDBLCLK()
	ON_WM_MOUSEMOVE()
	ON_WM_MOUSEWHEEL()
	ON_WM_VSCROLL()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CBoxList message handlers

void CBoxList::OnPaint() 
{
	CPaintDC dc(this); // device context for painting


	CFont font,font2;
	CFont *fntOld;

	CRect rect;

	//* client rect for window dimensions *//
	GetClientRect(rect);

	//* Create GDI objects *//
	font.CreatePointFont(120, "Marlett");			// Used for arrow icon
	font2.CreatePointFont(100, "MS Sans Serif");	// Normal text



	//* Set DC options *
	fntOld = buf.SelectObject(&font);


	buf.SetBkMode(TRANSPARENT);
	buf.SetTextAlign(TA_CENTER);


	//* Clear DC *
	buf.FillSolidRect(rect, GetSysColor(COLOR_3DFACE));



	//////////////////////////////
	// *** Draw Buttons ***		//
	//////////////////////////////

	if(m_bShowButtons)
	{
		BtnClr clrUp,clrDown;
		
		if(m_bHasMouse && btnUp.PtInRect(m_MousePos))
		{
			clrUp = btn_hover;
			if(m_MousePress == 0)
				clrUp = btn_down;
		}
		else
			clrUp = btn_normal;

		if(m_bHasMouse && btnDown.PtInRect(m_MousePos))
		{
			clrDown = btn_hover;
			if(m_MousePress == m_nPageSize+2)
				clrDown = btn_down;
		}
		else
			clrDown = btn_normal;

	buf.FillSolidRect(btnUp.left+3,btnUp.top+3,btnUp.Width()-6,btnUp.Height()-6, clrUp.face);
	buf.FillSolidRect(btnDown.left+3,btnDown.top+3,btnDown.Width()-6,btnDown.Height()-6, clrDown.face);

	buf.TextOut(btnUp.Width()/2, btnUp.top, "\x35");
	buf.TextOut(btnDown.Width()/2, btnDown.top, "\x36");

	buf.Draw3dRect(btnUp, 0, 0);
	buf.Draw3dRect(btnDown, 0, 0);

	buf.Draw3dRect(btnUp.left+1, btnUp.top+1, btnUp.Width()-2,btnUp.Height()-2, clrUp.highlight, clrUp.shadow);
	buf.Draw3dRect(btnUp.left+2, btnUp.top+2, btnUp.Width()-4,btnUp.Height()-4, clrUp.border, clrUp.border);

	buf.Draw3dRect(btnDown.left+1, btnDown.top+1, btnDown.Width()-2,btnDown.Height()-2, clrDown.highlight, clrDown.shadow);
	buf.Draw3dRect(btnDown.left+2, btnDown.top+2, btnDown.Width()-4,btnDown.Height()-4, clrDown.border, clrDown.border);

	}

	// Reset font and alignment for owner drawing
	//buf.SelectObject(&font2);
	buf.SetTextAlign(TA_LEFT);

	for(int i = 0; i < m_nPageSize; i++)
	{
		int y = (m_bShowButtons*m_nButtonHeight)+i*(m_nItemHeight+m_nItemPadding)+1;

		int nItem = i+m_nScrollPos;

		if(nItem == m_nSelected)
			buf.FillSolidRect(itemRect[i], RGB(160,180,255));

		buf.Draw3dRect(itemRect[i], 0, 0);

		buf.SetViewportOrg(0, y);
		m_pParent->SendMessage(BM_DRAWITEM, WPARAM(nItem), LPARAM(&buf));
		buf.SetViewportOrg(0, 0);
	}

	buf.SelectObject(fntOld);

	dc.BitBlt(0,0,rect.Width()-m_nScrollBarWidth-2, rect.Height(), &buf, 0, 0, SRCCOPY);
}


void CBoxList::OnLButtonDown(UINT nFlags, CPoint point) 
{
	int curPress = -1;

	if(btnUp.PtInRect(point))
	{
		curPress = 0;

		m_nScrollDir = -1;
		ScrollBy(-1);

		SetTimer(0, 400, NULL);
	}

	for(int i = 0; i < m_nPageSize; i++)
	{
		if(itemRect[i].PtInRect(point))
		{
			curPress = i+1;
			m_nSelected = m_nScrollPos+i;

			m_pParent->SendMessage(BM_DRAGITEM);
			m_pParent->SendMessage(BM_SELECTITEM);
		}
	}
	if(btnDown.PtInRect(point))
	{
		curPress = m_nPageSize+2;

		m_nScrollDir = 1;
		ScrollBy(1);

		SetTimer(0, 400, NULL);
	}

	m_MousePress = curPress;
	Invalidate(FALSE);

	SetFocus();

	CWnd::OnLButtonDown(nFlags, point);
}

void CBoxList::OnLButtonUp(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	KillTimer(0);
	KillTimer(1);

	m_MousePress = -1;

	Invalidate(FALSE);

	m_pParent->SendMessage(BM_CANCELDRAGITEM);
	
	CWnd::OnLButtonUp(nFlags, point);
}

void CBoxList::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	OnLButtonDown(nFlags, point);

	m_pParent->SendMessage(BM_DBLCLICK, m_nSelected);

	//CWnd::OnLButtonDblClk(nFlags, point);
}

void CBoxList::OnTimer(UINT nIDEvent) 
{
	// TODO: Add your message handler code here and/or call default
	if((m_nScrollDir == -1 && !btnUp.PtInRect(m_MousePos)) ||
	   (m_nScrollDir ==  1 && !btnDown.PtInRect(m_MousePos)))
	   return;

	switch(nIDEvent)
	{
	case 0:
		KillTimer(0);
		SetTimer(1, 75, NULL);
		break;
	case 1:
		ScrollTo(m_nScrollPos+m_nScrollDir);
		break;
	}
	CWnd::OnTimer(nIDEvent);
}



void CBoxList::OnMouseMove(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	m_bHasMouse = true;
	m_MousePos = point;
	
	int curHover = -1;

	if(btnUp.PtInRect(point)) curHover = 0;
	for(int i = 0; i < m_nPageSize; i++)
		if(itemRect[i].PtInRect(point))
			curHover = i+1;
	if(btnDown.PtInRect(point)) curHover = m_nPageSize+2;

	if(curHover != -1 && curHover != m_MouseHover)
	{
		m_MouseHover = curHover;
		Invalidate(FALSE);
	}
	else
	{
		m_MouseHover = -1;
		Invalidate(FALSE);
	}

	CWnd::OnMouseMove(nFlags, point);
}

BOOL CBoxList::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt) 
{
	// TODO: Add your message handler code here and/or call default
	//ScrollBy(zDelta);
	
	return FALSE;//CWnd::OnMouseWheel(nFlags, zDelta, pt);
}


void CBoxList::OnVScroll( UINT nSBCode, UINT nPos, CScrollBar* pScrollBar )
{
	switch(nSBCode)
	{
	case SB_THUMBTRACK:
		ScrollTo(nPos, FALSE);
		break;
	case SB_LINEUP:
		ScrollBy(-1, FALSE);
		break;
	case SB_LINEDOWN:
		ScrollBy(1, FALSE);
		break;
	case SB_PAGEUP:
		ScrollBy(-m_nPageSize+1, FALSE);
		break;
	case SB_PAGEDOWN:
		ScrollBy(m_nPageSize-1, FALSE);
		break;
	case SB_BOTTOM:
		ScrollTo(m_nListSize-m_nPageSize, FALSE);
		break;
	case SB_TOP:
		ScrollTo(0,FALSE);
		break;
	}

	pScrollBar->SetScrollPos(m_nScrollPos);

	CRect rect;
	GetClientRect(rect);

	// Exclude scroll bar from painting
	rect.right -= (m_nScrollBarWidth+2);

	InvalidateRect(rect, FALSE);
}



/////////////////////////////////////////////////////////////////////////////
// CBoxList functions

void CBoxList::CreateList(int x, int y, int nItemWidth, int nItemHeight, int nListSize, int nPageSize, CWnd *parent, UINT nID)
{
	// Creates the window as a child of the specified window
	Create(NULL, "_box", WS_CHILD, CRect(0,0,0,0),parent,nID);

	m_pParent = parent;

	m_nID = nID;

	m_nItemWidth = nItemWidth;
	m_nItemHeight = nItemHeight;

	m_nXPos = x;
	m_nYPos = y;

	m_nListSize = nListSize;
	m_nPageSize = nPageSize;

	m_wndScrollBar.Create(SBS_VERT|WS_CHILD|WS_VISIBLE, CRect(0,0,0,0), this, 0);


	buf.CreateCompatibleDC(GetDC());
	SelectObject((HDC)buf, CreateCompatibleBitmap((HDC)*GetDC(), 1000, 1000));


	CalcLayout();

	ShowWindow(SW_NORMAL);
}


void CBoxList::CalcLayout()
{
	CRect rect;
	GetWindowRect(rect);

	// total height and width of window
	int nHeight	= (m_nButtonHeight+m_nButtonPadding)*2*m_bShowButtons+ (!m_bShowButtons*2) + (m_nItemHeight+m_nItemPadding)*m_nPageSize;
	int nWidth	= m_nItemWidth + 2 + m_nScrollBarWidth;

	rect.bottom	= rect.top + nHeight;
	rect.right	= rect.left + nWidth;

	ScreenToClient(&rect);

	rect.top	= m_nYPos;
	rect.left	= m_nXPos;
	rect.bottom+= m_nYPos;
	rect.right += m_nXPos;

	MoveWindow(rect);


	// position buttons and item rectangles
	if(m_bShowButtons)
	{
		btnUp.left = 0;
		btnUp.right = m_nItemWidth;
		btnUp.top = 0;
		btnUp.bottom = m_nButtonHeight;
	} else btnUp = CRect(0,0,0,0);

	for(int i = 0; i < m_nPageSize; i++)
	{
		itemRect[i].left = 0;
		itemRect[i].right = m_nItemWidth;

		itemRect[i].top = i*(m_nItemHeight+m_nItemPadding)+btnUp.bottom+m_nButtonPadding;
		itemRect[i].bottom = itemRect[i].top + m_nItemHeight;
	}

	if(m_bShowButtons)
	{
		btnDown.left = 0;
		btnDown.right = m_nItemWidth;
		btnDown.top = itemRect[m_nPageSize-1].bottom+m_nButtonPadding;
		btnDown.bottom = btnDown.top+m_nButtonHeight;
	} else btnDown = CRect(0,0,0,0);

	// Set scroll bar position
	CRect scrollRect;

	scrollRect.left = m_nItemWidth;
	scrollRect.right = m_nItemWidth+m_nScrollBarWidth;
	scrollRect.top = 0;
	scrollRect.bottom = nHeight-2;

	m_wndScrollBar.MoveWindow(scrollRect);

	// Set scroll size/range
	SCROLLINFO scr;
	scr.fMask = SIF_PAGE;

	scr.nPage = m_nPageSize;

	m_wndScrollBar.SetScrollInfo(&scr, TRUE);

	m_wndScrollBar.SetScrollRange(0, m_nListSize-1);
}





// 'Get' attributes

CSize CBoxList::GetItemSize()
{
	return CSize(m_nItemWidth, m_nItemHeight);
}

int CBoxList::GetListSize()
{
	return m_nListSize;
}

int CBoxList::GetPageSize()
{
	return m_nPageSize;
}


int CBoxList::GetScroll()
{
	return m_nScrollPos;
}

int CBoxList::GetSelected()
{
	return m_nSelected;
}

int CBoxList::GetItemHeight()
{
	return m_nItemHeight;
}

int CBoxList::GetItemWidth()
{
	return m_nItemWidth;
}

int CBoxList::GetButtonHeight()
{
	if(m_bShowButtons)
		return m_nButtonHeight;
	else
		return 0;
}

int CBoxList::GetButtonPadding()
{
	return m_nButtonPadding;
}

int CBoxList::GetItemPadding()
{
	return m_nItemPadding;
}

COLORREF CBoxList::GetBtnColor(int nColorSet, int nItem)
{
	BtnClr *clrset;

	switch(nColorSet)
	{
	case COLORSET_NORMAL:clrset = &btn_normal;break;
	case COLORSET_HOVER: clrset = &btn_hover; break;
	case COLORSET_DOWN:  clrset = &btn_down;  break;
	}

	COLORREF clr = ((COLORREF*)clrset)[nItem];

	return clr;
}



// 'Set' attributes
void CBoxList::SetItemSize(int cx, int cy)
{
	m_nItemWidth = cx;
	m_nItemHeight = cy;
}

void CBoxList::SetItemSize(CSize size)
{
	m_nItemWidth = size.cx;
	m_nItemHeight = size.cy;
}


void CBoxList::SetListSize(int nItems)
{
	m_nListSize = nItems;
	CalcLayout();
}

void CBoxList::SetPageSize(int nPage)
{
	m_nPageSize = nPage;
	CalcLayout();
}

void CBoxList::SetItemPadding(int nPad)
{
	m_nItemPadding = nPad;
}

void CBoxList::SetButtonPadding(int nPad)
{
	m_nButtonPadding = nPad;
}

void CBoxList::SetButtonHeight(int nHeight, bool bRecalc)
{
	m_nButtonHeight = nHeight;

	if(bRecalc)
		CalcLayout();
}
void CBoxList::SetBtnColors(BtnClr *normal, BtnClr *hover, BtnClr *down)
{
	if(normal) btn_normal = *normal;
	if(hover) btn_hover = *hover;
	if(down) btn_down = *down;
}





/////////////////////////////////////////////////////////////////////
// Operations

void CBoxList::ShowButtons(bool bShow, bool bRecalc)
{
	m_bShowButtons = bShow;

	if(bRecalc)
		CalcLayout();
}


void CBoxList::ScrollTo(int nPos, bool bRedraw)
{
	m_nScrollPos = nPos;

	if(m_nScrollPos < 0) m_nScrollPos = 0;
	if(m_nScrollPos > m_nListSize-m_nPageSize) m_nScrollPos = m_nListSize-m_nPageSize;

	if(bRedraw)
	{
		m_wndScrollBar.SetScrollPos(m_nScrollPos, FALSE);
		Invalidate(FALSE);
	}
}


void CBoxList::ScrollBy(int nScroll, bool bRedraw)
{
	m_nScrollPos += nScroll;

	if(m_nScrollPos < 0) m_nScrollPos = 0;
	if(m_nScrollPos > m_nListSize-m_nPageSize) m_nScrollPos = m_nListSize-m_nPageSize;

	if(bRedraw)
	{
		m_wndScrollBar.SetScrollPos(m_nScrollPos, FALSE);
		Invalidate(FALSE);
	}
}


void CBoxList::SetSelected(int nSel, bool bAlwaysScroll)
{
	m_nSelected = nSel;

	if(nSel < m_nScrollPos || bAlwaysScroll)
		ScrollTo(nSel);
	else if(nSel >= m_nScrollPos + m_nPageSize)
		ScrollTo(nSel - m_nPageSize + 1);

	Invalidate(FALSE);
}


void CBoxList::KillMouseFocus()
{
	// This should be called by the parent window on a mousemove event
	m_bHasMouse = false;
	m_MouseHover = -1;
	m_MousePos = (-1,-1);
	Invalidate(TRUE);
}


