// EditorPaletteDlg.cpp : implementation file
//

#include "stdafx.h"
#include "MapEdit.h"
#include "EditorPaletteDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CEditorPaletteDlg dialog


CEditorPaletteDlg::CEditorPaletteDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CEditorPaletteDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CEditorPaletteDlg)
	//}}AFX_DATA_INIT

	tileset = 0;
	palette = 0;
}


void CEditorPaletteDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CEditorPaletteDlg)
	DDX_Control(pDX, IDC_SLIDER_R, m_Slider_R);
	DDX_Control(pDX, IDC_SLIDER_G, m_Slider_G);
	DDX_Control(pDX, IDC_SLIDER_B, m_Slider_B);
	DDX_Control(pDX, IDC_COMBO_FLASHTYPE, m_Combo_Flash);
	DDX_Control(pDX, IDC_COMBO_SPRITEPAL, m_Combo_Spritepal);
	DDX_Control(pDX, IDC_COMBO_TILESET, m_Combo_Tileset);
	DDX_Control(pDX, IDC_COMBO_PAL, m_Combo_Pal);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CEditorPaletteDlg, CDialog)
	//{{AFX_MSG_MAP(CEditorPaletteDlg)
	ON_CBN_EDITCHANGE(IDC_COMBO_TILESET, OnEditchangeComboTileset)
	ON_CBN_SELCHANGE(IDC_COMBO_TILESET, OnSelchangeComboTileset)
	ON_CBN_EDITCHANGE(IDC_COMBO_PAL, OnEditchangeComboPal)
	ON_CBN_SELCHANGE(IDC_COMBO_PAL, OnSelchangeComboPal)
	ON_WM_PAINT()
	ON_BN_CLICKED(IDC_BTN_APPLY, OnBtnApply)
	ON_BN_CLICKED(IDC_CHECK_SPRITEPAL, OnCheckSpritepal)
	ON_BN_CLICKED(IDC_BTN_SAVE, OnBtnSave)
	ON_WM_LBUTTONDOWN()
	ON_WM_HSCROLL()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CEditorPaletteDlg message handlers


BOOL CEditorPaletteDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here
	SelectPalette(tileset,palette);

	sel_pal = 0;
	sel_color = 0;

	m_Slider_R.SetRange(0, 31);
	m_Slider_G.SetRange(0, 31);
	m_Slider_B.SetRange(0, 31);
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CEditorPaletteDlg::OnPaint() 
{
	CPaintDC dc(this); // device context for painting

	CDC buf;
	CBitmap bmp;

	buf.CreateCompatibleDC(&dc);
	bmp.CreateCompatibleBitmap(&dc, 600, 400);

	CBitmap *pBmpOld = buf.SelectObject(&bmp);


	buf.FillSolidRect(195, 95, 260, 200, GetSysColor(COLOR_3DFACE));

	for(int i = 0; i < 6; i++)
	{
		int x = 200;
		int y = 100 + i*15;

		Rom->map_pal[tileset][palette][i].Draw(&buf, x, y, TRUE);
		palrect[i] = CRect(x, y, x+16*5, y+14);
	}

	int x = palrect[sel_pal].left;
	int y = palrect[sel_pal].top;

	buf.Draw3dRect(x-2, y-2, 16*5+4, 14, 0, 0);


	// blit the buffer to the window
	dc.BitBlt(195, 95, 90, 100, &buf, 195, 95, SRCCOPY);


	DrawPalette(&dc, 300, 100, 8, Rom->map_pal[tileset][palette][sel_pal]);

	Palette pal = Rom->map_pal[tileset][palette][sel_pal];

	dc.FillSolidRect(410, 185, 32, 32, RGB(pal[sel_color].r*8, pal[sel_color].g*8, pal[sel_color].b*8));


	buf.SelectObject(pBmpOld);
}


void CEditorPaletteDlg::OnLButtonDown(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default

	for(int i = 0; i < 6; i++)
	{
		if(palrect[i].PtInRect(point))
		{
			sel_pal = i;
			Invalidate(FALSE);
		}
	}

	for(i = 0; i < 16; i++)
	{
		if(colorrect[i].PtInRect(point))
		{
			sel_color = i;
			Invalidate(FALSE);
		}
	}
	
	CDialog::OnLButtonDown(nFlags, point);
}


void CEditorPaletteDlg::SelectPalette(int set, int pal)
{
	tileset = set;
	palette = pal;

	for(int i = 0; i < 8; i++)
		m_Combo_Pal.DeleteString(0);

	for(i = 0; i < Rom->map_pal[set].num_pal; i++)
	{
		CString s;
		s.Format("%d", i);
		m_Combo_Pal.AddString(s);
	}

	m_Combo_Tileset.SetCurSel(set);
	m_Combo_Pal.SetCurSel(pal);

	DisplayData();
}


void CEditorPaletteDlg::DrawPalette(CDC *dc, int x, int y, int w, Palette pal)
{
	for(int i = 0; i < 16; i++)
	{
		int xx = x + ((i%w)*18);
		int yy = y + ((i/w)*18);

		CRect rect(xx, yy, xx+16, yy+16);

		colorrect[i] = rect;

		dc->DrawFrameControl(rect, DFC_BUTTON, DFCS_BUTTONPUSH|DFCS_PUSHED);
		rect.DeflateRect(1,1,1,1);
		dc->FillSolidRect(rect, RGB(pal[i].r*8, pal[i].g*8, pal[i].b*8));
	}
}



void CEditorPaletteDlg::DisplayData()
{
	SetDlgItemText(IDC_EDIT_EVENT, FormatSNESAdr(Rom->map_pal[tileset][palette].event));
	SetDlgItemText(IDC_EDIT_PALPTR, FormatSNESAdr(Rom->map_pal[tileset][palette].pal_ptr));

	m_Combo_Flash.SetCurSel(Rom->map_pal[tileset][palette].flash);


	if(Rom->map_pal[tileset][palette].spr_pal == 0)
	{
		CheckDlgButton(IDC_CHECK_SPRITEPAL, false);
		GetDlgItem(IDC_COMBO_SPRITEPAL)->EnableWindow(false);
		SetDlgItemInt(IDC_COMBO_SPRITEPAL, 0);
	}
	else
	{
		CheckDlgButton(IDC_CHECK_SPRITEPAL, true);
		GetDlgItem(IDC_COMBO_SPRITEPAL)->EnableWindow(true);
		SetDlgItemInt(IDC_COMBO_SPRITEPAL, Rom->map_pal[tileset][palette].spr_pal-2);
	}
	


	Invalidate(FALSE);
}



void CEditorPaletteDlg::OnEditchangeComboTileset() 
{
	// TODO: Add your control notification handler code here
	int sel = GetDlgItemInt(IDC_COMBO_TILESET);

	if(sel < 32)
	{
		((CComboBox*)GetDlgItem(IDC_COMBO_TILESET))->SetCurSel(sel);
	}
}


void CEditorPaletteDlg::OnSelchangeComboTileset() 
{
	// TODO: Add your control notification handler code here
	int sel = ((CComboBox*)GetDlgItem(IDC_COMBO_TILESET))->GetCurSel();

	SelectPalette(sel, palette);
}


void CEditorPaletteDlg::OnEditchangeComboPal() 
{
	// TODO: Add your control notification handler code here
	int sel = GetDlgItemInt(IDC_COMBO_PAL);

	if(sel < Rom->map_pal[tileset].num_pal)
	{
		m_Combo_Pal.SetCurSel(sel);
	}
	else
	{
		sel = Rom->map_pal[tileset].num_pal-1;

		m_Combo_Pal.SetCurSel(sel);
	}
	SelectPalette(tileset, sel);
}


void CEditorPaletteDlg::OnSelchangeComboPal() 
{
	int sel = m_Combo_Pal.GetCurSel();
	SelectPalette(tileset, sel);
}


void CEditorPaletteDlg::OnBtnApply() 
{
	ApplyChanges();
}


void CEditorPaletteDlg::ApplyChanges()
{
	CString s;
	GetDlgItemText(IDC_EDIT_EVENT, s);
	Rom->map_pal[tileset][palette].event = DeformatSNESAdr(s);

	GetDlgItemText(IDC_EDIT_PALPTR, s);
	Rom->map_pal[tileset][palette].pal_ptr = DeformatSNESAdr(s);

	Rom->map_pal[tileset][palette].flash = m_Combo_Flash.GetCurSel();

	if(IsDlgButtonChecked(IDC_CHECK_SPRITEPAL))
		Rom->map_pal[tileset][palette].spr_pal = GetDlgItemInt(IDC_COMBO_SPRITEPAL) + 2;
	else
		Rom->map_pal[tileset][palette].spr_pal = 0;
}

void CEditorPaletteDlg::OnCheckSpritepal() 
{
	// TODO: Add your control notification handler code here
	GetDlgItem(IDC_COMBO_SPRITEPAL)->EnableWindow(IsDlgButtonChecked(IDC_CHECK_SPRITEPAL));
}


void CEditorPaletteDlg::OnBtnSave() 
{
	// TODO: Add your control notification handler code here
	Rom->Save();
}




void CEditorPaletteDlg::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	// TODO: Add your message handler code here and/or call default
	CSliderCtrl *s = (CSliderCtrl*)pScrollBar;

	if(slider_update)
		switch(s->GetDlgCtrlID())
		{
		case IDC_SLIDER_R:
			ChangeColor(s->GetPos(), -1, -1);
			break;
		case IDC_SLIDER_G:
			ChangeColor(-1, s->GetPos(), -1);
			break;
		case IDC_SLIDER_B:
			ChangeColor(-1, -1, s->GetPos());
			break;
		}

	CDialog::OnHScroll(nSBCode, nPos, pScrollBar);
}


void CEditorPaletteDlg::SetBarPos(int r, int g, int b)
{
	m_Slider_R.SetPos(r);
	m_Slider_G.SetPos(g);
	m_Slider_B.SetPos(b);
}

void CEditorPaletteDlg::ChangeColor(int r, int g, int b)
{
	if(r != -1) Rom->map_pal[tileset][palette][sel_pal][sel_color].r = r;
	if(g != -1) Rom->map_pal[tileset][palette][sel_pal][sel_color].g = g;
	if(b != -1) Rom->map_pal[tileset][palette][sel_pal][sel_color].b = b;

	Invalidate(FALSE);
}
