// Rom.cpp: implementation of the CRom class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MapEdit.h"
#include "Rom.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CRom::CRom()
{
	is_loaded = FALSE;
	next_adr = 0xF0200;

	music_playing = -1;

	gcount = 0;
}

CRom::~CRom()
{
	// Might want to clean up...
	for(int i = 0; i < 768; i++)
	{
		delete comprTable[i];
	}
}




// all functions should eventually go here...

void CRom::LoadEnemyMap()
{
	// seek to data location
	fseek(RomFile, 0x101A80, SEEK_SET);

	for(int y = 0; y < 160; y++)
	{
		for(int x = 0; x < 128; x++)
		{
			enemy_map[y][x] = fgetc(RomFile);
			fgetc(RomFile);
		}
	}

	fflush(RomFile);
}

void CRom::LoadSectorMap()
{
	// seek to data location
	fseek(RomFile, 0x17B400, SEEK_SET);

	for(int y = 0; y < 80; y++)
	{
		for(int x = 0; x < 32; x++)
		{
			sector_map[y][x] = fgetc(RomFile);
			sector_map[y][x]+= fgetc(RomFile) << 8;
		}
	}
}

void CRom::WriteSectorMap()
{
	// seek to data location
	fseek(RomFile, 0x17B400, SEEK_SET);

	for(int y = 0; y < 80; y++)
	{
		for(int x = 0; x < 32; x++)
		{
			fputc(sector_map[y][x] & 255, RomFile);
			fputc((sector_map[y][x] >> 8) & 255, RomFile);
		}
	}
}





void CRom::LoadEnemyData()
{
	long int      offset = 0;
	int           a;
	int           b;
	unsigned char ch;

	fseek(RomFile, 0x1597e7, SEEK_SET);
	fflush(RomFile);

	for (a = 0; a < 230; a++)
	{
		enemies[a].number = a;
		enemies[a].address = 1415143 + offset;
		enemies[a].theFlag = fgetc(RomFile);

		for (b = 0; b < 25; b++)
		{
			ch = fgetc(RomFile);

			if (ch == 0)
				enemies[a].name[b] = 0;
			else
				enemies[a].name[b] = ch - 48;
		}

		ch = fgetc(RomFile);  // Unknown, varies from 0-3 or so
						// unknown5.txt

		enemies[a].type = fgetc(RomFile); // 0 - normal, 1 - insect, 2 - metal

		enemies[a].insidePic = fgetc(RomFile);
		enemies[a].insidePic += (fgetc(RomFile) << 8);

		enemies[a].outsidePic = fgetc(RomFile);
		enemies[a].outsidePic += (fgetc(RomFile) << 8);

		enemies[a].runFlag = fgetc(RomFile);

		enemies[a].hp = fgetc(RomFile);
		enemies[a].hp += (fgetc(RomFile) * 256);

		enemies[a].pp = fgetc(RomFile);
		enemies[a].pp += (fgetc(RomFile) * 256);

		enemies[a].exp = fgetc(RomFile);
		enemies[a].exp += (fgetc(RomFile) * 256);
		enemies[a].exp += (fgetc(RomFile) * 256 * 256);
		enemies[a].exp += (fgetc(RomFile) * 256 * 256 * 256);

		enemies[a].money = fgetc(RomFile);
		enemies[a].money += (fgetc(RomFile) * 256);

		enemies[a].movement = fgetc(RomFile);
		ch = fgetc(RomFile); // Appears to always be 0

		enemies[a].startAddress = fgetc(RomFile);
		enemies[a].startAddress += (fgetc(RomFile) << 8);
		enemies[a].startAddress += (fgetc(RomFile) << 16);
		enemies[a].startAddress += (fgetc(RomFile) << 24);

		enemies[a].dieAddress = fgetc(RomFile);
		enemies[a].dieAddress += (fgetc(RomFile) << 8);
		enemies[a].dieAddress += (fgetc(RomFile) << 16);
		enemies[a].dieAddress += (fgetc(RomFile) << 24);

		enemies[a].palette = fgetc(RomFile);

		enemies[a].level = fgetc(RomFile);

		enemies[a].music = fgetc(RomFile);

		enemies[a].offense = fgetc(RomFile);

		ch = fgetc(RomFile); // Appears to always be 0

		enemies[a].defense = fgetc(RomFile);

		ch = fgetc(RomFile);  // Unknown, but mostly 0, sometimes 1
						// unknown4.txt

		enemies[a].speed = fgetc(RomFile);
		enemies[a].guts = fgetc(RomFile);
		enemies[a].iq = fgetc(RomFile);

		ch = fgetc(RomFile);  // Appears to vary from 0-4 or so
		ch = fgetc(RomFile);  // Appears to vary from 0-3
		ch = fgetc(RomFile);  // Appears to vary from 0-3
		ch = fgetc(RomFile);  // Appears to vary from 0-3
		ch = fgetc(RomFile);  // Appears to vary from 0-3
		enemies[a].missRate = fgetc(RomFile);  // 0 = almost never, 255 = always

		enemies[a].order = fgetc(RomFile);

		enemies[a].action[0] = fgetc(RomFile);
		enemies[a].action[0] += (fgetc(RomFile) << 8);
		enemies[a].action[1] = fgetc(RomFile);
		enemies[a].action[1] += (fgetc(RomFile) << 8);
		enemies[a].action[2] = fgetc(RomFile);
		enemies[a].action[2] += (fgetc(RomFile) << 8);
		enemies[a].action[3] = fgetc(RomFile);
		enemies[a].action[3] += (fgetc(RomFile) << 8);

		enemies[a].finalAction = fgetc(RomFile);
		enemies[a].finalAction += (fgetc(RomFile) << 8);

		enemies[a].amigo[0] = fgetc(RomFile);
		enemies[a].amigo[1] = fgetc(RomFile);
		enemies[a].amigo[2] = fgetc(RomFile);
		enemies[a].amigo[3] = fgetc(RomFile);

		enemies[a].finalArg = fgetc(RomFile);  // dunno, always 0, except Carbon Dog, who has 83
						// Mr. A says: could this be an argument for the final action?

		ch = fgetc(RomFile);  // seems to be size, or maybe precedence, possibly
						// linked to calling enemies, unknown3.txt

		enemies[a].bossFlag = fgetc(RomFile);

		enemies[a].freq = fgetc(RomFile);
		enemies[a].item = fgetc(RomFile);

		enemies[a].status = fgetc(RomFile);
		enemies[a].dieSound = fgetc(RomFile);

		ch = fgetc(RomFile);  // Has to do with battle layout, 0 or 1
						// unknown2.txt

		enemies[a].maxCall = fgetc(RomFile);  // Very max = 8

		ch = fgetc(RomFile);  // Unknown, numbers vary somewhat, maybe enemy type
						// unknown1.txt

		offset += 94;
	}
}


void CRom::LoadItemData()
{
   // Assumes rom is in read binary
   // Assumes item is an array of 256 Item types

   char ch;
   int  offset = 0;

   fseek(RomFile, 0x155200, SEEK_SET);
   fflush(RomFile);

   for (int i = 0; i < 256; i++)
   {
      items[i].number = i;

      items[i].address = 0x155200 + offset;

      for (int j = 0; j < 25; j++)
      {
         ch = fgetc(RomFile);

         if (ch == 0)
            items[i].name[j] = 0;
         else
            items[i].name[j] = ch - 48;
      }

      items[i].type = fgetc(RomFile);
      items[i].cost = fgetc(RomFile);
      items[i].cost += (fgetc(RomFile) << 8);
      items[i].ownership = fgetc(RomFile);
      items[i].effect = fgetc(RomFile);
      items[i].effect += (fgetc(RomFile) << 8);
      items[i].strength = fgetc(RomFile);
      items[i].increase = fgetc(RomFile);
      items[i].extraPower = fgetc(RomFile);
      items[i].specialProperties = fgetc(RomFile);
      items[i].descAddress = fgetc(RomFile);
      items[i].descAddress += (fgetc(RomFile) << 8);
      items[i].descAddress += (fgetc(RomFile) << 16);
      items[i].descAddress += (fgetc(RomFile) << 24);

      offset += 39;
   }
}



void CRom::SaveEnemyMap()
{
	if(!is_loaded) return;

	// seek to data location
	fseek(RomFile, 0x101A80, SEEK_SET);

	for(int y = 0; y < 160; y++)
	{
		for(int x = 0; x < 128; x++)
		{
			fputc(enemy_map[y][x], RomFile);
			fputc(0, RomFile);
		}
	}

	fflush(RomFile);
}









#define _R(c) ((c>>16)&255)
#define _G(c) ((c>>8)&255)
#define _B(c) (c&255)

void CRom::DrawMain(CDC *buffer, int cx, int cy, int nx, int ny, int xOffset, int yOffset, int sel_sector, SelectedSector *sectorlist)
{
	// nx = width
	// ny = height
	// x/yOffset = position

	// NEW:
	// cx,cy = buffer offset
	//

	// if cx and cy are zero, the entire map will be rendered

	// if cx is, for example, screen_w/32-1, and cy = 0,
	// only the right edge will be drawn

	BITMAPINFO bmi = { { 40, nx*32, ny*-32/*-232*/, 1, 32, BI_RGB, 0, 0, 0, 0, 0 } };
	HBITMAP bm;
	int *data;
	int xx, yy, x, y, tx, ty;
	bm = CreateDIBSection((HDC)*buffer, &bmi, DIB_RGB_COLORS, (void **)&data, NULL, 0);

	bool overlay = sel_sector>0;
	int clr = 0;
	int gray = 0;
	int R,G,B;

	unsigned int set,_set,tile,palette,b;

	if(is_loaded) for(yy = 0; yy < ny; yy++)
	{
		for(xx = 0; xx < nx; xx++)
		{
			if(sel_sector)
			{
				for(int i = 0; i < sel_sector; i++)
				{
					if(xx+xOffset >= sectorlist[i].x*8 &&
					   xx+xOffset <  sectorlist[i].x*8 + 8 &&
					   yy+yOffset >= sectorlist[i].y*4 &&
					   yy+yOffset <  sectorlist[i].y*4 + 4)
					   overlay = 0;
				}
			}

			if(xx+xOffset >= 0 && yy+yOffset >= 0 && xx+xOffset < 256 && yy+yOffset < 320) {
				// draw all 16 minitiles in a specific map tile
				for(y = 0; y < 4; y++) for(x = 0; x < 4; x++) {
					//if(yy+yOffset >= 320 || xx+xOffset 
					set	= global_sets[(yy+yOffset)/4][(xx+xOffset)/8];
					_set = set_conv[set];
					tile = map_tiles[yy+yOffset][xx+xOffset] + (local_sets[yy+yOffset][xx+xOffset]*256);
					palette = global_pal[(yy+yOffset)/4][(xx+xOffset)/8];

					b = arranges[_set][tile][y*4+x];//[(xx + 0) * 4 + yy][y * 4 + x];
					//b = x;
					// draw all pixels in a minitile
					for(ty = 0; ty < 8; ty++) for(tx = 0; tx < 8; tx++) {
						// x,y = minitile coordinate, so multiplied by 8 to draw in correct minitile
						clr = mpal[set] [palette] [((b & 0x1C00) >> 10) - 2] [tilecolor(_tilesets[_set][b & 0x1FF], tx, ty)];

						if(overlay)
						{
							gray = (_R(clr) + _G(clr) + _B(clr))/3;

							R = (_R(clr)/12 + gray);
							G = (_G(clr)/12 + gray);
							B = (_B(clr)/12 + gray);

							if(R > 255) R = 255;
							if(G > 255) G = 255;
							if(B > 255) B = 255;

							clr = (R<<16) | (G<<8) | (B);
						}

						data[32*xx + 8*x + (b & 0x4000 ? 7 - tx : tx) + nx*32 * 
							(32*yy + 8*y + (b & 0x8000 ? 7 - ty : ty))] = clr;
							//mpal[set] [palette] [((b & 0x1C00) >> 10) - 2] [tilecolor(_tilesets[_set][b & 0x1FF], tx, ty)];
					}
				}
			}
			overlay = sel_sector>0;
		}
	}

	SetDIBitsToDevice((HDC)*buffer, cx*32, cy*32, nx*32, ny*32, 0, 0, 0, ny*32, data, &bmi, DIB_RGB_COLORS);
	DeleteObject(bm);
}

void CRom::DrawTile(CDC *dc, int cx, int cy, int tile_x, int tile_y)
{
	BITMAPINFO bmi = { { 40, 32, -32, 1, 32, BI_RGB, 0, 0, 0, 0, 0 } };
	HBITMAP bm;
	int *data;

	int tile = map_tiles[tile_y][tile_x] + local_sets[tile_y][tile_x]*256;
	int set = global_sets[tile_y/4][tile_x/8];
	int pal = global_pal[tile_y/4][tile_x/8];

	int pset = set;
	set = set_conv[set];

	int /*xx, yy,*/ x, y, tx, ty;
	bm = CreateDIBSection(dc->m_hDC, &bmi, DIB_RGB_COLORS, (void **)&data, NULL, 0);

	// draw all 16 minitiles in a specific map tile
	for(y = 0; y < 4; y++) for(x = 0; x < 4; x++) {
		int b = arranges[set][tile][y*4+x];//[(xx + 0) * 4 + yy][y * 4 + x];
		// draw all pixels in a minitile
		for(ty = 0; ty < 8; ty++) for(tx = 0; tx < 8; tx++) {
			// x = minitile coordinate, so multiplied by 8 to draw in correct minitile
			data[33*0 + 8*x + (b & 0x4000 ? 7 - tx : tx) + 32 * 
				(33*0 + 8*y + (b & 0x8000 ? 7 - ty : ty))] =
				mpal[pset][pal][((b & 0x1C00) >> 10) - 2][
				tilecolor(_tilesets[set][b & 0x1FF], tx, ty)];
		}
	}
	SetDIBitsToDevice(dc->m_hDC, cx*32, cy*32, 32, 32, 0, 0, 0, 32, data, &bmi, DIB_RGB_COLORS);
	DeleteObject(bm);
}





void CRom::DrawView(CDC *buffer, int cx, int cy, int xOffset, int yOffset, bool** clip, int sel_sector, SelectedSector *sectorlist)
{

	// cx = width
	// cy = height
	// x/yOffset = tilemap coordinates

	// clip = drawing properties array

	// sel_sector = number of entries in sectorlist
	// sectorlist = information about selection (parts that should be grayed)

	BITMAPINFO bmi = { { 40, cx*32, cy*-32/*-232*/, 1, 32, BI_RGB, 0, 0, 0, 0, 0 } };
	HBITMAP bm;
	int *data;
	int xx, yy, x, y, tx, ty;
	bm = CreateDIBSection((HDC)*buffer, &bmi, DIB_RGB_COLORS, (void **)&data, NULL, 0);



	//HBITMAP src;
	//src = (HBITMAP)SelectObject((HDC)*buffer, bm);

	//GetDIBits((HDC)*buffer, src, 0, cy*32, (void**)data, &bmi, DIB_RGB_COLORS);

	//SelectObject((HDC)*buffer, src);

	bool overlay = sel_sector>0;
	int clr = 0;
	int gray = 0;
	int R,G,B;

	int set,_set,tile,palette,b;

	if(is_loaded) for(yy = 0; yy < cy; yy++)
	{
		for(xx = 0; xx < cx; xx++)
		{
			if(clip[xx][yy]) continue;

			if(sel_sector)
			{
				for(int i = 0; i < sel_sector; i++)
				{
					if(xx+xOffset >= sectorlist[i].x*8 &&
					   xx+xOffset <  sectorlist[i].x*8 + 8 &&
					   yy+yOffset >= sectorlist[i].y*4 &&
					   yy+yOffset <  sectorlist[i].y*4 + 4)
					   overlay = 0;
				}
			}

			if(xx+xOffset >= 0 && yy+yOffset >= 0 && xx+xOffset < 256 && yy+yOffset < 320) {
				// draw all 16 minitiles in a specific map tile
				for(y = 0; y < 4; y++) for(x = 0; x < 4; x++) {
					set	= global_sets[(yy+yOffset)/4][(xx+xOffset)/8];
					_set = set_conv[set];
					tile = map_tiles[yy+yOffset][xx+xOffset] + (local_sets[yy+yOffset][xx+xOffset]*256);
					palette = global_pal[(yy+yOffset)/4][(xx+xOffset)/8];

					b = arranges[_set][tile][y*4+x];//[(xx + 0) * 4 + yy][y * 4 + x];
					//b = x;
					// draw all pixels in a minitile
					for(ty = 0; ty < 8; ty++) for(tx = 0; tx < 8; tx++) {
						// x,y = minitile coordinate, so multiplied by 8 to draw in correct minitile
						clr = mpal[set] [palette] [((b & 0x1C00) >> 10) - 2] [tilecolor(_tilesets[_set][b & 0x1FF], tx, ty)];

						if(overlay)
						{
							gray = (_R(clr) + _G(clr) + _B(clr))/3;

							R = (_R(clr)/12 + gray);
							G = (_G(clr)/12 + gray);
							B = (_B(clr)/12 + gray);

							if(R > 255) R = 255;
							if(G > 255) G = 255;
							if(B > 255) B = 255;

							clr = (R<<16) | (G<<8) | (B);
						}

						data[32*xx + 8*x + (b & 0x4000 ? 7 - tx : tx) + cx*32 * 
							(32*yy + 8*y + (b & 0x8000 ? 7 - ty : ty))] = clr;
							//mpal[set] [palette] [((b & 0x1C00) >> 10) - 2] [tilecolor(_tilesets[_set][b & 0x1FF], tx, ty)];
					}
				}
			}
			overlay = sel_sector>0;
		}
	}

	SetDIBitsToDevice((HDC)*buffer, 0, 0, cx*32, cy*32, 0, 0, 0, cy*32, data, &bmi, DIB_RGB_COLORS);

	DeleteObject(bm);

	//DeleteObject(src);
}







void CRom::DrawSet(CDC *buffer, int px, int py, int set, int pal, int nx, int ny, int offset)
{
	// draws an array of tiles (in order) from set; width nx, height ny, starting at line offset
		BITMAPINFO bmi = { { 40, 132, -ny*33/*-264*/, 1, 32, BI_RGB, 0, 0, 0, 0, 0 } };
		HBITMAP bm;
		int *data;
		int xx, yy, x, y, tx, ty;
		bm = CreateDIBSection((HDC)*buffer, &bmi, DIB_RGB_COLORS, (void **)&data, NULL, 0);

		// draw all map tiles
		if(is_loaded) for(yy = 0; yy < ny; yy++) for(xx = 0; xx < nx; xx++)

			// draw all 16 minitiles in a specific map tile
			for(y = 0; y < 4; y++) for(x = 0; x < 4; x++)
			{
				int b = arranges[set_conv[set]][xx+(yy+offset)*4/*+(lSet*256)*/][y * 4 + x];
				//b = x;
				// draw all pixels in a minitile
				for(ty = 0; ty < 8; ty++) for(tx = 0; tx < 8; tx++)
				{
					// x = minitile coordinate, so multiplied by 8 to draw in correct minitile
					data[33*xx + 1 + 8*x + (b & 0x4000 ? 7 - tx : tx) + 132 * 
						(33*yy + 1 + 8*y + (b & 0x8000 ? 7 - ty : ty))] =
						mpal[set][pal][((b & 0x1C00) >> 10) - 2][
						tilecolor(_tilesets[set_conv[set]][b & 0x1FF], tx, ty)];
				}
			}


		SetDIBitsToDevice((HDC)*buffer, px, py, 132, ny*33/*264*/, 0, 0, 0, ny*33/*264*/, data, &bmi, DIB_RGB_COLORS);
		DeleteObject(bm);
}









// Returns the door (if any) at the specified location
DoorEntry* CRom::GetDoorAt(int x, int y)
{
	int area = (int(y/32))*32 + (x/32);
	int door = -1;

	int dx = x%32;
	int dy = y%32;

	for(int i = 0; i < door_entry[area].num_doors; i++)
	{
		if(door_entry[area][i].x == dx &&
		   door_entry[area][i].y == dy)
		{
			return &door_entry[area][i];
		}
	}


	return NULL;
}


void CRom::GetDoorAt(int x, int y, int &door_area, int &door_index)
{
	int area = (int(y/32))*32 + (x/32);

	int dx = x%32;
	int dy = y%32;

	door_area = -1;
	door_index = -1;

	for(int i = 0; i < door_entry[area].num_doors; i++)
	{
		if(door_entry[area][i].x == dx &&
		   door_entry[area][i].y == dy)
		{
			door_area = area;
			door_index = i;
		}
	}
}


SpriteEntry* CRom::GetSpriteAt(int x, int y, int &spr_area, int &spr_index)
{
	int area = 0;//(int(y/256))*256 + (x/256);

	int dx = x%256;
	int dy = y%256;

	spr_area = -1;
	spr_index = -1;


	for(int i = 0; i < object_areas[area].num_sprites; i++)
	{
		SpriteEntry s = object_areas[area].sprites[i];
		int w = sprites[tpt[s.tptNum].sprite].width*8;
		int h = sprites[tpt[s.tptNum].sprite].height*8;

		// Visual bounds
		int vx = s.x - (w/2);
		int vy = s.y - (h - 8);

		CRect rect(vx, vy, vx+w, vy+h);
		CPoint pt(dx, dy);

		int q = 1;
		if(rect.PtInRect(pt))
		{
			spr_area = area;
			spr_index = i;
			return &object_areas[area].sprites[i];
		}
	}

/*
*	Sprite/mouse collision code from old buttondown handler in view
*
		for(int i = 0; i < n_disp_spr; i++) {
			int sx = disp_spr[i].x;
			int sy = disp_spr[i].y;
			int sw = disp_spr[i].w;
			int sh = disp_spr[i].h;

			if(x >= sx && x <= sx+sw && y >= sy && y <= sy+sh) {
				drag_area = disp_spr[i].area;
				drag_sprite = disp_spr[i].spr;

				spr_offx = m_x - sx-(sw/2);
				spr_offy = m_y - sy-(sh-8);

				is_sprite_dragging = TRUE;
				bFound = TRUE;
			}
		}/**/

	return NULL;
}




void CRom::GetExpandedRange(int &start, int &end)
{
	// This function should be called during the save procedure,
	// _before_ any data is written. It will search all the data
	// objects in the ROM object to identify the space used in the
	// expanded area, so that it may be wiped prior to writing.

	// This needs to search every single object that could be
	// written into the expanded area.

	// Hmmm... perhaps it would be convenient to have a global
	// datapack containing all expandable entries... 'twould
	// make iteration in this step a wee bit easier.

	CDataPack stuff;
	int i;

	// Add all the data entries to the package
	for(i = 0; i < CEnemyCluster::num_entries; i++)
		stuff.Add(&enemy_clusters[i]);

	for(i = 0; i < CEnemyBattleGroup::num_entries; i++)
		stuff.Add(&enemy_groups[i]);

	for(i = 0; i < CDoorEntry::num_entries; i++)
		stuff.Add(&door_entry[i]);


	// Now, iterate through the entire package
	int n = stuff.GetCount();
	start = 0x400200;
	end = -1;
	for(i = 0; i < n; i++)
	{
		// if the object's address is in the expanded area
		int adr = stuff[i].GetAddress();

		if(adr >= 0x300200)
		{
			// If address is before current starting value, reset start
			if(adr < start)
				start = adr;
			else if(adr > end)
				end = adr + stuff[i].GetSize();
		}
	}
}




bool CRom::PlayMusic(int n)
{
	StopMusic();

	if(n == 0 || n == 4)
	{
		music_playing = n;
		return true;
	}

	char filename[256];
	sprintf(filename, "%s\\music\\%x.spc", Path, n);

	// Make sure the file exists
	FILE* file = fopen(filename, "r");
	if(!file)
		return false;
	else
		fclose(file);

	music_playing = n;

	if(spc.Load(filename))
		spc.Play();

	return true;
}

bool CRom::StopMusic()
{
	spc.Stop();
	music_playing = -1;

	return true;
}



