// Rom.h: interface for the CRom class.
//
// This will be the huge pile of junk that contains all the data
// from the ROM pertinent to this editor's functions. This should
// help make the whole thing a little more object-oriented...
// (_little_ being the keyword :O)
//
// Most of the previously global crap is gonna go in here. The rest
// should get placed where it belongs: within the appropriate
// framework class.
//
//////////////////////////////////////////////////////////////////////



#if !defined(AFX_ROM_H__60F750C0_CAE0_4B4B_B147_AC9D4CB5FC53__INCLUDED_)
#define AFX_ROM_H__60F750C0_CAE0_4B4B_B147_AC9D4CB5FC53__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#define TYPE_CODE 0
#define TYPE_ARGLIST 1


//////////////////////////////////////////////////////////////////////
// Includes

#include "classes.h"
#include "spcplay.h"
#include "romclasses.h"







//////////////////////////////////////////////////////////////////////
// CSprite class


#define NORTH	  0		// Values indicating directional frames
#define EAST	  2
#define SOUTH	  4
#define WEST	  6
#define NORTHEAST 8
#define SOUTHEAST 10
#define SOUTHWEST 12
#define NORTHWEST 14
#define FIRST	  0		// OR'ed with directional flag to retrieve first of two frames
#define SECOND	  1		// OR'ed with directional flag to retrieve second frame

#define FRAME_STD	8	// Normal NPC sprites
#define FRAME_EXTRA	9	// Some enemies using an additional resting frame
#define FRAME_FULL	16	// Only used by PC objects

#define FLAG_MPAL 0x16	// Use palette from map

class CRom;

// Encapsulates one sprite as represented in the sprite table
class CSprite : public CDataEntry
{
//////////////////////////////////////////////////////////////////////
// Default implementation
public:

	enum class_info
	{
		ptr_address = 0x2F153F,			// Address of pointer table
		ptr_size = 4,					// Size of each entry in pointer table
		def_data_start = 0x2F1C7F,		// Location of original data
		def_data_extent = 0x2F4C40,		// Original extent of data
		num_entries = 464,				// Number of entries for this class
	};

	void WriteObject(FILE*, unsigned long address);
	void WritePtr(FILE*, unsigned long address);
	void ReadObject(FILE* file, int index, DWORD data = 0);

	int GetAddress();
	int GetSize();

///////////////////////////////////////////////////////////////////////
// Class-specific data and behavior

	// Actual entry data

	int width;	// Width and height for all frames used in the sprite
	int height;

	unsigned char yy;		// Unknown 1

	unsigned char flags;

	unsigned char uu[4];	// Unknown 2

	unsigned char bank;				// Sprite bank
	unsigned short frame_ptr[16];	// Array of sprite pointers (relative to bank #)


	// Logical redundancies
	unsigned char palette;
	unsigned short frames[16];		// Array of sprite indeces
	bool flip[16];					

	int num_frames;					// Can be either 8, 9, or 16
	bool exframes;


	// Functions
	void Draw(CDC* dc, int x, int y, int frame, CRom* pRom);
	void Draw(CDC* dc, int x, int y, int frame, int area, int entry, CRom* pRom);
};





//////////////////////////////////////////////////////////////////////
// CMapPalette class


// MapPalette structure
// A collection of six palettes for use in map stuffage
struct MapPalette
{
	Palette pal[6];			// The six sixteen-color palettes used

	// Palette flags
	unsigned short event;	// Event flag to toggle this palette
	unsigned short pal_ptr;	// Possible: pointer to post-event palette?
	unsigned short spr_pal;	// Index of palette to use for sprites
	unsigned short flash;	// Flashing style
	unsigned short uu1;
	unsigned short uu2;

	Palette& operator[](int n) { return pal[(n%6)]; };
};



class CMapPalGroup : public CDataEntry
{
//////////////////////////////////////////////////////////////////////
// Default implementation
public:

	enum class_info
	{
		ptr_address = 0x2F12FB,			// Address of pointer table
		ptr_size = 4,					// Size of each entry in pointer table
		def_data_start = 0x1A7EA7,		// Location of original data
		def_data_extent = 0x1AFCA7,		// Original extent of data
		num_entries = 32,				// Number of entries for this class
	};

	void WriteObject(FILE*, unsigned long address);
	void WritePtr(FILE*, unsigned long address);
	void ReadObject(FILE* file, int index, DWORD data = 0);

	int GetAddress();
	int GetSize();


///////////////////////////////////////////////////////////////////////
// Class-specific data and behavior

	int num_pal;			// Number of palettes contained in the group
	MapPalette pal[8];		// Maximum eight palettes per tileset

	MapPalette& operator[](int);	// Returns a reference to the specified palette

};

//////////////////////////////////////////////////////////////////////
// End unimplemented section








//////////////////////////////////////////////////////////////////////
// Structures

struct CCNode {
	int level;				// depth, i.e., CC byte # (sorta)
	short value;			// byte value (-1 for arguments)

	unsigned char type;		// code byte or argument
	unsigned char num_arg;	// number of argument bytes
	unsigned char args[32];	// argument list
	bool arg_multiplier;	// if true, number of arguments remaining is multiplied by the
							// value of the first

	bool isTerminator;

	int num_nodes;			// Number of child nodes
	CCNode *nodes;			// Pointer to array of child nodes
	
	void Free()				// Deletes child nodes before deleting self
	{
		for(int i = 0; i < num_nodes; i++)
		{
			nodes[i].Free();
		}
		delete this;
	};

	// Destructor calls free to delete memory

};

struct String {
	int address;

	char* str;
	int len;
};


struct EnemyBattleGroup
{
	unsigned char entrycount;

	unsigned short runflag;
	unsigned char flaguse;
	unsigned char letterbox;

	struct
	{
		unsigned char num;
		unsigned char enemy;
	} enemies[32];

	// For sanity reasons, battle groups are limited to 32
	// entries, although the EB ROM has no such restriction.
	// (Also 'cuz I'm too lazy to bother with reallocation
	//  of a dynamic array)
} /*BATTLEGROUPS[484]*/;
  // Room for 484 battle entries in pointer table
	


struct EnemyCluster
{
	unsigned short event;	// Event flag that toggles between primary and secondary
	unsigned char p_pri;		// Global probability for group encounters
	unsigned char p_sec;		// Unknown value

	unsigned char n_pri;	// Number of primary entries
	unsigned char n_sec;	// Number of secondary entries

	bool has_primary;
	bool has_secondary;

	struct
	{
		unsigned char prb;
		unsigned short group;
	} primary[8], secondary[8];

	// DEBUG
	
	unsigned long address;
	unsigned char data[64];


	// A maximum of eight groups is allowed in each subcluster
	// (The cumulative probability for all groups in the cluster
	//  must be eight out of eight)
} /*ENEMYCLUSTERS[204]*/;
  // Room for 204 placement groups in pointer table


struct Enemy {
	long          address;			// Address in ROM
	unsigned char theFlag;			//
	char          name[25];
	unsigned char number;
	unsigned char insidePic;
	int           outsidePic;
	unsigned char runFlag;
	unsigned int  hp;
	unsigned int  pp;
	unsigned int  exp;
	unsigned int  money;
	unsigned char speed;
	unsigned int  startAddress;
	unsigned int  dieAddress;
	unsigned char palette;
	unsigned char offense;
	unsigned char defense;
	unsigned char item;
	unsigned char freq;
	unsigned char status;
	int           action[4];		// actions
	int           finalAction;
	unsigned char level;
	unsigned char guts;
	unsigned char iq;
	unsigned char bossFlag;
	unsigned char dieSound;
	unsigned char movement;
	unsigned char order;
	unsigned char amigo[4];			// action arguments
	unsigned char finalArg;
	unsigned char maxCall;
	unsigned char missRate;
	unsigned char type;
	unsigned char music;
	unsigned char unknown;
}/*ENEMY[230]*/;


struct Item
{
   long          address;
   char          name[25];
   unsigned char number;
   unsigned char type;
   int           cost;
   unsigned char ownership;
   int           effect;
   char          strength;
   unsigned char increase;
   char          extraPower;
   unsigned char specialProperties;
   unsigned int  descAddress;
}/*ITEM[254]*/;




//////////////////////////////////////////////////////////////////////
// CRom class

class CRom  
{
public:
	void WriteSectorMap();
	void LoadSectorMap();
	void DrawTile(CDC *dc, int cx, int cy, int tile_x, int tile_y);
	bool StopMusic();
	bool PlayMusic(int n);
	SpriteEntry* GetSpriteAt(int x, int y, int &spr_area, int &spr_index);
	void GetExpandedRange(int &start, int &end);
	void GetDoorAt(int x, int y, int& door_area, int& door_index);
	DoorEntry* GetDoorAt(int x, int y);
	void LoadEnemyData();

	void LoadItemData();

	CRom();
	virtual ~CRom();


	void LoadAppData(FILE* rom);
	void CreateAppData(SystemEntry *appdata);

	// loading/saving functions
	void Load();
	void Save();

	//void LoadDoorEntries(int &first, int &last, bool firstload, void _cdecl update(int));
	//void WriteEntryTable(unsigned int address, int &first, int &last, void _cdecl update(int));

	void WriteDoorDestinations(int &next_adr, void _cdecl update(int));


	void LoadEnemyMap();
	void SaveEnemyMap();


	// data modification functions
	void AddDestination(int type, int pointer, int x, int y, int direction, int eventFlag, bool checkFlag, int style);
	void RemoveDestination(int destination);
	void ChangeDestination(int destination, int pointer, int x, int y, int eventFlag, bool checkFlag, int style);


	//void AddDoor(int area, int x, int y, int flags, int dest, int destindex, int dir, int rope, bool esc_end);
	//void RemDoor(int area, int door);
	//void ChangeDoor(int area, int door, int x, int y, int flags, int dest);
	

	void AddSprite(int area, int x, int y, int tptNum);
	void RemSprite(int area, int sprite);
	void ChangeSprite(int area, int sprite, int x, int y, int tptNum);


	//int GetDETLength();

	void DrawView(CDC *buffer, int cx, int cy, int xOffset, int yOffset, bool** clip, int sel_sector = 0, SelectedSector *sectorlist = NULL);

	void DrawMain(CDC *dc, int cx, int cy, int nx, int ny, int xOffset, int yOffset, int sel_sector = 0, SelectedSector *sectorlist = NULL);
	void DrawSet(CDC *buffer, int x, int y, int set, int pal, int nx, int ny, int offset);



	// Text functions
void CreateCCTable(char* filename, CCNode table[32]);
CCNode* AddNode(CCNode *node, int level, int value, int type, int num_arg, bool multi, bool terminate);

void CreateCompressionTable(FILE* rom, char* tbl[]);

String* ReadString(FILE* rom, unsigned long address, CCNode ccTable);
char* ParseHeader(char str[], int len, CCNode ccTable, char* comprTable[]);
//char* ParseString(char str[], int len, CCNode ccTable, char* comprTable[]);
//char* ParseGrouping(char str[]);

//int GetStringLength(char str[]);
//char* DeparseString(char str[]);

//int WriteString(FILE* rom, String s, int address);

int SearchNode(CCNode *node, int val);
void AddCode(char* str, int val, int pos);
bool isCC(char ch);

//int LoadRawText(FILE* rom, String strlist[], CCNode ccTable, char* comprTable[]);
//int LoadEXPText(FILE *rom, String strlist[], CCNode ccTable);
//int LoadTPTText(FILE* rom, String strlist[], CCNode ccTable);

//bool CheckFile(CString szFile);

//unsigned int HexToSNES(unsigned int);
//unsigned int SNESToHex(unsigned int);



/////////////////////////////////////////////////////////////////////
//	Global ROM data
	CString FileName;					// filename currently being edited
	CString Path;
	FILE* RomFile;						// one and only file object
	BOOL is_loaded;// = FALSE;			// TRUE if a file is currently loaded
	


/////////////////////////////////////////////////////////////////////
//	ROM System data
	int next_adr;// = 0xF0200;			// free space counter

	SystemArea sys;						// system area
	SystemEntry *appdata;				// application-specific system data

	int det_first;						// address of start of door entry table
	int det_last;						// address of end of door entry table
	int ddt_last;						// address of end of door destination table

	bool FIRSTLOAD;						// TRUE if ROM has not been edited with this app before


/////////////////////////////////////////////////////////////////////
//	Map data
//	Note: All matrices are arranged with [y][x] dimensions
	unsigned char map_tiles[320][256];			// tilemap
	unsigned char local_sets[320][256];			// local tileset data
	unsigned char global_sets[80][32];			// global tileset data
	unsigned char global_pal[80][32];			// palette overlay
	unsigned char global_music[80][32];			// music overlay
	unsigned short enemy_map[320][128];			// map overlay for enemy data
	unsigned short sector_map[80][32];			// map sector properties

	ObjectArea object_areas[1280];		// 8x8 areas for object placement

/////////////////////////////////////////////////////////////////////
// Doors

	CDoorEntry door_entry[1280];



/////////////////////////////////////////////////////////////////////
//	Enemy data

	CEnemyBattleGroup enemy_groups[484];
	CEnemyCluster enemy_clusters[204];

	//EnemyBattleGroup enemy_groups[484];		//	Battle groups
	//EnemyCluster	 enemy_clusters[204];	//	Map placement clusters
	Enemy			 enemies[230];			//	Enemy stats


/////////////////////////////////////////////////////////////////////
//	Item data
	Item items[254];	// Items


/////////////////////////////////////////////////////////////////////
//	Door data
	DestinationList dest;					// list of door destinations


/////////////////////////////////////////////////////////////////////
//	Music data
	MapMusicEntry musictable[164];			// map music table
	CSPCPlayer spc;							// muzak player :D
	int music_playing;						// currently playing music


/////////////////////////////////////////////////////////////////////
//	Sprite data
	
	CSprite sprites[464];		// Sprite table
	TPTEntry tpt[1583];			// Text pointer table (non-expandable for now)

	SpriteGfx gfx2[1300];		// Sprite graphics
	//SpriteGfx sprite_gfx[1144];	// The actual sprite graphics
	Palette pal[8];				// Universal sprite palettes


	int gcount;


/////////////////////////////////////////////////////////////////////
//	Text data
	char* comprTable[768];	// compression string table
	CCNode ccTable;			// initial control tree node


/////////////////////////////////////////////////////////////////////
//	Graphics data
	CMapPalGroup map_pal[32];					// Map palettes (SNES arrangement)

	int				set_conv[32];				// Converter from graphics set to tileset
	char			_tilesets[20][1024][32];	// Tileset graphics
	unsigned short	arranges[20][1024][16];		// Tile arrangements
	unsigned int tileset_pointers[20];			// 0x2F125B
	unsigned int arrange_pointers[20];			// 0x2F12AB
	unsigned int mpal[32][8][6][16];			// Map palettes (RGB)
	unsigned short mpal_snes[32][8][6][16];		// Map palettes (SNES);
};



#endif // !defined(AFX_ROM_H__60F750C0_CAE0_4B4B_B147_AC9D4CB5FC53__INCLUDED_)
