#include "stdafx.h"
#include "classes.h"
#include "mapstuff.h"

#include "Rom.h"


void LoadSystemArea(FILE* rom, SystemArea *sys)
{
	for(int i = 0; i < 32; i++) {
		int address = GetPointer(rom, 0x2E9400+(i*0x10));

		sys->entries[i].index = i;

		if(address != 0) {
			sys->entries[i].is_valid = true;
			sys->entries[i].address = address;

			fseek(rom, 0x2E9400+(i*0x10)+4, SEEK_SET);

			sys->entries[i].size = fgetc(rom);
			sys->entries[i].size += fgetc(rom) << 8;

			sys->entries[i].data = new unsigned char[sys->entries[i].size];

			fseek(rom, sys->entries[i].address, SEEK_SET);

			for(int j = 0; j < sys->entries[i].size; j++) {
				sys->entries[i].data[j] = fgetc(rom);
			}
		}
		else
			sys->entries[i].is_valid = false;
	}
}

void WriteSystemArea(FILE* rom, SystemArea sys)
{
	int offset = 0;		// how much data we've already written

	for(int i = 0; i < 32; i++) {
		// if it's valid, we'll need to give it an address and write its data + pointer
		if(sys.entries[i].is_valid == true) {
			int address = 0x2E9600+offset;		// address to write data to
												// (0x2E9600 is the base for system data)

			// write pointer and size
			PutPointer(rom, 0x2E9400+i*0x10, address);

			fseek(rom, 0x2E9400+(i*0x10)+4, SEEK_SET);
			fputc(sys.entries[i].size & 255, rom);
			fputc((sys.entries[i].size >> 8) & 255, rom);

			// now write the data
			fseek(rom, address, SEEK_SET);
			for(int j = 0; j < sys.entries[i].size; j++) {
				fputc(sys.entries[i].data[j], rom);
			}
			offset+=sys.entries[i].size;
		}
	}
	fflush(rom);
}


void GetROMStuff(FILE* rom, unsigned char mt[320][256], unsigned char ls[320][256], unsigned char gs[80][32], unsigned char gp[80][32], unsigned char gm[80][32])
{
	int i, j;

	// map tiles
	for(i = 0; i < 320; i++) {
		fseek(rom, line_adr[i], SEEK_SET);
		for(j = 0; j < 256; j++) {
			mt[i][j] = fgetc(rom);
		}
	}

	// global tilesets
	for(i = 0; i < 80; i++) {
		fseek(rom, 0x17AA00+(i*32), SEEK_SET);
		for(j = 0; j < 32; j++) {
			int ch = fgetc(rom);
			gs[i][j] = (ch &248)>> 3;
			gp[i][j] = ch & 7;
		}
	}

	// music
	for(i = 0; i < 80; i++) {
		fseek(rom, 0x1CD837+(i*32), SEEK_SET);
		for(j = 0; j < 32; j++) {
			gm[i][j] = fgetc(rom)-1;
		}
	}

	// local tilesets
	for(i = 0; i < 80; i++) {
		for(j = 0; j < 256; j++) {
			if(i%2==0)
				fseek(rom, 0x175200+(256*(i/2))+j, SEEK_SET);
			else
				fseek(rom, 0x178200+(256*(i/2))+j, SEEK_SET);
			char ch = fgetc(rom);

			int t1,t2,t3,t4;
			t4 = (ch >> 6) & 3;
			t3 = (ch >> 4) & 3;
			t2 = (ch >> 2) & 3;
			t1 = (ch /**/) & 3;

			int y = (i*4);
			int x = j;

			ls[y][x]   = t1%4;
			ls[y+1][x] = t2%4;
			ls[y+2][x] = t3%4;
			ls[y+3][x] = t4%4;
		}
	}
}



void SetROMStuff(FILE* rom, unsigned char mt[320][256], unsigned char ls[320][256], unsigned char gs[80][32], unsigned char gp[80][32], unsigned char gm[80][32])
{
	int i,j;

	// tile placement
	for(i = 0; i < 320; i++) {
		fseek(rom, line_adr[i], SEEK_SET);
		for(j = 0; j < 256; j++) {
			fputc(mt[i][j], rom);
		}
	}
	fflush(rom);

	// music
	for(i = 0; i < 80; i++) {
		fseek(rom, 0x1CD837+(i*32), SEEK_SET);
		for(j = 0; j < 32; j++) {
			fputc(gm[i][j]+1,rom);
		}
	}
	fflush(rom);

	// global sets/palettes
	for(i = 0; i < 80; i++) {
		fseek(rom, 0x17AA00+(i*32), SEEK_SET);
		for(j = 0; j < 32; j++) {
			int ch = (gs[i][j]<<3) | gp[i][j];
			fputc(ch, rom);
		}
	}
	fflush(rom);

	// local tilesets
	for(i = 0; i < 80; i++) {
		for(j = 0; j < 256; j++) {
			if(i%2==0)
				fseek(rom, 0x175200+(256*(i/2))+j, SEEK_SET);
			else
				fseek(rom, 0x178200+(256*(i/2))+j, SEEK_SET);
			int ch = 0;

			int y = (i*4);
			int x = j;

			ch += ls[y][x];
			ch += ls[y+1][x] << 2;
			ch += ls[y+2][x] << 4;
			ch += ls[y+3][x] << 6;

			fputc(ch, rom);
		}
	}
	fflush(rom);
}




void LoadDoorDestinations(FILE* rom, DestinationList *dest, SystemEntry *appdata, void _cdecl update(int))
{
	// door editor system area data must be loaded before this gets called.

	if(!appdata->is_valid) {
		dest->num_dest = 977;
	}
	else {
		dest->num_dest =  appdata->data[0];
		dest->num_dest += appdata->data[1] << 8;
	}

	// allocate space for destination table
	dest->entries = new DestinationEntry[dest->num_dest];

	int offset = 0;	//start at 0xF0200 + 0

	for(int i = 0; i < dest->num_dest; i++) {
		if(!appdata->is_valid) {
			// load current door type from a static array
			dest->entries[i].type = default_dest_types[i];
		}
		else {
			dest->entries[i].type = appdata->data[i+16];
		}

		fseek(rom, 0xF0200+offset, SEEK_SET);
		dest->entries[i].address = offset;
		
		// initialize values
		dest->entries[i].eventFlag = 0;
		dest->entries[i].checkFlag = 0;
		dest->entries[i].direction = 0;
		dest->entries[i].style = 0;
		dest->entries[i].x = 0;
		dest->entries[i].y = 0;
		dest->entries[i].pointer = 0;

		// now check stuff, do different things based on door's type.
		int ch;

		switch(dest->entries[i].type) {
		case DEST_DOOR:		// standard door. easy.
			dest->entries[i].pointer = fgetc(rom);
			dest->entries[i].pointer+= fgetc(rom) << 8;
			dest->entries[i].pointer+= fgetc(rom) << 16;
			fgetc(rom);

			dest->entries[i].eventFlag = fgetc(rom);

			ch = fgetc(rom);

			dest->entries[i].eventFlag+= (ch & 0x7F) << 8;
			dest->entries[i].checkFlag = (ch & 0x80) >> 7 != 0;


			dest->entries[i].y = fgetc(rom);

			ch = fgetc(rom);	// temp

			dest->entries[i].y+= (ch & 0x3F) << 8;
			dest->entries[i].direction = (ch & 0xC0) >> 6;


			dest->entries[i].x = fgetc(rom);
			dest->entries[i].x+= fgetc(rom) << 8;

			dest->entries[i].style = fgetc(rom);

			offset+= 11;
			break;

		case DEST_OBJECT: case DEST_PERSON:		// just text pointer
			dest->entries[i].pointer = fgetc(rom);
			dest->entries[i].pointer+= fgetc(rom) << 8;
			dest->entries[i].pointer+= fgetc(rom) << 16;

			offset+= 4;
			break;

		case DEST_SWITCH:	// event flag, pointer
			dest->entries[i].eventFlag = fgetc(rom);

			ch = fgetc(rom);

			dest->entries[i].eventFlag+= (ch & 0x7F) << 8;
			dest->entries[i].checkFlag = (ch & 0x80) >> 7 != 0;

			dest->entries[i].pointer = fgetc(rom);
			dest->entries[i].pointer+= fgetc(rom) << 8;
			dest->entries[i].pointer+= fgetc(rom) << 16;

			offset+= 6;
			break;
		}

		// use function pointer to update progress
		int pos = 1000/dest->num_dest;
		update(pos);
	}
}




void CRom::WriteDoorDestinations(int &next_adr, void _cdecl update(int))
{
	int ch;
	int offset = 0;

	for(int i = 0; i < dest.num_dest; i++) {
		int t = dest.entries[i].type;
		int size = (t == DEST_DOOR ? 11 : ((t == DEST_OBJECT || t == DEST_PERSON) ? 4 : (t == DEST_SWITCH? 6 : -1)));

		int address = next_adr;
		next_adr+= size;

		if(size != 11 && size != 4 && size != 6)
			AfxMessageBox("incorrect destination size!");

		if(address > 0xF8000)
		{
			AfxMessageBox("Door entry write error");
			return;
		}

		fseek(RomFile, address, SEEK_SET);

		dest.entries[i].address = address - 0xF0200;//offset;

		switch(dest.entries[i].type) {
		case DEST_DOOR:
			fputc(dest.entries[i].pointer&255, RomFile);
			fputc((dest.entries[i].pointer >> 8) & 255, RomFile);
			fputc((dest.entries[i].pointer >> 16) & 255, RomFile);
			fputc(0, RomFile);

			ch = dest.entries[i].eventFlag + (dest.entries[i].checkFlag*0x8000);
			fputc(ch & 255, RomFile);
			fputc((ch >> 8) & 255, RomFile);

			ch = dest.entries[i].y;

			fputc(dest.entries[i].y & 255, RomFile);

			fputc(((dest.entries[i].y >> 8) | (dest.entries[i].direction) << 6) & 255, RomFile);

			fputc(dest.entries[i].x & 255, RomFile);
			fputc((dest.entries[i].x >> 8) & 255, RomFile);

			fputc(dest.entries[i].style, RomFile);

			offset+= 11;
			break;

		case DEST_OBJECT:
		case DEST_PERSON:
			fputc(dest.entries[i].pointer & 255, RomFile);
			fputc((dest.entries[i].pointer >> 8) & 255, RomFile);
			fputc((dest.entries[i].pointer >> 16) & 255, RomFile);
			fputc(0, RomFile);

			offset+= 4;
			break;

		case DEST_SWITCH:
			ch = dest.entries[i].eventFlag + (dest.entries[i].checkFlag*0x8000);
			fputc(ch & 255, RomFile);
			fputc((ch >> 8) & 255, RomFile);

			fputc(dest.entries[i].pointer & 255, RomFile);
			fputc((dest.entries[i].pointer >> 8) & 255, RomFile);
			fputc((dest.entries[i].pointer >> 16) & 255, RomFile);
			fputc(0, RomFile);

			offset+= 6;
			break;
		}
		update(-1);
	}

	//ddt_last = 0xF0200+offset;	// end of destination table

	fflush(RomFile);
}





/****OBSOLETE CODE****/
/*void CRom::LoadDoorEntries(int &first, int &last, bool firstload, void _cdecl update(int))
{
/*	for(int i = 0; i < 1280; i++) {
		int address = SNESToHex(GetPointer(RomFile, 0x100200+(i*4)));
		int num_doors;

		fseek(RomFile, address, SEEK_SET);

		object_areas[i].door_address = address;

		num_doors = fgetc(RomFile);
		fgetc(RomFile);

		if(firstload) {
			switch(i) {
			case 49:
			case 81:
			case 97:
			case 98:
			case 140:
			case 300:
			case 333:
			case 395:
			case 562:
			case 604:
			case 613:
			case 681:
			case 688:
			case 752:
			case 815:
			case 816:
			case 911:
			case 1136:
				num_doors--;
				break;
			case 1073:
				num_doors-=2;
				break;
			}
		}

		object_areas[i].num_doors = num_doors;
		if(num_doors < 256 && num_doors >= 0) object_areas[i].doors = new DoorEntry[num_doors];	// allocate space for our door data

		for(int j = 0; j < num_doors; j++) {
			object_areas[i].doors[j].address = address + 2 + (j*5);
			object_areas[i].doors[j].y = fgetc(RomFile);
			object_areas[i].doors[j].x = fgetc(RomFile);
			object_areas[i].doors[j].flags = fgetc(RomFile);

			int ch = fgetc(RomFile);
			ch += fgetc(RomFile) << 8;

			object_areas[i].doors[j].destination =  ch;

			object_areas[i].doors[j].direction = (ch & 0x300) >> 8;

			object_areas[i].doors[j].ropetype = ch;

			object_areas[i].doors[j].is_esc_end = (ch == 0x8000);

		}
		update(!(i%4));
	}
	first = object_areas[0].doors[0].address - 2;
	last = first+GetDETLength();
}*/




/****OBSOLETE CODE****/
/*void CRom::WriteEntryTable(unsigned int address, int &first, int &last, void _cdecl update(int))
{
	/*
	// writes the entry table to the specified address.
	// use FindFreeRange to find the starting address for this data

	// this function also modifies the pointer table as it writes
	// the door entry data.

	int offset = 0;

	// start with the standard loop
	for(int i = 0; i < 1280; i++) {
		PutPointer(RomFile, 0x100200+(i*4), HexToSNES(address+offset));

		fseek(RomFile, address+offset, SEEK_SET);

		fputc(object_areas[i].num_doors, RomFile);
		fputc(0, RomFile);

		for(int j = 0; j < object_areas[i].num_doors; j++) {
			object_areas[i].doors[j].address = address+offset+2;
			fputc(object_areas[i].doors[j].y, RomFile);
			fputc(object_areas[i].doors[j].x, RomFile);
			fputc(object_areas[i].doors[j].flags, RomFile);
			switch(object_areas[i].doors[j].flags) {
			case DEST_SWITCH:
			case DEST_DOOR:
			case DEST_OBJECT:
			case DEST_PERSON:
				fputc(object_areas[i].doors[j].destination & 255, RomFile);
				fputc((object_areas[i].doors[j].destination >> 8) & 255, RomFile);
				break;
			case DEST_ROPE:
				fputc(object_areas[i].doors[j].ropetype & 255, RomFile);
				fputc((object_areas[i].doors[j].ropetype >> 8) & 255, RomFile);
				break;
			case DEST_STAIR:
				fputc(0, RomFile);
				fputc((object_areas[i].doors[j].direction) & 255, RomFile);
				break;
			case DEST_ESCALATOR:
				fputc(0, RomFile);
				if(object_areas[i].doors[j].is_esc_end)
					fputc(0x80, RomFile);
				else
					fputc((object_areas[i].doors[j].direction) & 255, RomFile);
				break;
			}
			offset+=5;
		}
		offset+=2;

		update(-1);
	}
	first = object_areas[0].doors[0].address - 2;
	last = first+GetDETLength();
}*/







/****OBSOLETE CODE****/
/*void CRom::AddDoor(int area, int x, int y, int flags, int dest, int destindex, int dir, int rope, bool esc_end)
{
	DoorEntry *tempdoors = new DoorEntry[object_areas[area].num_doors];

	for(int i = 0; i < object_areas[area].num_doors; i++)
		tempdoors[i] = object_areas[area].doors[i];

	object_areas[area].doors = new DoorEntry[object_areas[area].num_doors+1];

	for(i = 0; i < object_areas[area].num_doors; i++)
		object_areas[area].doors[i] = tempdoors[i];

	int door = object_areas[area].num_doors;

	object_areas[area].doors[door].x = x;
	object_areas[area].doors[door].y = y;
	object_areas[area].doors[door].flags = flags;
	object_areas[area].doors[door].destination = dest;

	object_areas[area].doors[door].destindex = destindex;
	object_areas[area].doors[door].direction = dir;
	object_areas[area].doors[door].ropetype = rope;
	object_areas[area].doors[door].is_esc_end = esc_end;

	delete tempdoors;

	object_areas[area].num_doors++;
}*/

/****OBSOLETE CODE****
/// CROM INTERNALIZED
void CRom::RemDoor(int area, int door)
{
/**	DoorEntry *tempdoors = new DoorEntry[object_areas[area].num_doors];

	for(int i = 0; i < object_areas[area].num_doors; i++)
		tempdoors[i] = object_areas[area].doors[i];

	object_areas[area].num_doors--;

	for(i = door; i < object_areas[area].num_doors; i++)
		tempdoors[i] = tempdoors[i+1];

	object_areas[area].doors = new DoorEntry[object_areas[area].num_doors];

	for(i = 0; i < object_areas[area].num_doors; i++)
		object_areas[area].doors[i] = tempdoors[i];

	delete tempdoors;
}/**/


/****OBSOLETE CODE****/
// CROM INTERNALIZED
/*void CRom::ChangeDoor(int area, int door, int x, int y, int flags, int dest)
{
	object_areas[area].doors[door].x = x;
	object_areas[area].doors[door].y = y;
	object_areas[area].doors[door].flags = flags;
	object_areas[area].doors[door].destination = dest;
}*/


//CROM INTERNALIZED
void CRom::AddSprite(int area, int x, int y, int tptNum)
{
	SpriteEntry *tempsprites = new SpriteEntry[object_areas[area].num_sprites];

	for(int i = 0; i < object_areas[area].num_sprites; i++)
		tempsprites[i] = object_areas[area].sprites[i];

	object_areas[area].sprites = new SpriteEntry[object_areas[area].num_sprites+1];

	for(i = 0; i < object_areas[area].num_sprites; i++)
		object_areas[area].sprites[i] = tempsprites[i];

	int sprite = object_areas[area].num_sprites;

	object_areas[area].sprites[sprite].x = x;
	object_areas[area].sprites[sprite].y = y;
	object_areas[area].sprites[sprite].tptNum = tptNum;

	delete tempsprites;

	object_areas[area].num_sprites++;
}

// CROM INTERNALIZED
void CRom::RemSprite(int area, int sprite)
{
	SpriteEntry *tempsprites = new SpriteEntry[object_areas[area].num_sprites];

	for(int i = 0; i < object_areas[area].num_sprites; i++)
		tempsprites[i] = object_areas[area].sprites[i];

	object_areas[area].num_sprites--;

	for(i = sprite; i < object_areas[area].num_sprites; i++)
		tempsprites[i] = tempsprites[i+1];

	object_areas[area].sprites = new SpriteEntry[object_areas[area].num_sprites];

	for(i = 0; i < object_areas[area].num_sprites; i++)
		object_areas[area].sprites[i] = tempsprites[i];

	delete tempsprites;
}


// CROM INTERNALIZED
void CRom::ChangeSprite(int area, int sprite, int x, int y, int tptNum)
{
	object_areas[area].sprites[sprite].x = x;
	object_areas[area].sprites[sprite].y = y;
	object_areas[area].sprites[sprite].tptNum = tptNum;
}


// CROM INTERNALIZED
void CRom::AddDestination(int type, int pointer, int x, int y, int direction, int eventFlag, bool checkFlag, int style)
{
	// again, the entire array needs to be backed up during this operation
	DestinationEntry *temp = new DestinationEntry[dest.num_dest];

	for(int i = 0; i < dest.num_dest; i++)
		temp[i] = dest.entries[i];

	dest.entries = new DestinationEntry[dest.num_dest+1];

	for(i = 0; i < dest.num_dest; i++)
		dest.entries[i] = temp[i];

	dest.entries[dest.num_dest].type = type;
	dest.entries[dest.num_dest].pointer = pointer;
	dest.entries[dest.num_dest].x = x;
	dest.entries[dest.num_dest].y = y;
	dest.entries[dest.num_dest].direction = direction;
	dest.entries[dest.num_dest].eventFlag = eventFlag;
	dest.entries[dest.num_dest].checkFlag = checkFlag;
	dest.entries[dest.num_dest].style = style;

	dest.num_dest++;

	delete temp;

}

void CRom::RemoveDestination(int destination)
{
	DestinationEntry *temp = new DestinationEntry[dest.num_dest];

	for(int i = 0; i < dest.num_dest; i++)
		temp[i] = dest.entries[i];

	dest.num_dest--;

	for(i = destination; i < dest.num_dest; i++)
		temp[i] = temp[i+1];

	dest.entries = new DestinationEntry[dest.num_dest];

	for(i = 0; i < dest.num_dest; i++)
		dest.entries[i] = temp[i];

	delete temp;

	// okay, we're not done yet.
	// here's the tricky part:
	// all the destination entries after the one we just deleted will get bumped back
	// one position. This also means that a very large number of door entries will
	// now have incorrect destination indexes attached to them.
	// basically, we have to find all the door entries that use those destinations above
	// the one we just deleted, and decrement their destination index values, so they'll
	// be in line with the new ones.
	// as for any doors that used the destination we just deleted?
	// heheh, well, they get set to zero.

	for(i = 0; i < 1280; i++) {
		for(int j = 0; j < door_entry[i].num_doors; j++) {
			if(door_entry[i][j].destindex > destination)
				door_entry[i][j].destindex--;

			if(door_entry[i][j].destindex == destination) {
				door_entry[i][j].destindex = 0;
				door_entry[i][j].null_dest = true;
			}
		}
	}
}


void CRom::ChangeDestination(int destination, int pointer, int x, int y, int eventFlag, bool checkFlag, int style)
{
	dest.entries[destination].pointer = pointer;
	dest.entries[destination].x = x;
	dest.entries[destination].y = y;
	dest.entries[destination].eventFlag = eventFlag;
	dest.entries[destination].checkFlag = checkFlag;
	dest.entries[destination].style = style;
}




/****OBSOLETE CODE****
int CRom::GetDETLength()
{
	int size = 0;
	for(int i = 0; i < 1280; i++) {
		size += 2;
		for(int j = 0; j < object_areas[i].num_doors; j++)
			size += 5;
	}

	return size;
}*/

int GetDestIndex(DestinationEntry *destinations, int num_dest, int destination)
{
	// retrieves the index of a given destination pointer.
	// destination data must be loaded prior to calling.
	// if a a destination matching the pointer is not found,
	// this returns -1, signifying that the object's destination
	// is indeterminate.
	int retVal = 0;

	for(int i = 0; i < num_dest; i++) {
		if(destinations[i].address == destination) {		// match found!
			retVal = i;
			break;
		}
	}

	return retVal;
}







bool CheckFile(CString szFile)
{
	// This function checks a file against various
	// criteria to make sure it is valid before loading.
	// There's the basic stuff, like making sure the file
	// can be opened and written to, or check to make sure
	// it's not a folder or some stupid thing like that.

	FILE* stream;
	char str[12];
	int idiot;

	stream = fopen(szFile, "rb+");

	if(stream) {
		fseek(stream, 0x0101C0, SEEK_SET);
		for(int i = 0; i < 11; i++)
			str[i] = fgetc(stream);
		str[11] = 0;
		if(strcmp(str, "EARTH BOUND"))
			idiot = AfxMessageBox("This doesn't look like an EarthBound ROM. Continuing may cause damage to your ROM. Do you want to continue anyway?", MB_YESNO | MB_ICONQUESTION);
		else
			return true;
		if(idiot == IDYES)
			return true;

		fclose(stream);
	}
	else {
		AfxMessageBox("An error occurred while attempting to open this file. Check to make sure the file is not read-only, and is not currently in use by another application.");
	}

	return false;
}


int CheckExpand(FILE* rom)
{
	int retVal = 0;
	int error;

	fseek(rom, 0, SEEK_END);
	long len = ftell(rom);

	if(len == 0x400200) {
		retVal = 1;
	}
	else if(len == 0x300200) {
		error = AfxMessageBox("WARNING\n\nThis ROM must be expanded before door data can be saved.\nDo you want to expand the ROM now?\n\n(Choosing \"No\" will skip saving door data)", MB_YESNOCANCEL);
		if(error == IDYES) {
			ExpandROM(rom);
			retVal = 1;
		}
		if(error == IDCANCEL) {
			retVal = -1;
		}
	}
	else {
		AfxMessageBox("ROM is unrecognized size.");
		retVal = -1;
	}

	return retVal;
}


int ExpandROM(FILE* rom)
{
	ZeroRange(rom, 0x300200, 0x400200, 1);
	return 1;
}



unsigned int HexToSNES(unsigned int address)
{
   // Converts good-ol regular hex addresses into the SNES-style addresses.
   // In this case, it's assumed the ROM is a hi-rom with a header.

   unsigned int retVal;

   retVal = (address + 0xc00000) - 0x200;

   return retVal;
}

unsigned int SNESToHex(unsigned int address)
{
   // Converts an SNES-style address (hi-rom, with header) to a nice
   // handy hex address.

   unsigned int retVal;

   retVal = (address - 0xc00000) + 0x200;

   return retVal;
}

unsigned int GetPointer(FILE* rom, unsigned int address)
{
	// Retrieves a 32-bit SNES pointer stored at address
	int curPos = ftell(rom);

	unsigned int retVal;

	fseek(rom, address, SEEK_SET);

	retVal =  fgetc(rom);
	retVal += fgetc(rom) << 8;
	retVal += fgetc(rom) << 16;

	fseek(rom, curPos, SEEK_SET);

	return retVal;
}

int PutPointer(FILE* rom, unsigned int address, unsigned int newPointer)
{
	// Write a 32-bit SNES pointer to address
	// NOTE: newPointer should be in SNES format, not standard hexadecimal
	// (Unless, of course, you _want_ to write a hex pointer into the ROM)
	// ...
	// (Moron.)
	int curPos = ftell(rom);

	fseek(rom, address, SEEK_SET);

	fputc(newPointer & 255, rom);
	fputc((newPointer >> 8) & 255, rom);
	fputc((newPointer >> 16) & 255, rom);

	fseek(rom, curPos, SEEK_SET);

	return 1;
}


CString FormatSNESAdr(int adr)
{
	// Formats a raw integer into a hex-format string with the
	// little '$' sign in front of it
	CString str = "                        ";
	char temp[24];

	itoa(adr, temp, 16);

	for(int i = 0; i < 24; i++)
		str.SetAt(i, temp[i]);

	str.Insert(0,'$');
	str.MakeUpper();

	return str;
}

int DeformatSNESAdr(CString adr)
{
	// The opposite of above; takes a formatted address and converts
	// it into the actual integer address
	int result;

	adr.Remove('$');

	result = strtoul((LPCSTR)adr, NULL, 16);

	return result;
}


int GetNextAddress(int size)
{
	// Returns the next available address in the region addressable
	// by the various two-byte pointer tables. (0xF0200-0x100200)
	// Returns -1 if there is no more room.

	int retVal = -1;

	if(cur_address + size > 0xF8000)
	{
		AfxMessageBox("Welcome to B0rksville;\nPopulation: you");
	}
	else
	{
		retVal = cur_address;
		cur_address += size;
	}

	return retVal;
}


unsigned long FindFreeRange(FILE* rom, unsigned long startAt, int length)
{
	// simple recursive function
	// finds the highest occurrence of a data-free block of size [length]
	// checks for normal expanded area data (i.e., 255 [00]s, one [02], and so on)
	// if it finds an interrupt in the pattern, it will skip that area and
	// search just beyond it, thus leaving any user-modified data intact (hopefully)

	int ch;

	for(unsigned long i = startAt; i > startAt-length; i--) {
		fseek(rom, i, SEEK_SET);

		ch = fgetc(rom);

		if(((i+1)%0x100) == 0) {
			if(ch != 2) {
				return FindFreeRange(rom, i-1, length);
			}
		}
		else {
			if(ch != 0) {
				return FindFreeRange(rom, i-1, length);
			}
		}
	}

	return i;
}

int ZeroRange(FILE* rom, unsigned int start, unsigned int end, int type)
{
	// Zeroes out a specified range with the specified range type
	// 0: Normal; all zeroes
	// 1: Text: 02's every 255 bytes

	int curPos = ftell(rom);

	fseek(rom, start, SEEK_SET);

	for(unsigned int i = start; i < end; i++) {
		if(type == 1 && ((i+1)%0x100) == 0)
			fputc(2, rom);
		else
			fputc(0, rom);
	}

	fseek(rom, curPos, SEEK_SET);

	return 1;
}





/*void LoadDoorData(FILE* rom)
{
	LoadDoorDestinations(rom, &dest);
	LoadDoorEntries(rom, object_areas, FIRSTLOAD);

	for(int i = 0; i < 1280; i++) {
		for(int j = 0; j < object_areas[i].num_doors; j++) {
			object_areas[i].doors[j].destindex = GetDestIndex(dest.entries, dest.num_dest, object_areas[i].doors[j].destination);
		}
	}
}*/




/*void SaveDoorData(FILE* rom, DestinationList dest)
{
	if(FIRSTLOAD) {
		// delete old stuff
		ZeroRange(rom, 0x0F0200, 0x0F284E, 0);	// destination table
	}
	ZeroRange(rom, det_first, det_last, (HexToSNES(det_first) >= 0xF00000));

	WriteDoorDestinations(rom, dest);

	// okay, the old door data's been wiped out, and we've got the door destinations done
	// time to set the addresses for the door destinations now
	for(int i = 0; i < 1280; i++) {
		for(int j = 0; j < object_areas[i].num_doors; j++) {
			object_areas[i].doors[j].destination = dest.entries[object_areas[i].doors[j].destindex].address;
		}
	}

	int det_length = GetDETLength(object_areas);
	int write_adr = FindFreeRange(rom, 0x400000, det_length);

	WriteEntryTable(rom, object_areas, write_adr);

	CreateAppData();
	WriteSystemArea(rom, sys);
}*/





// graphics functions







void DrawTile(FILE* rom, long address, int xx, int yy, int pal, unsigned char sprite[80][64])
{
   unsigned char ch;
   int           i;
   int           y;
   char          sprLine[8][8];

   fseek(rom, address, SEEK_SET);

   for (y = 0; y < 8; y++)
   {
      ch = fgetc(rom);

      for (i = 0; i < 8; i++)
         sprLine[y][7 - i] = (ch & (1 << i)) >> i;

      ch = fgetc(rom);

      for (i = 0; i < 8; i++)
         sprLine[y][7 - i] += ((ch & (1 << i)) >> i) << 1;
   }

   for (y = 0; y < 8; y++)
   {
      ch = fgetc(rom);

      for (i = 0; i < 8; i++)
         sprLine[y][7 - i] += ((ch & (1 << i)) >> i) << 2;

      ch = fgetc(rom);

      for (i = 0; i < 8; i++)
         sprLine[y][7 - i] += ((ch & (1 << i)) >> i) << 3;
   }

   for (y = 0; y < 8; y++)
   {
      for (i = 0; i < 8; i++)
      {
         if (sprLine[y][i] == 0)
            sprite[yy + y][xx + i] = 0;
         else
            sprite[yy + y][xx + i] = /*(16 * pal) +*/ sprLine[y][i];
      }
   }
}

void GetSprite(FILE* rom, SpriteGfx *sprite)//int spriteNum, unsigned char spr[80][64], int* width, int* height)
{
   // width and height and palette will have to be predetermined and then
   // selected in this function.  Calling functions do not require this info.
   // Same goes for address of each sprite.  This function just takes a
   // sprite "number", which is determined by where it appears in the rom.

   unsigned int   address   = sprite->address;
   int            offset    = 0;
   int            pal       = sprite->palette;
   int			  width		= sprite->width;
   int			  height	= sprite->height;

   //address = sprite[spriteNum].address;
   
   //*width = sprite[spriteNum].width;
   //*height = sprite[spriteNum].height;
   //pal = sprite[spriteNum].palette;

   for (int y = 0; y < 80; y++)
   {
      for (int x = 0; x < 64; x++)
      {
         sprite->spr[y][x] = 0;
      }
   }

{
   for (int a = 0; a < height; a++)
   {
      for (int i = 0; i < width; i++)
      {
         DrawTile(rom, address + offset, i * 8, a * 8, pal, sprite->spr);
         offset += 32;
      }
   }
}
}


void ChangeTile(FILE* rom, unsigned int loc, unsigned char newTile[8][8])
{
   unsigned char num;
   unsigned char num2;
   unsigned char num3;
   unsigned char num4;
   int i,j;

   fseek(rom, loc, SEEK_SET);
   fflush(rom);

   for (j = 0; j < 8; j++)
   {
      num = 0;
      num2 = 0;
      num3 = 0;
      num4 = 0;

      for (i = 0; i < 8; i++)
      {
         num += ((newTile[j][i] & 0x01) << (7 - i));
         num2 += (((newTile[j][i] & 0x02) >> 1) << (7 - i));
         num3 += (((newTile[j][i] & 0x04) >> 2) << (7 - i));
         num4 += (((newTile[j][i] & 0x08) >> 3) << (7 - i));
      }

      fseek(rom, loc, SEEK_SET);
      fputc(num, rom);
      fputc(num2, rom);

      fseek(rom, loc + 0x10, SEEK_SET);
      fputc(num3, rom);
      fputc(num4, rom);

      loc += 2;
   }
}

void ChangeSprite(FILE* rom, SpriteGfx sprite, unsigned char newTile[80][64])
{
   // Assumes rom is in write binary format.
   // spriteNum is the raw sprite number to be edited.  This is the #
   // associated with the original sprite viewer program, and goes from 0
   // to 1143.
   // newTile is the new sprite info to be stored over the sprite you
   // indicate.  Width and height are automatically trimmed to the correct
   // size.

   // NOTE: I gotta do something about that static array here, there's one
   // in the getsprite routine too, and that's A LOT of wasted memory,
   // plus creating static variables like this is bad design.  I'll figure
   // something out later.

   unsigned char temp[8][8];
   int           offset = 0;
   int           sx;
   int           sy;
   unsigned int  loc	= sprite.address;
   unsigned char width	= sprite.width;
   unsigned char height	= sprite.height;
   int y,x,iy,ix;

   for (y = 0; y < height; y++)
   {
      for (x = 0; x < width; x++)
      {
         sy = y * 8;
         sx = x * 8;
         
         for (iy = sy; iy < sy + 8; iy++)
         {
            for (ix = sx; ix < sx + 8; ix++)
            {
               temp[iy - sy][ix - sx] = newTile[iy][ix];
            }
         }

         ChangeTile(rom, loc + offset, temp);
         offset += 32;
      }
   }
}

void GetSpriteRGB(FILE* rom, int index, Palette &pal) // int r[16], int g[16], int b[16])
{
   int           bgrBlock;
   unsigned char ch;
   unsigned char ch2;

   // the palette at 0x30200 is the first one, and is the one used
   // by human enemies to make their faces blue.
   // i think #8 could be the moonside neony palette

   fseek(rom, 0x30200 + index * 32, SEEK_SET);

   ch = fgetc(rom);		// the first two bytes are insignificant, I guess
   ch = fgetc(rom);

   pal[0].r = 0;
   pal[0].g = 0;
   pal[0].b = 0;


   int i;
   for (i = 1; i < 16; i++)
   {
      ch = fgetc(rom);
      ch2 = fgetc(rom);

      bgrBlock = ch2 << 8;		// shift the second byte eight bits to the left
      bgrBlock += ch;			// add the first byte
      bgrBlock &= 0x7FFF;		// Ignore high-order bit

      pal[i].b = bgrBlock >> 10;
      pal[i].g = (bgrBlock & 0x03e0) >> 5;
      pal[i].r = bgrBlock & 0x001f;


      pal[i].r *= 8;				// multiply by eight to put the value into a 0-255 range
      pal[i].g *= 8;
      pal[i].b *= 8;

   }
}

void SetSpriteRGB(FILE* rom, int pal, int r[16], int g[16], int b[16])
{
   int           bgrBlock;
   int cr,cg,cb;

   // the palette at 0x30200 is the first one, and is the one used
   // by human enemies to make their faces blue.
   // i think #8 could be the moonside neony palette

   fseek(rom, (0x30200 + pal * 32)+2, SEEK_SET);

   for (int i = 1; i < 16; i++)
   {
	  cr = r[i];
	  cg = g[i];
	  cb = b[i];

	  cb <<= 10;
	  cg <<= 5;

	  bgrBlock = cb;
	  bgrBlock += cg;
	  bgrBlock += cr;

	  fputc(bgrBlock & 255, rom);
	  fputc((bgrBlock >> 8) & 255,rom);
   }
   fflush(rom);
}






void SpriteGfx::Draw(CDC *pDC, int x, int y, Palette pal, bool flip)
{
	int xx,yy;
	for(int i = 0; i < height*8; i++)
	{
		for(int j = 0; j < width*8; j++)
		{
			int c = spr[i][j];

			xx = x+j;
			yy = y+i;

			if(flip)
				xx = x + width*8 - (j+1);

			if(c != 0) {
				pDC->SetPixel(xx, yy, RGB(pal[c].r, pal[c].g, pal[c].b));
			}
		}
	}
}

void Palette::Draw(CDC *dc, int x, int y, bool to24bit)
{
	// Debugging: draws all colors in the palette to the specified dc
	CRect rect(x,y,x+5,y+10);

	for(int i = 0; i < 16; i++)
	{
		if(to24bit)
			dc->FillSolidRect(rect, RGB(clrs[i].r*8, clrs[i].g*8, clrs[i].b*8));
		else
			dc->FillSolidRect(rect, RGB(clrs[i].r, clrs[i].g, clrs[i].b));
		rect.OffsetRect(5,0);
	}
}




void LoadTPT(FILE* rom, TPTEntry *tpt, void _cdecl update(int))
{
	int offset = 0;

	for(int i = 0; i < 1583; i++) {
		fseek(rom, 0xF8B96+offset, SEEK_SET);

		tpt[i].type = fgetc(rom);

		tpt[i].sprite = fgetc(rom);
		tpt[i].sprite+= fgetc(rom) << 8;

		tpt[i].dir = fgetc(rom) >> 1;

		tpt[i].movement1 = fgetc(rom);
		tpt[i].movement2 = fgetc(rom);

		tpt[i].flag = fgetc(rom);
		tpt[i].flag+= fgetc(rom) << 8;

		tpt[i].flaguse = fgetc(rom);

		tpt[i].pointer = fgetc(rom);
		tpt[i].pointer+= fgetc(rom) << 8;
		tpt[i].pointer+= fgetc(rom) << 16;

		fgetc(rom);		// null pointer byte

		if(tpt[i].type == 2) {	// item
			tpt[i].item = fgetc(rom);

			tpt[i].itempointer = 0;
		}
		else {
			tpt[i].item = 0;

			tpt[i].itempointer = fgetc(rom);
			tpt[i].itempointer+= fgetc(rom) << 8;
			tpt[i].itempointer+= fgetc(rom) << 16;
		}

		if((i%13) == 0)
			update(-1);

		offset+= 0x11;
	}
}

void WriteTPT(FILE* rom, TPTEntry *tpt)
{
	int offset = 0;

	for(int i = 0; i < 1583; i++) {
		fseek(rom, 0xF8B96+offset, SEEK_SET);

		fputc(tpt[i].type, rom);

		fputc(tpt[i].sprite & 255, rom);
		fputc((tpt[i].sprite >> 8) & 255, rom);

		fputc(tpt[i].dir<<1, rom);

		fputc(tpt[i].movement1, rom);
		fputc(tpt[i].movement2, rom);

		fputc(tpt[i].flag & 255, rom);
		fputc((tpt[i].flag >> 8) & 255, rom);
		fputc(tpt[i].flaguse, rom);

		fputc(tpt[i].pointer & 255, rom);
		fputc((tpt[i].pointer >> 8) & 255, rom);
		fputc((tpt[i].pointer >> 16) & 255, rom);
		fputc(0, rom);

		if(tpt[i].type == 2) {
			fputc(tpt[i].item, rom);
		}
		else {
			fputc(tpt[i].itempointer & 255, rom);
			fputc((tpt[i].itempointer >> 8) & 255, rom);
			fputc((tpt[i].itempointer >> 16) & 255, rom);
			fputc(0, rom);
		}

		offset+= 0x11;
	}
}



void LoadMusicTable(FILE* rom, MapMusicEntry entries[164])
{
	int ptr_address;
	fseek(rom, 0x06B39, SEEK_SET);
	ptr_address = fgetc(rom);
	ptr_address+= fgetc(rom) << 8;
	ptr_address+= fgetc(rom) << 16;
	ptr_address-= 0xBFFDFE;

	for(int i = 0; i < 164; i++)
	{
		fseek(rom, ptr_address+i*2, SEEK_SET);

		int adr;
		adr = fgetc(rom);
		adr+= fgetc(rom) << 8;

		fseek(rom, 0xF0200+adr, SEEK_SET);

		int flag;
		int music;

		entries[i].num_events = 0;

		do
		{
			flag = fgetc(rom);
			flag+= fgetc(rom) << 8;
			music = fgetc(rom);
			fgetc(rom);

			entries[i].events[entries[i].num_events].flag = flag;
			entries[i].events[entries[i].num_events].music = music;
			entries[i].num_events++;
		}while(flag != 0);
	}
}


void WriteMusicTable(FILE *rom, MapMusicEntry entries[164], int &next_adr)
{
	// start writing at 
	//int address = 0x0F5C39 - 0xF0200;
	//int offset = 0;

	//int adr;


	// rewrite the ASM instruction that loads the pointer table so
	// that it points to the new pointer table at 0xF8A00
	fseek(rom, 0x06B39, SEEK_SET);
	fputc(0xFE, rom);
	fputc(0x87, rom);
	fputc(0xCF, rom);

	for(int i = 0; i < 164; i++)
	{
		int size = entries[i].num_events*4;
		int address = next_adr;
		next_adr+= size;

		if(address > 0xF8000)
		{
			AfxMessageBox("Music Table write error");
			return;
		}

		fseek(rom, 0xF8A00+i*2, SEEK_SET);

		// write pointer
		fputc((address-0xF0200) & 255, rom);
		fputc((address-0xF0200) >> 8, rom);

		//if(i == 163)
		//{
		//	CString s;
		//	s.Format("0x%X - 0x%X", adr+0xF0200, adr + 0xF0200 + entries[i].num_events*4);
		//	AfxMessageBox(s);
		//}


		/*if(adr+0xF0200 > (0x0F63DD - entries[i].num_events*4))
		{
			AfxMessageBox("w00t!");
			return;
		}*/

		// write entry
		fseek(rom, address, SEEK_SET);

		for(int j = 0; j < entries[i].num_events; j++)
		{
			fputc(entries[i].events[j].flag & 255, rom);
			fputc((entries[i].events[j].flag >> 8)& 255, rom);
			fputc(entries[i].events[j].music, rom);
			fputc(0, rom);
			//offset+=4;
		}
		//CString s;
		//s.Format("Final write address: 0x%X", ftell(rom));
		//AfxMessageBox(s);
	}
}



void LoadSpriteEntries(FILE* rom, ObjectArea *areas, bool firstload, void _cdecl update(int))
{
	int address;
	int num_sprites;

	int ptr_address;
	fseek(rom, 0x02461, SEEK_SET);
	ptr_address = fgetc(rom);
	ptr_address+= fgetc(rom) << 8;
	ptr_address+= fgetc(rom) << 16;
	ptr_address-= 0xBFFE00;


	for(int i = 0; i < 1280; i++) {
		fseek(rom, ptr_address+(i*2), SEEK_SET);

		address = fgetc(rom);
		address += fgetc(rom) << 8;

		if(address == 0) continue;

		address += 0xF0200;

		fseek(rom, address, SEEK_SET);

		areas[i].spr_address = address;

		num_sprites = fgetc(rom);
		fgetc(rom);

		areas[i].num_sprites = num_sprites;
		if(num_sprites < 256 && num_sprites >= 0) areas[i].sprites = new SpriteEntry[num_sprites];

		for(int j = 0; j < num_sprites; j++) {
			areas[i].sprites[j].address = address + 2 + (j*5);

			areas[i].sprites[j].tptNum = fgetc(rom);
			areas[i].sprites[j].tptNum+= fgetc(rom) << 8;

			areas[i].sprites[j].tptNum--;

			areas[i].sprites[j].y = fgetc(rom);
			areas[i].sprites[j].x = fgetc(rom);
		}

		if(i%10 == 0)
			update(-1);
	}
}

void WriteSpriteTable(FILE* rom, ObjectArea *areas, int &next_adr)
{
	// start writing sprites at 0x0F6DE7, and when we run out of room,
	// begin putting them at the end of the door entries
	//int next_adr = ddt_last;

	//int address = 0x0F6DE7;
	//int offset = 0;

	// write pointer to pointer table
	fseek(rom, 0x02461, SEEK_SET);
	fputc(0x00, rom);
	fputc(0x7E, rom);
	fputc(0xCF, rom);

	for(int i = 0; i < 1280; i++)
	{
		// location that current sprite block will extend to
		//int block_end = address+offset+2+(areas[i].num_sprites*4);

		//if(block_end > 0x0F8B91) {
		//	address = next_adr;
		//	offset = 0;
		//}
		int size = 2 + (areas[i].num_sprites*4);	// size required for sprite block
		int address = next_adr;						// Address to which we can write
		next_adr+= size;

		if(address > 0xF8000)
		{
			AfxMessageBox("Sprite Table write error");
			return;
		}

		// write pointer
		fseek(rom, 0x0F8000+i*2, SEEK_SET);

		if(areas[i].num_sprites == 0) {
			fputc(0, rom);
			fputc(0, rom);
		}
		else {
			fputc((address-0xF0200) & 255, rom);
			fputc(((address-0xF0200) >> 8) & 255, rom);

			fseek(rom, address, SEEK_SET);

			fputc(areas[i].num_sprites, rom);
			fputc(0, rom);
			//offset+=2;

			for(int j = 0; j < areas[i].num_sprites; j++)
			{
				areas[i].sprites[j].address = address+j*4;

				fputc((areas[i].sprites[j].tptNum+1) & 255, rom);
				fputc(((areas[i].sprites[j].tptNum+1) >> 8) & 255, rom);

				fputc(areas[i].sprites[j].y, rom);
				fputc(areas[i].sprites[j].x, rom);

				//offset+=4;
			}
		}
	}
}












/////////////////////////////////////////////////////////////////////
// Text functions

String* CRom::ReadString(FILE* rom, unsigned long address, CCNode ccTable)
{
	// Reads a string from ROM at address.
	// Returns a struct containing a pointer to the string and
	// the string's length. This function does not do any formatting
	// or decompression. If the string is greater than 8192 characters
	// long, it will be truncated at that point.

	char buffer[8192];

	int pos = 0;
	int len = 0;

	CCNode* activeNode = &ccTable;

	bool opcode = false;
	int arglevel = 0;

	unsigned char ch;

	fseek(rom, address, SEEK_SET);

	for(int i = 0; i < 8192; i++)
	{
		ch = fgetc(rom);
		len++;

		if(arglevel > 0)
		{
			buffer[pos++] = ch;

			arglevel--;
			continue;
		}

		if(!opcode)
		{
			if(isCC(ch))
			{
				opcode = true;
			}
		}

		if(opcode)
		{
			int n = SearchNode(activeNode, ch);
			if(n == -1)
			{
				opcode = false;
				activeNode = &ccTable;
				buffer[pos++] = ch;
				continue;				
			}
			else
			{
				activeNode = &activeNode->nodes[n];
			}
			
			if(activeNode->type == TYPE_ARGLIST)
			{
				arglevel = activeNode->num_arg-1;
				buffer[pos++] = ch;
			}
			else if(activeNode->type == TYPE_CODE)
			{
				buffer[pos++] = ch;
			}

			if(activeNode->level == 0 && activeNode->value == 0x02)
				break;

			if(activeNode->isTerminator)
			{
				opcode = false;
				activeNode = &ccTable;
			}
		}
		else
		{
			buffer[pos++] = ch;
		}
	}

	buffer[pos++] = 0;

	String *s;
	s = new String;
	s->address = address;
	s->str = new char[pos];
	s->len = len;
	
	for(i = 0; i < pos; i++)
		s->str[i] = buffer[i];

	return s;
}



/*int WriteString(FILE* rom, String s, int address)
{
	fseek(rom, address, SEEK_SET);

	for(int i = 0; i < s.len; i++)
	{
		fputc(s.str[i], rom);
	}
	fflush(rom);

	return 0;
}







/*char* ParseString(char str[], int len, CCNode ccTable, char* comprTable[])
{
	char buffer[8192];

	int pos = 0;	// position in returned string
	int ipos = 0;

	CCNode* activeNode = &ccTable;

	bool opcode = false;
	int arglevel = 0;

	int curcode;		// initial byte of current control code

	unsigned char ch;

	for(int i = 0; i < len; i++)
	{
		ch = str[ipos++];
		//len++;

		// if there are arguments left to count, decrement arg level and continue
		if(arglevel > 0)
		{
			AddCode(buffer, ch, pos);
			pos+=4;

			arglevel--;
			continue;
		}

		// if opcode was not previously set
		if(!opcode)
		{
			if(isCC(ch))
			{
				curcode = ch;
				opcode = true;
			}
		}


		// do more stuff

		// if opcode is true, ...

		// note that none of this will happen when there are arguments in the queue,
		// so the active node will still be the original argument node when the arguments
		// finally close, so everything will proceed in an orderly fashion.
		// (an argument node will just run through a few extra bytes and add them to the string,
		// whereas other nodes involve adding only one character to the string)

		if(opcode)
		{

			// search for a subnode matching the current byte
			// (as of initial iteration, activeNode is the root node)
			int n = SearchNode(activeNode, ch);
			if(n == -1)
			{
				// if no match was found, this is an invalid code
				opcode = false;
				activeNode = &ccTable;
			}
			else
			{
				// set the active node to a subnode of itself
				activeNode = &activeNode->nodes[n];
			}
			

			// okay, now we have a byte from the ROM, and have associated it
			// with a node. depending on the node type, do stuff

			if(activeNode->type == TYPE_ARGLIST)
			{
				// set arg level to num_arg - 1 because we've already read
				// one of the argument bytes
				arglevel = activeNode->num_arg-1;

				if(curcode == 0x15 || curcode == 0x16 || curcode == 0x17)
				{
					int c = ((curcode-0x15)*256)+ch;

					int p = 0;

					while(comprTable[c][p] != 0)
					{
						buffer[pos++] = comprTable[c][p++];
					}
				}
				else
				{
					AddCode(buffer, ch, pos);
					pos+=4;
				}
			}
			if(activeNode->type == TYPE_CODE)
			{
				// only add a bracketed code if the current code is not a compression code
				if(curcode != 0x15 && curcode != 0x16 && curcode != 0x17)
				{
					AddCode(buffer, ch, pos);
					pos+=4;
				}
			}



			// if the current node is a first-byte code with value 2, terminate text block
			if(activeNode->level == 0 && activeNode->value == 0x02)
				break;


			// if the active node is a terminator node,  stop stuff
			if(activeNode->isTerminator)
			{
				opcode = false;

				// reset the active node so we start from the beginning
				activeNode = &ccTable;
			}
		}
		else
		{
			// if the current byte is not a control code,
			// add it to the string as-is
			buffer[pos++] = ch-0x30;
		}
	}

	buffer[pos++] = 0;

	char* newstr = new char[pos+1];
	for(i = 0; i < pos; i++)
		newstr[i] = buffer[i];

	newstr[pos] = 0;

	return newstr;
}*/


char* CRom::ParseHeader(char str[], int len, CCNode ccTable, char* comprTable[])
{
	char buffer[8192];

	int pos = 0;
	int ipos = 0;

	CCNode* activeNode = &ccTable;

	bool opcode = false;
	int arglevel = 0;

	int curcode;

	unsigned char ch;

	for(int i = 0; i < len; i++)
	{
		ch = str[ipos++];

		if(arglevel > 0)
		{
			arglevel--;
			continue;
		}

		if(!opcode)
		{
			if(isCC(ch))
			{
				curcode = ch;
				opcode = true;
			}
		}

		if(opcode)
		{
			int n = SearchNode(activeNode, ch);
			if(n == -1)
			{
				opcode = false;
				activeNode = &ccTable;
			}
			else
			{
				activeNode = &activeNode->nodes[n];
			}
			
			if(activeNode->type == TYPE_ARGLIST)
			{
				arglevel = activeNode->num_arg-1;

				if(curcode == 0x15 || curcode == 0x16 || curcode == 0x17)
				{
					int c = ((curcode-0x15)*256)+ch;

					int p = 0;

					while(comprTable[c][p] != 0)
					{
						buffer[pos++] = comprTable[c][p++];
					}
				}
			}

			if(activeNode->level == 0 && activeNode->value == 0x02)
				break;

			if(activeNode->isTerminator)
			{
				opcode = false;
				activeNode = &ccTable;
			}
		}
		else
		{
			buffer[pos++] = ch-0x30;
		}
	}

	buffer[pos++] = 0;

	char* newstr = new char[pos+1];
	for(i = 0; i < pos; i++)
		newstr[i] = buffer[i];

	newstr[pos] = 0;

	return newstr;
}


/*char* ParseGrouping(char str[])
{
	int pos = 0;

	int len = strlen(str);

	do
	{
		char* s = strstr(str, "][");

		if(s == NULL)
		{
			break;
		}
		else
		{
			pos = s - str;

			str[pos] = ' ';

			for(int j = pos+1; j < len; j++)
			{
				str[j] = str[j+1];
			}
		}
	}while(pos != -1);

	return str;
}


/*int GetStringLength(char str[])
{
	// This is similar to DeparseString (below);
	// but this function does not deparse the string, it just counts
	// the number of characters.
	// Naturally, the value returned represents the string's un-
	// compressed length.
	// (xlen is the length of the code-formatted string)
	// (Returns -1 if a bracket error is encountered)

	int len = 0;
	int bracketlevel = 0;
	int ch;

	int xlen = strlen(str);

	for(int i = 0; i < xlen; i++)
	{
		if(str[i] == '\0')
		{
			CString x;
			x.Format("%d", i);
			AfxMessageBox(x);
			break;
		}

		if(str[i] == '[')
		{
			bracketlevel++;
			continue;
		}

		if(str[i] == ']')
		{
			bracketlevel--;
			continue;
		}

		if(bracketlevel < 0) {
			return -1;
		}


		if(bracketlevel > 0)
		{
			int j = 0;

			do
			{
				ch = str[i++];

				if(ch == ']')
				{
					len++;
					j = 0;
					i-=2;
					break;
				}

				if(ch == ' ' || ch == ']')
				{
					if(j != 0)
					{
						len++;
						j = 0;
					}

				}
				else
					j++;

				if(j == 32) break;

				
			}while(ch != ']');
		}

		else
		{
			len++;
		}

	}

	if(bracketlevel != 0)
	{
		return -1;
	}

	return len;
}


/*char* DeparseString(char str[])
{
	// This function takes a visually formatted string (like the user
	// input into the text editing box) and converts it into a string
	// that is ready to be written directly to the ROM.

	char buffer[8192];
	char *newstr;	// this gets allocated later

	int pos = 0;

	int bracketlevel = 0;

	int ch;

	int len = strlen(str);

	for(int i = 0; i < len; i++)
	{
		if(str[i] == '[')
		{
			bracketlevel++;
			continue;
		}

		if(str[i] == ']')
		{
			bracketlevel--;
			continue;
		}

		if(bracketlevel < 0) {
			AfxMessageBox("WARNING: Bracket error detected.");
			return NULL;
		}

		// if we're still inside brackets (if bracketlevel > 0),
		// get everything up to and including the next space, or
		// up to the next closing bracket.

		if(bracketlevel > 0)
		{
			char t[32];
			int j = 0;

			do
			{
				ch = str[i++];

				if(ch == ']')
				{
					t[j] = 0;

					int a = strtoul(t, NULL, 16)&0xFF;
					buffer[pos++] = a;

					j = 0;
					i-=2;
					break;
				}

				if(ch == ' ' || ch == ']')
				{
					if(j != 0)
					{
						t[j] = 0;

						int a = strtoul(t, NULL, 16)&0xFF;
						buffer[pos++] = a;

						j = 0;
					}

				}
				else
					t[j++] = ch;

				if(j == 32) break;

				
			}while(ch != ']');
		}

		else
			buffer[pos++] = str[i]+0x30;

	}

	if(bracketlevel != 0)
	{
		AfxMessageBox("WARNING: Bracket error detected.");
		return NULL;
	}


	buffer[pos++] = 0;

	newstr = new char[pos];
	for(i = 0; i < pos; i++)
		newstr[i] = buffer[i];

	return newstr;
}*/

int CRom::SearchNode(CCNode *node, int val)
{
	// note: if there's only one subnode of the specified node,
	// and it's an argument list, return its index regardless of
	// the specified value.

	if(node->num_nodes == 1 && node->nodes[0].type == TYPE_ARGLIST)
	{
		return 0;
	}

	int retVal = -1;
	for(int i = 0; i < node->num_nodes; i++)
	{
		if(node->nodes[i].value == val)
			retVal = i;
	}

	return retVal;
}


bool CRom::isCC(char ch)
{
	bool retVal = false;

	if(ch >= 0 && ch <= 0x1F)
		retVal = true;

	return retVal;
}

void CRom::AddCode(char* str, int val, int pos)
{
	char b[5];
	if(val < 0x10)
		sprintf(b, "[0%X]", val);
	else
		sprintf(b, "[%X]", val);

	str[pos++] = b[0];
	str[pos++] = b[1];
	str[pos++] = b[2];
	str[pos++] = b[3];
}



void CRom::CreateCompressionTable(FILE* rom, char* tbl[])
{
	// loads compression strings from rom.
	// assumes tbl is allocated with 768 pointers to strings

	char buffer[256];
	int adr;
	char ch;
	int pos;

	for(int i = 0; i < 768; i++)
	{
		fseek(rom, 0x8CFED+(i*4), SEEK_SET);

		pos = 0;

		adr = fgetc(rom);
		adr+= fgetc(rom) << 8;
		adr+= fgetc(rom) << 16;

		fseek(rom, SNESToHex(adr), SEEK_SET);
		for(int j = 0; j < 256; j++)
		{
			ch = fgetc(rom);
			buffer[pos++] = ch;
			if(ch == 0) break;
		}

		tbl[i] = new char[pos+1];

		for(j = 0; j < pos+1; j++)
			tbl[i][j] = buffer[j]==0 ? 0 : buffer[j]-0x30;
	}
}


void CRom::CreateCCTable(char* filename, CCNode table[])
{
	FILE* file;
	char buffer[1024];
	unsigned char codestr[1024];
	unsigned char nodestr[1024];
	unsigned char descstr[1024];
	int bpos = 0;
	int pos = 0;
	int len = 0;

	file = fopen("codelist.txt", "r");

	if(!file) AfxMessageBox("Holy crap and a half!");

	if(file)
	{
		fseek(file, 0, SEEK_SET);

		while(!feof(file))
		{
			fgets(buffer, 1024, file);

			while(buffer[bpos] != ',' && bpos < 1024)
			{
				if(buffer[bpos] != ' ')
				{
					codestr[pos++] = buffer[bpos];
				}
				bpos++;
			}
			codestr[pos] = 0;
			len = pos;
			pos = 0;
			bpos++;

			while(buffer[bpos] != ';' && bpos < 1024)
			{
				if(buffer[bpos] != '"')
				{
					descstr[pos++] = buffer[bpos];
				}
				bpos++;
			}
			descstr[pos] = 0;
			bpos = pos = 0;

			//printf("%s\t:\t", codestr);

			// now convert the string into a series of single bytes
			for(int i = 0; i < len; i)
			{
				char t[3];
				t[0] = codestr[i++];
				t[1] = codestr[i++];
				t[2] = '\0';

				nodestr[pos++] = (unsigned char)strtoul(t, 0, 16);
			}
			nodestr[pos] = 0;
			len = pos;
			pos = 0;


			//printf("%s\n", nodestr);

			// now loop through the code string and add a series of nodes

			int prev = 0;
			bool arg_counting = false;
			int arg_count = 0;
			int args[32];

			CCNode* activeNode = table;
			int nodeLevel = 0;

			for(i = 0; i < len; i++)
			{
				//printf("%X: ", nodestr[i]);
				// if we encounter an argument value (0xAA, 0xBB, etc.),
				// start counting arguments

				// stop counting when the last code no is no longer equal to the current
				if(arg_counting)
				{
					if(prev != nodestr[i])
					{
						//printf("STOP! %d arguments!\n", arg_count);

						// now add latent node
						activeNode = AddNode(activeNode, nodeLevel, nodestr[i-1], TYPE_ARGLIST, arg_count, false, false);
						nodeLevel++;
						arg_count = 0;
						arg_counting = false;
					}
				}
				if((nodestr[i] - 0xAA)%0x11 == 0 && nodestr[i] != 0 && !arg_counting)
				{
					//printf("COUNTING ARGUMENTS!\n");
					arg_counting = true;
				}

				if(arg_counting)
				{
					// if we're still counting arguments, increment the argument count
					// and add the current value to the argument list
					args[arg_count] = nodestr[i];
					arg_count++;
				}
				else
				{
					// add default code node here
					activeNode = AddNode(activeNode, nodeLevel, nodestr[i], TYPE_CODE, NULL, false, (i+1 == len));
					nodeLevel++;
					//printf("\n");
				}
				
				
				prev = nodestr[i];
			}
			if(arg_counting)
			{
				// if we're still counting arguments, tidy up
				
				//printf("STOP! %d arguments!\n", arg_count);

				AddNode(activeNode, nodeLevel, nodestr[i-1], TYPE_ARGLIST, arg_count, false, true);

				arg_counting = false;
				arg_count = 0;
			}

			//printf("\n");
		}
	}
	fclose(file);
	delete file;
}


CCNode* CRom::AddNode(CCNode *node, int level, int value, int type, int num_arg, bool multi, bool terminate)
{
	// Adds a subnode with specified properties to the provided node

	// if a node of the specified value already exists, return a pointer to it
	for(int i = 0; i < node->num_nodes; i++)
	{
		if(node->nodes[i].value == value)
			return &node->nodes[i];
	}

	CCNode *temp = new CCNode[node->num_nodes];

	for(i = 0; i < node->num_nodes; i++)
		temp[i] = node->nodes[i];

	node->nodes = new CCNode[node->num_nodes+1];

	for(i = 0; i < node->num_nodes; i++)
		node->nodes[i] = temp[i];

	int n = node->num_nodes;

	node->nodes[n].level = level;
	node->nodes[n].value = value;
	node->nodes[n].type = type;
	node->nodes[n].num_arg = num_arg;
	node->nodes[n].arg_multiplier = multi;
	node->nodes[n].isTerminator = terminate;

	node->nodes[n].num_nodes = 0;

	delete temp;

	node->num_nodes++;

	// return a pointer to the node we just created
	return &node->nodes[n];

}





















// Graphics decompression

int tilecolor(char *tile, int x, int y) {
	int c = 0;
	if(tile[ 0 + 2*y] & (128 >> x)) c  = 1;
	if(tile[ 1 + 2*y] & (128 >> x)) c |= 2;
	if(tile[16 + 2*y] & (128 >> x)) c |= 4;
	if(tile[17 + 2*y] & (128 >> x)) c |= 8;
	return c;
}



/* The decompressor function.
 * Takes a pointer to the compressed block, a pointer to the buffer
 * which it decompresses into, . Returns the number of bytes uncompressed,
 * or -1 if decompression failed.
 */

int decomp(unsigned char *cdata, unsigned char *buffer, int maxlen) {
	unsigned char *bpos = buffer, *bpos2, tmp;
	while(*cdata != 0xFF) {
		int cmdtype = *cdata >> 5;
		int len = (*cdata & 0x1F) + 1;
		if(cmdtype == 7) {
			cmdtype = (*cdata & 0x1C) >> 2;
			len = ((*cdata & 3) << 8) + *(cdata + 1) + 1;
			cdata++;
		}
		if(bpos + len > &buffer[maxlen]) return -1;
		cdata++;
		if(cmdtype >= 4) {
			bpos2 = &buffer[(*cdata << 8) + *(cdata + 1)];
			if(bpos2 >= &buffer[maxlen]) return -1;
			cdata += 2;
		}
		switch(cmdtype) {
			case 0:
				memcpy(bpos, cdata, len);
				cdata += len;
				bpos += len;
				break;
			case 1:
				memset(bpos, *cdata++, len);
				bpos += len;
				break;
			case 2:
				if(bpos + 2 * len > &buffer[maxlen]) return -1;
				while(len--) {
					*(short *)bpos = *(short *)cdata;
					bpos += 2;
				}
				cdata += 2;
				break;
			case 3:
				tmp = *cdata++;
				while(len--) *bpos++ = tmp++;
				break;
			case 4:
				if(bpos2 + len > &buffer[maxlen]) return -1;
				memcpy(bpos, bpos2, len);
				bpos += len;
				break;
			case 5:
				if(bpos2 + len > &buffer[maxlen]) return -1;
				while(len--) {
					tmp = *bpos2++;
					/* reverse the bits */
					tmp = ((tmp >> 1) & 0x55) | ((tmp << 1) & 0xAA);
					tmp = ((tmp >> 2) & 0x33) | ((tmp << 2) & 0xCC);
					tmp = ((tmp >> 4) & 0x0F) | ((tmp << 4) & 0xF0);
					*bpos++ = tmp;
				}
				break;
			case 6:
				if(bpos2 - len + 1 < buffer) return -1;
				while(len--) *bpos++ = *bpos2--;
				break;
			case 7:
				return -1;
		}
	}
	return bpos - buffer;
}


// overloaded decompressor; takes a FILE object and an address
// rather than a pointer to a buffer
int _decomp(FILE* rom, unsigned int pos, unsigned char *buffer, int maxlen) {

	unsigned char *cdata;
	cdata = new unsigned char[maxlen];
	fseek(rom, pos, SEEK_SET);
	for(int i = 0; i < maxlen; i++)
		cdata[i] = fgetc(rom);

	int retVal = decomp(cdata, buffer, maxlen);
	delete cdata;
	return retVal;

}