/////////////////////////////////////////////////////////////////////////////////////////////////
// Classes.h: declarations for data structures

#if !defined(CLASSES_INCLUDED)
#define CLASSES_INCLUDED


#define DEST_SWITCH 0
#define DEST_ROPE 1
#define DEST_DOOR 2
#define DEST_ESCALATOR 3
#define DEST_STAIR 4
#define DEST_OBJECT 5
#define DEST_PERSON 6


#include "destinationtypes.h"



struct SystemEntry {
	bool is_valid;			// Validity

	int index;		// Ordering

	long address;	// Location of data
	int size;		// The amount of space necessary for this entry (in bytes)

	unsigned char *data;		// Should be allocated with length size

	~SystemEntry()
	{
		delete data;
	}
};


struct SystemArea {
	SystemEntry entries[32];
};


struct DoorEntry {
	unsigned int address;

	short x;
	short y;

	unsigned short destination;
	unsigned short destindex;

	unsigned char direction;

	unsigned char ropetype;

	bool is_esc_end;

	bool null_dest;

	unsigned char flags;

};

struct SpriteEntry {
	unsigned int address;

	short x;
	short y;

	unsigned short tptNum;
};


struct SelectedSector {
	unsigned short x; // tile coord / 8
	unsigned short y; // tile coord / 4
};



// this struct is used to track information about which sprites are currently on-screen.
// this is used so we can tell which sprite the user has clicked on, etc.
struct DISPSPRINFO {
	unsigned short x, y;
	unsigned char w, h;
	unsigned short area, spr;
};

struct ObjectArea {
	unsigned int door_address;
	unsigned int spr_address;

	//int num_doors;
	//DoorEntry *doors;

	int num_sprites;
	SpriteEntry *sprites;

	~ObjectArea()
	{
		delete sprites;
		//delete doors;
	};
};

// Destination entry struct is okay; no data allocation, so it's good to go
struct DestinationEntry {
	int address;	// used for index locating

	int type;

	int x;
	int y;

	int direction;

	int pointer;

	int eventFlag;
	bool checkFlag;		// last bit of event data

	int style;
};

// This needs work. Maintaining a pointer to an array of entries
// is not only a pain in the ass to manage, it's slow. New dest
// list class will use a CPtrArray. Yay MFC!
struct DestinationList {
	DestinationEntry *entries;
	int num_dest;

	~DestinationList()
	{
		delete entries;
	};
};

struct MusicEvent {
	int flag;
	int music;
};

struct MapMusicEntry {
	MusicEvent events[256];
	int num_events;
};



struct TPTEntry {
	int type;

	int sprite;

	int dir;
	int movement1;
	int movement2;

	int flag;
	int flaguse;

	int pointer;
	int itempointer;

	int item;
};

struct Palette {
	struct COLOR
	{
		unsigned char r, g, b;
	} clrs[16];

	COLOR& operator[](int n) { return clrs[n]; };

	//int r[16];
	//int g[16];
	//int b[16];

	void Draw(CDC* pDC, int x, int y, bool to24bit = FALSE);
};

struct SpriteGfx {
	unsigned int address;
	int width;
	int height;
	int palette;

	unsigned char spr[80][64];

	void Draw(CDC *pDC, int x, int y, Palette pal, bool flip = false);
};


// some data
static int cur_address = 0xF0200;


// functions


// graphics
void DrawMapTile(HDC dc, int cx, int cy, int set, int tile, int palette);

// data serialization
void GetROMStuff(FILE* rom, unsigned char[320][256],unsigned char[320][256],unsigned char[80][32],unsigned char[80][32],unsigned char[80][32]);
void SetROMStuff(FILE* rom, unsigned char[320][256],unsigned char[320][256],unsigned char[80][32],unsigned char[80][32],unsigned char[80][32]);

void LoadDoorEntries(FILE* rom, ObjectArea areas[1280], int&, int&, bool firstload, void _cdecl update(int));
void LoadSpriteEntries(FILE* rom, ObjectArea areas[1280], bool firstload, void _cdecl update(int));
void LoadDoorDestinations(FILE* rom, DestinationList *dest, SystemEntry *appdata, void _cdecl update(int));

void WriteEntryTable(FILE* rom, ObjectArea *areas, unsigned int address, int&, int&, void _cdecl update(int));
void WriteSpriteTable(FILE* rom, ObjectArea *areas, int &next_adr);
void WriteDoorDestinations(FILE* rom, DestinationList dest, int&, void _cdecl update(int));

void LoadSystemArea(FILE* rom, SystemArea* sys);
void WriteSystemArea(FILE* rom, SystemArea sys);


void LoadMusicTable(FILE* rom, MapMusicEntry entries[164]);
void WriteMusicTable(FILE *rom, MapMusicEntry entries[164], int &next_adr);


// load-only accessory data

void DrawTile(FILE* rom, long address, int xx, int yy, int pal, unsigned char sprite[80][64]);
//void GetSprite(FILE* rom, int spriteNum, unsigned char sprite[80][64], int* width, int* height);
void GetSprite(FILE* rom, SpriteGfx *sprite);

void GetSpriteRGB(FILE*, int, Palette&);
void GetSpriteRGB(FILE*, int, int[], int[], int[],int[],int[],int[]);
void SetSpriteRGB(FILE*, int, int[], int[], int[]);

void ChangeTile(FILE*, unsigned int, unsigned char[8][8]);
void ChangeSprite(FILE*, int, unsigned char[80][64]);

void LoadSprites(FILE* rom, SpriteGfx *sprites, Palette *pal);

void LoadTPT(FILE* rom, TPTEntry *tpt, void _cdecl update(int));
void WriteTPT(FILE* rom, TPTEntry *tpt);

// data modification
//void AddDoor(ObjectArea *areas, int area, int x, int y, int flags, int dest, int destindex, int dir, int rope, int esc_end);
//void RemDoor(ObjectArea *areas, int area, int door);
void ChangeDoor(ObjectArea *areas, int area, int door, int x, int y, int flags, int dest);

//void AddSprite(ObjectArea *areas, int area, int x, int y, int tptNum);
void RemSprite(ObjectArea *areas, int area, int sprite);
void ChangeSprite(ObjectArea *areas, int area, int sprite, int x, int y, int tptNum);

//void AddDestination(DestinationList *dest, int type, int pointer, int x, int y, int direction, int eventFlag, bool checkFlag, int style);
//void RemoveDestination(DestinationList *dest, int destination, ObjectArea *areas);
//void ChangeDestination(DestinationList *dest, int destination, int pointer, int x, int y, int eventFlag, bool checkFlag, int style);

//int GetDETLength(ObjectArea *areas);

void SaveDoorData();


// utilities
bool CheckFile(CString szFile);
int CheckExpand(FILE* rom);
int ExpandROM(FILE* rom);

unsigned int HexToSNES(unsigned int);
unsigned int SNESToHex(unsigned int);

CString FormatSNESAdr(int adr);
int DeformatSNESAdr(CString adr);

// data management

unsigned int GetPointer(FILE* rom, unsigned int);
int PutPointer(FILE* rom, unsigned int, unsigned int);

int GetNextAddress(int size);

unsigned long FindFreeRange(FILE* rom, unsigned long startAt, int length);
int ZeroRange(FILE* rom, unsigned int start, unsigned int end, int type);

int GetDestIndex(DestinationEntry *destinations, int num_dest, int destination);



int tilecolor(char *tile, int x, int y);
int decomp(unsigned char *cdata, unsigned char *buffer, int maxlen);
int _decomp(FILE* rom, unsigned int pos, unsigned char *buffer, int maxlen);




#endif