// MapEdit.cpp : Defines the class behaviors for the application.
//
// NOTE:
// Most of the source for this program is right here in this file,
// including the classes for the app's core, frame, and view.
// This is primarily to save myself the hassle of converting
// old code to a new architecture.
// In other words, it's much more convenient for me to have each
// class in the same scope as a handy set of global attributes.
// In _other_ words, I completely butchered the MFC document/view
// architecture so I wouldn't have to change my existing data code.
// In retrospect, that was a really sloppy approach; but it works,
// so feh.

#include "stdafx.h"
#include <float.h>
#include <direct.h>

#include "MapEdit.h"
#include "mapstuff.h"
#include "InfoDlg.h"
#include "SaveProgDlg.h"
#include "Rom.h"

#include "MainFrm.h"
#include "MapEditDoc.h"
#include "MapEditView.h"

#include "globals.h"
#include "classes.h"

#include "DestinationDlg.h"
#include "DoorDlg.h"
#include "SpriteDlg.h"
#include "MapMusicDlg.h"
#include "EnemyClusterDlg.h"
#include "EditorPaletteDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


// standard parameter definitions
#define SYSTEM_ID 3

#define UNDO_COUNT 1000

#define UNDO_SECTOR 0
#define UNDO_VALUE 1

#define TILE_PLACE 0		//
#define TILE_PICK 1			//
#define TILE_INFO 2			//
#define TILE_SELECT 3		// 

#define MODE_MAP 0			// Map editing mode
#define MODE_DOOR 1			// Door mode
#define MODE_SPRITE 2		// Sprite placement mode
#define MODE_ENEMY 3		// Enemy placement mode

#define TOOL_NONE 0			// No tool selected
#define TOOL_SELECT 1		// Selection marquee tool
#define TOOL_PAN 2			// Panning tool
#define TOOL_MAGNIFY 3		// Magnification tool

#define DOOR_NORMAL 0
#define DOOR_DESTSEEK 1
#define DOOR_ADD 2

#define SPRITE_NORMAL 0
#define SPRITE_ADD 1

#define SEEK_DESTDLGONLY 0
#define SEEK_DOORDLGFIRST 1


// additional structures

// Mouse struct is updated in the appropriate message handlers
struct MOUSEINFO
{
	int x;	//  
	int y;	// 

	UINT b;		// Buttons
} mouse;

struct INPUTINFO
{
	enum key_data
	{
		ctrl=0x1,
		alt=0x2,
		shift=0x4
	};
	UINT keys;
} input;

#define RENDER_NONE	0
#define RENDER_MAP	1
#define RENDER_TILE	2

struct DRAWINFO
{
	UINT render;
	unsigned short cx,cy;
	unsigned short tx,ty;
	bool kill_draw;
} draw;



// contains information about tile changes
struct UndoInfo {
	int type;

	int x;
	int y;

	// tile states
	int old_tile;
	int new_tile;
};


struct Clipboard {
	int isValid;
	int isUseable;
	unsigned char sector[4][8];
	unsigned char tileset[4][8];
};



// color animation data
COLORREF anim_clrs[16] = {
	RGB(0,		0,	255),
	RGB(64,		0,	255),
	RGB(128,	0,	255),
	RGB(192,	0,	255),

	RGB(255,	0,	255),

	RGB(255,	0,	192),
	RGB(255,	0,	128),
	RGB(255,	0,	64),

	RGB(255,	0,	128),
	RGB(255,	0,	192),
	RGB(255,	0,	255),

	RGB(192,	0,	255),
	RGB(128,	0,	255),
	RGB(64,		0,	255),
	RGB(0,		0,	255),
	RGB(0,		0,	255)
};

COLORREF boxcolor;

int drawcount = 0;
int cur_frame = 0;
int num_frames = 16;
int anim_speed = 150;
int timer_update = 0;




/***********************************\
| global attributes and behavior	|
\***********************************/

CString cstrpath;
CString* CEditorEnemyGroupDlg::curpath = &cstrpath;

//char curpath[256];

// function declarations

void LoadDoorData(FILE* rom);

void CreateAppData();
void LoadAppData(FILE* rom);

//void AddUndoState(int type, int &val_old);
void DoUndo();
void Copy();
void Paste();

BOOL IsCloseable();

//void Load();
//void Save();

void UpdateProg(int pos = -1);




// NEW GRAPHICS STUFF


int palstart[20] = {
	0x1A7EA7, 0x1A81A7, 0x1A83E7, 0x1A86E7, 0x1A8867,
	0x1A89E7, 0x1A8CE7, 0x1A9227, 0x1A92E7, 0x1A9467,
	0x1A98E7, 0x1ACEE7, 0x1AC8E7, 0x1AE7A7, 0x1ADD27,
	0x1AD967, 0x1AC3A7, 0x1AAAE7, 0x1AC2E7, 0x1AE327
}; // Why is this global? ...I should probably move this




// mode selection
int edit_mode = MODE_MAP;			// current editing mode
int tilemode = TILE_PLACE;			// current tile placement mode
int door_mode = DOOR_NORMAL;		// current door mode
int door_type = DEST_DOOR;
int door_rope = 0;
int sprite_mode = SPRITE_NORMAL;
int sprite_place = 0;				// TPT sprite to be placed on the map

int tool_mode = 0;




/////////////////////////////////
// Probably temporary: ROM OBJECT
CRom theRom;






// GUI data

// all this global GUI data presents something of a problem...
// It's needed by many parts of the program from within many
// different classes (perhaps a problem in itself), so it's hard
// to find one class in which to put it.

BOOL muzak = TRUE;

int xOffset = 0;					// Scroll offset from map origin
int yOffset = 0;					//


SelectedSector sectorlist[256];		// list of selected sectors
int sel_sector = 0;					// number of selected sectors

int place_tile = 0;					// Current tile for placement
int place_set = 0;					// Current tileset to use for placement

int mouse_x = 0;					// Current mouse position
int mouse_y = 0;					//

float zoomlevel = 1;				// zoom factor

BOOL needs_map_draw = TRUE;			// If TRUE, map is re-rendered

BOOL gridlines = TRUE;				// If TRUE, gridlines are drawn on the map

BOOL selectedinfo = TRUE;			// If TRUE, information about selected sectors is shown

BOOL spriteborder = TRUE;			// If TRUE, borders are drawn around sprites

BOOL sprite_outlines_only = FALSE;	// If TRUE, only sprite borders are shown, not actual sprite

BOOL drag_sprite_outline = FALSE;	// If TRUE, only sprite border is shown during dragging

BOOL color_anim = TRUE;				// If TRUE, object borders flash

BOOL auto_update = TRUE;			// Edit box handler control

BOOL infomode = FALSE;				// TRUE if tile info dialog is displayed

int infoX;							// Position of tile to display info for
int infoY;							//


DISPSPRINFO disp_spr[1024];			// array of sprites currently on-screen
int n_disp_spr = 0;



// drag 'n' drop data
int drag_area = -1;					// the area that contains the door being dragged
int drag_door = -1;					// the door being dragged
int drag_sprite = -1;
int spr_offx = 0;
int spr_offy = 0;
BOOL is_door_dragging = FALSE;		// TRUE if a door is being dragged
BOOL is_sprite_dragging = FALSE;


// door selection data
int sel_area = -1;					// area containing selected door
int sel_door = -1;					// door selected within area
int sel_sprite = -1;
BOOL is_door_selected = FALSE;		// TRUE if a door is selected
BOOL is_sprite_selected = FALSE;




// clipboard/state memory stuff

Clipboard clipboard;				// app clipboard

UndoInfo undobuffer[UNDO_COUNT+1];	// array of actions performed
int undocount = 0;					// number of actions performed
int undoptr = 0;					// current action

BOOL undoable = FALSE;				// TRUE if undo is available
BOOL redoable = FALSE;				// TRUE if redo is available


// windows
CDialogBar *dlgBar;					// pointer to app's dialog bar

//CToolsDlg toolwin;				// tileset window
//CDoorDlg doorwin;					// door entry editor window
//CDestinationDlg destwin;			// door destination editor window
//CSpriteDlg spritewin;				// sprite entry editor window
//CMapMusicDlg musicwin;				// music table editor window
CSaveProgDlg saveprogwin;			// load/save progress window

// bitmaps
CDC buffer;							// drawing buffer
CDC mapbuf;							// buffer to hold base map
int buf_x,buf_y;					// buffer coordinates


CDC crosshair;						// crosshair image for destination seeking









void CRom::LoadAppData(FILE* rom)
{
	// load the PK Hack system area into memory
	LoadSystemArea(rom, &sys);

	appdata = &(sys.entries[SYSTEM_ID]);

	if(!appdata->is_valid)
		FIRSTLOAD = true;
	else
		FIRSTLOAD = false;
}


void CRom::CreateAppData(SystemEntry *appdata)
{
	// this should be called prior to WriteSystemArea

	appdata->is_valid = true;

	// write number of doors into system entry
	appdata->data = new unsigned char[dest.num_dest+16];
	appdata->size = dest.num_dest+16;	// first 16 bytes reserved

	appdata->data[0] = dest.num_dest & 255;
	appdata->data[1] = (dest.num_dest >> 8) & 255;

	for(int i = 0; i < dest.num_dest; i++) {
		appdata->data[i+16] = dest.entries[i].type;
	}
}































void UpdateProg(int pos)
{
	if(pos == -1)
		saveprogwin.m_Prog.StepIt();
	else
		saveprogwin.m_Prog.OffsetPos(pos);
}


void CRom::Load()
{
	int i;

	// Load editable data
	saveprogwin.SetWindowText("Loading");

	saveprogwin.ShowWindow(SW_SHOWNORMAL);
	saveprogwin.m_Prog.SetRange(0, 1650);
	saveprogwin.m_Prog.SetPos(0);

	saveprogwin.SetDlgItemText(IDC_TEXT, "Loading...");

	// basic map attributes
	GetROMStuff(RomFile, map_tiles, local_sets, global_sets, global_pal, global_music);

	LoadSectorMap();
	LoadEnemyMap();

	LoadEnemyData();

	LoadItemData();

	//LoadEnemyBattleGroups();
	for(i = 0; i < CEnemyBattleGroup::num_entries; i++)
	{
		enemy_groups[i].ReadObject(RomFile, i);
	}

	//LoadEnemyClusters();
	for(i = 0; i < CEnemyCluster::num_entries; i++)
	{
		enemy_clusters[i].ReadObject(RomFile, i);
	}




	// application data from system area
	LoadAppData(RomFile);

	// 
	// 1320:
	LoadDoorDestinations(RomFile, &dest, appdata, UpdateProg);


	// Door entry table
	//LoadDoorEntries(det_first, det_last, FIRSTLOAD, UpdateProg);

	for(i = 0; i < 1280; i++)
		door_entry[i].ReadObject(RomFile, i, (DWORD)FIRSTLOAD);

	det_first = door_entry[0].address;
	det_last = door_entry[1279].address + door_entry[1279].GetSize();


	for(i = 0; i < 1280; i++) {
		for(int j = 0; j < door_entry[i].num_doors; j++) {
			// New stuff
			// (Still needs to be changed for stuff);
			door_entry[i][j].destindex = GetDestIndex(dest.entries, dest.num_dest, door_entry[i][j].destination);

			// old stuff: object areas
			//object_areas[i].doors[j].destindex = GetDestIndex(dest.entries, dest.num_dest, object_areas[i].doors[j].destination);
		}
	}

	// 250 > 
	// Sprite map entries (TBI w/ CDataEntry)
	LoadSpriteEntries(RomFile, object_areas, FIRSTLOAD, UpdateProg);



	// Sprite table data (also loads graphics)
	for(i = 0; i < CSprite::num_entries; i++)
	{
		sprites[i].ReadObject(RomFile, i, (DWORD)this);
	}
	// Temp: load sprite palettes
	for(i = 0; i < 8; i++)
		GetSpriteRGB(RomFile, i, pal[i]);


	LoadTPT(RomFile, tpt, UpdateProg);

	LoadMusicTable(RomFile, musictable);


	// Map palettes
	for(i = 0; i < CMapPalGroup::num_entries; i++)
	{
		map_pal[i].ReadObject(RomFile, i);
	}


// 144 >

	// graphics decompression
	// load tileset pointers
	fseek(RomFile, 0x2F125B, SEEK_SET);
	for(i = 0; i < 20; i++)
	{
		tileset_pointers[i] = fgetc(RomFile);
		tileset_pointers[i]+= fgetc(RomFile) << 8;
		tileset_pointers[i]+= fgetc(RomFile) << 16;
		tileset_pointers[i]+= fgetc(RomFile) << 24;
		UpdateProg();
	}
	fseek(RomFile, 0x2F12AB, SEEK_SET);
	for(i = 0; i < 20; i++)
	{
		arrange_pointers[i] = fgetc(RomFile);
		arrange_pointers[i]+= fgetc(RomFile) << 8;
		arrange_pointers[i]+= fgetc(RomFile) << 16;
		arrange_pointers[i]+= fgetc(RomFile) << 24;
		UpdateProg();
	}

	// load correlation table
	for(i = 0; i < 32; i++)
	{
		fseek(RomFile, 0x2F121B+i*2, SEEK_SET);
		int ch = fgetc(RomFile);
		set_conv[i] = ch;
		//doorwin.set_conv[i] = ch;
		//destwin.set_conv[i] = ch;
		UpdateProg();
	}

	// decompress stuff
	for(i = 0; i < 20; i++)
	{
		_decomp(RomFile, tileset_pointers[i] - 0xBFFE00, (unsigned char*)_tilesets[i], 28673);
		UpdateProg();
		_decomp(RomFile, arrange_pointers[i] - 0xBFFE00, (unsigned char*)arranges[i], 32768);
		UpdateProg();
	}

	// palette
	for(i = 0; i < 32; i++)
	{
		fseek(RomFile, 0x2F12FB+i*4, SEEK_SET);
		int adr;
		adr = fgetc(RomFile);
		adr+= fgetc(RomFile) << 8;
		adr+= fgetc(RomFile) << 16;
		adr+= fgetc(RomFile) << 24;
		adr-=0xBFFE00;

		for(int j = 0; j < 8; j++)
		{
			for(int k = 0; k < 6; k++) {
				for(int x = 0; x < 16; x++) {
					//y = *(unsigned short *)&rom[palstart[tileset] + 192*palette + 32*i + 2*x];
					
					fseek(RomFile, adr+j*192 + 32*k + 2*x, SEEK_SET);

					unsigned short y;
					y = fgetc(RomFile);
					y+= fgetc(RomFile) << 8;

					mpal_snes[i][j][k][x] = y;

					int red		= (((y & 0x1F) >> 0) * 8);
					int green	= (((y & 0x3E0) >> 5) * 8);
					int blue	= (((y & 0x7C00) >> 10) * 8);

					mpal[i][j][k][x]  = red << 16;
					mpal[i][j][k][x] |= green << 8;
					mpal[i][j][k][x] |= blue;

					//mpal[i][j][k][x] -= 0x101010;

					//destwin.mpal[i][j][k][x]  = (((y & 0x1F) >> 0) * 8) << 16;
					//destwin.mpal[i][j][k][x] |= (((y & 0x3E0) >> 5) * 8) << 8;
					//destwin.mpal[i][j][k][x] |= (((y & 0x7C00) >> 10) * 8) << 0;

					//doorwin.mpal[i][j][k][x]  = (((y & 0x1F) >> 0) * 8) << 16;
					//doorwin.mpal[i][j][k][x] |= (((y & 0x3E0) >> 5) * 8) << 8;
					//doorwin.mpal[i][j][k][x] |= (((y & 0x7C00) >> 10) * 8) << 0;
				}
			}
		}
		UpdateProg();
	}


	// Text
	//---theRom.CreateCCTable("codelist.txt", &theRom.ccTable);
	theRom.CreateCompressionTable(RomFile, theRom.comprTable);

	saveprogwin.ShowWindow(SW_HIDE);



	// The ROM is loaded now
	is_loaded = true;

}

void CRom::Save()
{
	saveprogwin.SetWindowText("Saving");
	saveprogwin.ShowWindow(SW_SHOWNORMAL);

	int total = 1280 + dest.num_dest;

	saveprogwin.m_Prog.SetRange(0, total);
	saveprogwin.m_Prog.SetPos(0);

	int save_flag = CheckExpand(RomFile);

	if(save_flag == 0 || save_flag == 1)
	{
		saveprogwin.SetDlgItemText(IDC_TEXT, "Saving map data...");
		SetROMStuff(RomFile, map_tiles, local_sets, global_sets, global_pal, global_music);
		WriteSectorMap();
		SaveEnemyMap();
	}

	if(save_flag == 1)
	{


		// Find the expanded area range
		GetExpandedRange(det_first, det_last);

		if(det_last - det_first > 0)
			ZeroRange(RomFile, det_first, det_last, (HexToSNES(det_first) >= 0xF00000));


		// Wipe the free part of the two-byte pointer-addressable area
		ZeroRange(RomFile, 0xF0200, 0xF8000, 0);

		// Reset the counter for the two-byte pointer area
		next_adr = 0xF0200;


		/************************************/
		/* Write destination table			*/
		/************************************/
		saveprogwin.SetDlgItemText(IDC_TEXT, "Writing door destination table...");
		WriteDoorDestinations(next_adr, UpdateProg);


		// okay, the old door data's been wiped out, and we've got the door destinations done
		// time to set the addresses for the door destinations now
		for(int i = 0; i < 1280; i++) {
			for(int j = 0; j < door_entry[i].num_doors; j++) {
				door_entry[i][j].destination = dest.entries[door_entry[i][j].destindex].address;
			}
		}


		// Manage expanded area space
		saveprogwin.SetDlgItemText(IDC_TEXT, "Organizing data...");


		// Data packs
		CDataPack cluster_table;	// Enemy cluster table
		CDataPack group_table;		// Group table

		CDataPack expanded;			// Package of expanded area data



		// Add the enemy clusters to the enemy cluster pack
		// until we run out of room

		int curpos = CEnemyCluster::def_data_start;
		for(i = 0; i < CEnemyCluster::num_entries; i++)
		{
			curpos += enemy_clusters[i].GetSize();

			if(curpos <= CEnemyCluster::def_data_extent)
				// We're clear, so write to normal area
				cluster_table.Add(&enemy_clusters[i]);

			else
				// We're out of space, so write to expanded area
				expanded.Add(&enemy_clusters[i]);
		}

		curpos = CEnemyBattleGroup::def_data_start;
		// Skip the first entry (this should not be modified anyway)
		// Also, there's a little data hiccup right around that first
		// entry, and it doesn't seem safe to write anything there
		for(i = 1; i < CEnemyBattleGroup::num_entries; i++)
		{
			curpos += enemy_groups[i].GetSize();

			if(curpos <= CEnemyBattleGroup::def_data_extent)
			{
				group_table.Add(&enemy_groups[i]);
			}
			else
			{
				expanded.Add(&enemy_groups[i]);
			}
		}



		// The entire door table goes to the expanded area, because we
		// need the room in the two-byte pointer area
		for(i = 0; i < 1280; i++)
			expanded.Add(&door_entry[i]);



		// Find space for the expanded area package
		int det_length = expanded.GetSize();


		int write_adr = FindFreeRange(RomFile, 0x400000, det_length);


		saveprogwin.SetDlgItemText(IDC_TEXT, "Writing data packages...");



		/************************************/
		/* Write data packages				*/
		/************************************/
		cluster_table.WritePack(RomFile, CEnemyCluster::def_data_start);
		group_table.WritePack(RomFile, CEnemyBattleGroup::def_data_start);
		expanded.WritePack(RomFile, write_adr);




		// reset space trackers
		det_first = expanded.GetStart();
		det_last = expanded.GetEnd();


		// Palette stuff
		CDataPack paltable;

		for(i = 0; i < CMapPalGroup::num_entries; i++)
		{
			paltable.Add(&map_pal[i]);
		}

		paltable.WritePack(RomFile, CMapPalGroup::def_data_start);






		// Misc data that has not yet been OO-implemented

		WriteSpriteTable(RomFile, object_areas, next_adr);	// TBI

		WriteTPT(RomFile, tpt);	// Indeterminate

		WriteMusicTable(RomFile, musictable, next_adr);		// TBI



		// Sys data: TBI
		saveprogwin.SetDlgItemText(IDC_TEXT, "Writing system area...");
		CreateAppData(appdata);
		WriteSystemArea(RomFile, sys);
	}
	saveprogwin.ShowWindow(SW_HIDE);
}














void AddUndoState(int x, int y, int newtile)
{
	undobuffer[undoptr].x = x;
	undobuffer[undoptr].y = y;

	undobuffer[undoptr].new_tile = newtile;
	undobuffer[undoptr].old_tile = theRom.map_tiles[y][x] + theRom.local_sets[y][x]*256;

	undoptr++;
	undocount = undoptr;

	// if we've reached the end of the buffer, bump back the pointer and all of the
	// history entries (the first entry gets erased)
	if(undoptr > UNDO_COUNT)
	{
		undoptr = UNDO_COUNT;

		for(int i = 0; i < UNDO_COUNT; i++)
		{
			undobuffer[i] = undobuffer[i+1];
		}
	}

	undoable = TRUE;
	redoable = FALSE;
}


void DoUndo()
{
	// reset tiles to their original state and decrement undoptr

	undoptr--;

	int x = undobuffer[undoptr].x;
	int y = undobuffer[undoptr].y;

	theRom.map_tiles[y][x] = undobuffer[undoptr].old_tile%256;
	theRom.local_sets[y][x] = undobuffer[undoptr].old_tile/256;

	redoable = TRUE;

	if(undoptr <= 0)
	{
		undoptr = 0;
		undoable = FALSE;
	}
}

void DoRedo()
{
	// reset tiles to their altered state and increment undoptr

	int x = undobuffer[undoptr].x;
	int y = undobuffer[undoptr].y;

	theRom.map_tiles[y][x] = undobuffer[undoptr].new_tile%256;
	theRom.local_sets[y][x] = undobuffer[undoptr].new_tile/256;

	undoptr++;

	undoable = TRUE;

	if(undoptr >= undocount)
	{
		undoptr = undocount;
		redoable = FALSE;
	}
}












void Copy()
{
	/*clipboard.isValid = TRUE;
	for(int i = 0; i < 4; i++) {
		for(int j = 0; j < 8; j++) {
			clipboard.sector[i][j] = theRom.map_tiles[i+sector_y*4][j+sector_x*8];
			clipboard.tileset[i][j] = theRom.local_sets[i+sector_y*4][j+sector_x*8];
		}
	}*/
}

void Paste()
{
	/*AddUndoState(sector_x*8,sector_y*4,8,4);
	for(int i = 0; i < 4; i++) {
		for(int j = 0; j < 8; j++) {
			theRom.map_tiles[i+sector_y*4][j+sector_x*8] = clipboard.sector[i][j];
			theRom.local_sets[i+sector_y*4][j+sector_x*8] = clipboard.tileset[i][j]; 
		}
	}*/
}





/////////////////////////////////////////////////////////////////////////////
// CMapEditApp

BEGIN_MESSAGE_MAP(CMapEditApp, CWinApp)
	//{{AFX_MSG_MAP(CMapEditApp)
	ON_COMMAND(ID_APP_ABOUT, OnAppAbout)
	//}}AFX_MSG_MAP
	// Standard file based document commands
	ON_COMMAND(ID_FILE_NEW, CWinApp::OnFileNew)
	ON_COMMAND(ID_FILE_OPEN, CWinApp::OnFileOpen)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CMapEditApp construction

CMapEditApp::CMapEditApp()
{
	// TODO: add construction code here,
	// Place all significant initialization in InitInstance
}

/////////////////////////////////////////////////////////////////////////////
// The one and only CMapEditApp object

CMapEditApp theApp;

/////////////////////////////////////////////////////////////////////////////
// CMapEditApp initialization

BOOL CMapEditApp::InitInstance()
{
	AfxEnableControlContainer();

	// Standard initialization
	// If you are not using these features and wish to reduce the size
	//  of your final executable, you should remove from the following
	//  the specific initialization routines you do not need.

#ifdef _AFXDLL
	Enable3dControls();			// Call this when using MFC in a shared DLL
#else
	Enable3dControlsStatic();	// Call this when linking to MFC statically
#endif

	// Change the registry key under which our settings are stored.
	// TODO: You should modify this string to be something appropriate
	// such as the name of your company or organization.
	SetRegistryKey(_T("EarthBound Map Editor"));

	LoadStdProfileSettings();  // Load standard INI file options (including MRU)

	// Register the application's document templates.  Document templates
	//  serve as the connection between documents, frame windows and views.

	CSingleDocTemplate* pDocTemplate;
	pDocTemplate = new CSingleDocTemplate(
		IDR_MAINFRAME,
		RUNTIME_CLASS(CMapEditDoc),
		RUNTIME_CLASS(CMainFrame),       // main SDI frame window
		RUNTIME_CLASS(CMapEditView));
	AddDocTemplate(pDocTemplate);

	// Parse command line for standard shell commands, DDE, file open
	CCommandLineInfo cmdInfo;
	ParseCommandLine(cmdInfo);

	// Dispatch commands specified on the command line
	if (!ProcessShellCommand(cmdInfo))
		return FALSE;

	// The one and only window has been initialized, so show and update it.
	m_pMainWnd->ShowWindow(SW_SHOW);
	m_pMainWnd->UpdateWindow();

	return TRUE;
}





/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
		// No message handlers
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
		// No message handlers
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

// App command to run the dialog
void CMapEditApp::OnAppAbout()
{
	CAboutDlg aboutDlg;
	aboutDlg.DoModal();
}

/////////////////////////////////////////////////////////////////////////////
// CMapEditApp message handlers

CDocument* CMapEditApp::OpenDocumentFile(LPCTSTR lpszFileName) 
{
	AfxMessageBox("This feature is temporarily disabled in the 4.0 beta release due to some annoying problem or other. I'm sure you're not too lazy to just use the stupid \"Open\" button.");
	return NULL;
	/*
	CMainFrame* frm = (CMainFrame*)GetMainWnd();

	frm->CloseFile();

	CDocument* temp = CWinApp::OpenDocumentFile(lpszFileName);

	if(CheckFile(lpszFileName)) {
		frm->LoadFile(lpszFileName);
	}

	
	
	return temp;
*/
	//return CWinApp::OpenDocumentFile(lpszFileName);
}


void CMapEditApp::AddToRecentFileList(LPCTSTR lpszPathName) 
{
	//for(int i = 3; i > 1; i--) {
	//	MRU[i] = MRU[i-1];
	//}
	//MRU[0] = lpszPathName;

	CWinApp::AddToRecentFileList(lpszPathName);
}
















/////////////////////////////////////////////////////////////////////////////
// CMainFrame

IMPLEMENT_DYNCREATE(CMainFrame, CFrameWnd)

BEGIN_MESSAGE_MAP(CMainFrame, CFrameWnd)
	//{{AFX_MSG_MAP(CMainFrame)
	ON_WM_CREATE()
	ON_COMMAND(ID_FILE_OPEN, OnFileOpen)
	ON_CBN_SELCHANGE(IDC_COMBO_TILESET, OnSelectTileset)
	ON_EN_CHANGE(IDC_EDIT_X, OnEditX)
	ON_EN_CHANGE(IDC_EDIT_Y, OnEditY)
	ON_EN_CHANGE(IDC_EDIT_PALETTE, OnEditPalette)
	ON_EN_CHANGE(IDC_EDIT_MUSIC, OnEditMusic)
	ON_COMMAND(ID_VIEW_GRIDLINES, OnViewGridlines)
	ON_UPDATE_COMMAND_UI(ID_VIEW_GRIDLINES, OnUpdateViewGridlines)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	ON_COMMAND(ID_EDIT_PASTE, OnEditPaste)
	ON_UPDATE_COMMAND_UI(ID_EDIT_COPY, OnUpdateEditCopy)
	ON_UPDATE_COMMAND_UI(ID_EDIT_PASTE, OnUpdateEditPaste)
	ON_COMMAND(ID_EDIT_UNDO, OnEditUndo)
	ON_UPDATE_COMMAND_UI(ID_EDIT_UNDO, OnUpdateEditUndo)
	ON_WM_KEYDOWN()
	ON_WM_KEYUP()
	ON_COMMAND(ID_FILE_CLOSE, OnFileClose)
	ON_UPDATE_COMMAND_UI(ID_FILE_CLOSE, OnUpdateFileClose)
	ON_UPDATE_COMMAND_UI(ID_FILE_SAVE, OnUpdateFileSave)
	ON_COMMAND(ID_TOOLS_DESTINATIONEDIT, OnToolsDestination)
	ON_WM_TIMER()
	ON_COMMAND(ID_MODE_DOOR, OnModeDoor)
	ON_UPDATE_COMMAND_UI(ID_MODE_DOOR, OnUpdateModeDoor)
	ON_COMMAND(ID_MODE_MAP, OnModeMap)
	ON_UPDATE_COMMAND_UI(ID_MODE_MAP, OnUpdateModeMap)
	ON_COMMAND(ID_MODE_SPRITE, OnModeSprite)
	ON_UPDATE_COMMAND_UI(ID_MODE_SPRITE, OnUpdateModeSprite)
	ON_COMMAND(ID_TOOLS_ADDDOOR, OnToolsAddDoor)
	ON_UPDATE_COMMAND_UI(ID_TOOLS_ADDDOOR, OnUpdateToolsAddDoor)
	ON_UPDATE_COMMAND_UI(ID_TOOLS_DESTINATIONEDIT, OnUpdateToolsDestinationEdit)
	ON_COMMAND(ID_TOOLS_ADDSPRITE, OnToolsAddSprite)
	ON_UPDATE_COMMAND_UI(ID_TOOLS_ADDSPRITE, OnUpdateToolsAddSprite)
	ON_COMMAND(ID_VIEW_SPRITEBORDERS, OnViewSpriteborders)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SPRITEBORDERS, OnUpdateViewSpriteborders)
	ON_COMMAND(ID_VIEW_SPRITEOUTLINES, OnViewSpriteOutlines)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SPRITEOUTLINES, OnUpdateViewSpriteOutlines)
	ON_COMMAND(ID_VIEW_DRAGSPRITEOUTLINE, OnViewDragSpriteOutline)
	ON_UPDATE_COMMAND_UI(ID_VIEW_DRAGSPRITEOUTLINE, OnUpdateViewDragSpriteOutline)
	ON_COMMAND(ID_VIEW_BORDERFLASH, OnViewBorderFlash)
	ON_UPDATE_COMMAND_UI(ID_VIEW_BORDERFLASH, OnUpdateViewBorderFlash)
	ON_COMMAND(ID_TOOLS_REMOVEOBJECT, OnToolsRemoveObject)
	ON_UPDATE_COMMAND_UI(ID_TOOLS_REMOVEOBJECT, OnUpdateToolsRemoveObject)
	ON_COMMAND(ID_TOOLS_TPTEDITOR, OnToolsTPTEditor)
	ON_UPDATE_COMMAND_UI(ID_TOOLS_TPTEDITOR, OnUpdateToolsTPTEditor)
	ON_COMMAND(ID_TOOLS_ADDLADDER, OnToolsAddladder)
	ON_UPDATE_COMMAND_UI(ID_TOOLS_ADDLADDER, OnUpdateToolsAddladder)
	ON_COMMAND(ID_TOOLS_ADDROPE, OnToolsAddrope)
	ON_UPDATE_COMMAND_UI(ID_TOOLS_ADDROPE, OnUpdateToolsAddrope)
	ON_COMMAND(ID_TOOLS_MUSICTABLEEDITOR, OnToolsMusicEditor)
	ON_UPDATE_COMMAND_UI(ID_TOOLS_MUSICTABLEEDITOR, OnUpdateToolsMusicEditor)
	ON_COMMAND(ID_VIEW_TOOLBOX, OnViewToolBox)
	ON_UPDATE_COMMAND_UI(ID_VIEW_TOOLBOX, OnUpdateViewToolBox)
	ON_COMMAND(ID_VIEW_ATTRIBUTEBAR, OnViewAttributeBar)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ATTRIBUTEBAR, OnUpdateViewAttributeBar)
	ON_COMMAND(ID_VIEW_SELECTEDINFO, OnViewSelectedInfo)
	ON_UPDATE_COMMAND_UI(ID_VIEW_SELECTEDINFO, OnUpdateViewSelectedInfo)
	ON_COMMAND(ID_EDIT_REPEAT, OnEditRepeat)
	ON_UPDATE_COMMAND_UI(ID_EDIT_REPEAT, OnUpdateEditRepeat)
	ON_WM_SIZE()
	ON_COMMAND(ID_MODE_ENEMY, OnModeEnemy)
	ON_UPDATE_COMMAND_UI(ID_MODE_ENEMY, OnUpdateModeEnemy)
	ON_COMMAND(ID_HELP_TEST, OnHelpTest)
	ON_COMMAND(ID_TOOLS_ENEMYCLUSTER, OnToolsEnemycluster)
	ON_UPDATE_COMMAND_UI(ID_TOOLS_ENEMYCLUSTER, OnUpdateToolsEnemycluster)
	ON_COMMAND(ID_TOOLS_PALETTEEDITOR, OnToolsPaletteeditor)
	ON_UPDATE_COMMAND_UI(ID_TOOLS_PALETTEEDITOR, OnUpdateToolsPaletteeditor)
	ON_COMMAND(ID_VIEW_ZOOM_100, OnViewZoom100)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ZOOM_100, OnUpdateViewZoom100)
	ON_COMMAND(ID_VIEW_ZOOM_150, OnViewZoom150)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ZOOM_150, OnUpdateViewZoom150)
	ON_COMMAND(ID_VIEW_ZOOM_200, OnViewZoom200)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ZOOM_200, OnUpdateViewZoom200)
	ON_COMMAND(ID_VIEW_ZOOM_300, OnViewZoom300)
	ON_UPDATE_COMMAND_UI(ID_VIEW_ZOOM_300, OnUpdateViewZoom300)
	ON_COMMAND(ID_TOOLS_ENEMYGROUPEDITOR, OnToolsEnemygroupeditor)
	ON_UPDATE_COMMAND_UI(ID_TOOLS_ENEMYGROUPEDITOR, OnUpdateToolsEnemygroupeditor)
	ON_COMMAND(ID_FUNSTUFF_MUZAK, OnFunstuffMuzak)
	ON_UPDATE_COMMAND_UI(ID_FUNSTUFF_MUZAK, OnUpdateFunstuffMuzak)
	ON_WM_MOUSEMOVE()
	ON_COMMAND(ID_FILE_CLOSEROM, OnFileClose)
	ON_UPDATE_COMMAND_UI(ID_FILE_CLOSEROM, OnUpdateFileClose)
	ON_WM_SIZING()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

static UINT indicators[] =
{
	ID_SEPARATOR,           // status line indicator
	ID_INDICATOR_CAPS,
	ID_INDICATOR_NUM,
	ID_INDICATOR_SCRL,
};

/////////////////////////////////////////////////////////////////////////////
// CMainFrame construction/destruction

CMainFrame::CMainFrame()
{
	// TODO: add member initialization code here
	
}

CMainFrame::~CMainFrame()
{
}

int CMainFrame::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
	if (CFrameWnd::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	if (!m_wndToolBar.CreateEx(this, TBSTYLE_FLAT, WS_CHILD | WS_VISIBLE | CBRS_TOP
		| CBRS_GRIPPER | CBRS_TOOLTIPS | CBRS_FLYBY | CBRS_SIZE_DYNAMIC | CBRS_FLOAT_MULTI) ||
		!m_wndToolBar.LoadToolBar(IDR_MAINFRAME))
	{
		TRACE0("Failed to create toolbar\n");
		return -1;      // fail to create
	}

	if (!m_wndDlgBar.Create(this, IDR_MAINFRAME, 
		CBRS_ALIGN_TOP|CBRS_GRIPPER, 0x1000))
	{
		TRACE0("Failed to create dialogbar\n");
		return -1;		// fail to create
	}


	m_wndToolboxBar.Create(this, IDR_TOOLBOX, CBRS_ALIGN_LEFT|CBRS_FLOAT_MULTI, 0x1001);


	if (!m_wndStatusBar.Create(this) ||
		!m_wndStatusBar.SetIndicators(indicators,
		  sizeof(indicators)/sizeof(UINT)))
	{
		TRACE0("Failed to create status bar\n");
		return -1;      // fail to create
	}

	// TODO: Remove this if you don't want tool tips
	m_wndToolBar.SetBarStyle(m_wndToolBar.GetBarStyle() |
		CBRS_TOOLTIPS | CBRS_FLYBY);

	EnableDocking(CBRS_ALIGN_ANY);

	m_wndToolboxBar.EnableDocking(CBRS_ALIGN_LEFT|CBRS_ALIGN_RIGHT);
	DockControlBar(&m_wndToolboxBar, AFX_IDW_DOCKBAR_LEFT);

	m_wndToolBar.EnableDocking(CBRS_ALIGN_ANY);
	DockControlBar(&m_wndToolBar, AFX_IDW_DOCKBAR_TOP);

	m_wndDlgBar.EnableDocking(CBRS_ALIGN_TOP|CBRS_ALIGN_BOTTOM);
	DockControlBar(&m_wndDlgBar);


	LoadBarState("ControlBars");


	// create some bitmaps for later use

	CDC* dc = GetDC();

	buffer.CreateCompatibleDC(dc);
	SelectObject((HDC)buffer,CreateCompatibleBitmap((HDC)*GetDC(),1280,1024));

	mapbuf.CreateCompatibleDC(dc);
	SelectObject((HDC)mapbuf,CreateCompatibleBitmap((HDC)*GetDC(),1280,1024));


	crosshair.CreateCompatibleDC(dc);
	SelectObject((HDC)crosshair, LoadImage(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDB_CROSSHAIR2), IMAGE_BITMAP, 0, 0, 0));
	crosshair.SelectStockObject(GRAY_BRUSH);

	buffer.SetBkMode(TRANSPARENT);
	buffer.SetTextColor(RGB(244,244,244));

	ReleaseDC(dc);


	dlgBar = &m_wndDlgBar;

	// create some windows
	saveprogwin.Create(IDD_SAVEPROGDLG, this);



	// give toolbar pointer to ROM object
	m_wndToolboxBar.Rom = &theRom;

	m_wndToolboxBar.m_pParent = this;

	m_wndToolboxBar.Initialize(this);


	// set pretty animation timer
	SetTimer(0, anim_speed, NULL);

	// initialize some control bar thingies
	auto_update = FALSE;
	m_wndDlgBar.SetDlgItemInt(IDC_EDIT_X, 0);
	m_wndDlgBar.SetDlgItemInt(IDC_EDIT_Y, 0);
	auto_update = TRUE;

	m_wndStatusBar.SetPaneInfo(1, 1024, SBPS_NOBORDERS, 0);
	m_wndStatusBar.SetPaneInfo(2, 1025, SBPS_NORMAL, 100);
	m_wndStatusBar.SetPaneInfo(3, 1026, SBPS_NORMAL, 100);

	m_wndStatusBar.SetPaneText(1, "");
	m_wndStatusBar.SetPaneText(2, "");
	m_wndStatusBar.SetPaneText(3, "");


#ifndef _DEBUG
	theRom.CreateCCTable("codelist.txt", &theRom.ccTable);
#endif


	_getcwd(curpath, 256);

	cstrpath.Format("%s", curpath);

	theRom.Path.Format("%s", curpath);

	return 0;
}

BOOL CMainFrame::PreCreateWindow(CREATESTRUCT& cs)
{
	if( !CFrameWnd::PreCreateWindow(cs) )
		return FALSE;
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return TRUE;
}

BOOL CMainFrame::DestroyWindow() 
{
	// Save control bar states
	SaveBarState("ControlBars");
	return CFrameWnd::DestroyWindow();
}

/////////////////////////////////////////////////////////////////////////////
// CMainFrame diagnostics

#ifdef _DEBUG
void CMainFrame::AssertValid() const
{
	CFrameWnd::AssertValid();
}

void CMainFrame::Dump(CDumpContext& dc) const
{
	CFrameWnd::Dump(dc);
}

#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CMainFrame message handlers


void CMainFrame::OnSize(UINT nType, int cx, int cy) 
{
	//draw.kill_draw = true;
	CFrameWnd::OnSize(nType, cx, cy);
	//draw.kill_draw = false;
	
	//draw.kill_draw = false;
	// invalidate the map
	//needs_map_draw = TRUE;
	//draw.render = RENDER_MAP;

	// recalculate the layout of the toolbar
	if(m_wndToolboxBar.m_hWnd)	// only resize if it's been created
		m_wndToolboxBar.CalcLayout();
}



void CMainFrame::OnSizing(UINT fwSide, LPRECT pRect) 
{
	draw.kill_draw = true;
	CFrameWnd::OnSizing(fwSide, pRect);
}


void CMainFrame::OnTimer(UINT nIDEvent) 
{
	// TODO: Add your message handler code here and/or call default
	if(color_anim)
	{
		if(edit_mode == MODE_DOOR || (edit_mode == MODE_SPRITE && spriteborder))
		{
			timer_update = 1;
			GetActiveView()->Invalidate(FALSE);
		}
	}

	CFrameWnd::OnTimer(nIDEvent);
}


void CMainFrame::OnFileOpen() 
{

	// give the view a pointer to the toolbar (it needs it)
	((CMapEditView*)GetActiveView())->m_pToolbox = &m_wndToolboxBar;

	// Note: why does this have to be in the opening dealy?
	// stupid piece of crap...


	// Display OpenFileName dialog, get filename,
	// Call LoadFile to load data
	OPENFILENAME ofn;       // common dialog box structure
	char szFile[260];       // buffer for filename

	szFile[0] = NULL;

	// Initialize OPENFILENAME
	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.lpstrFile = szFile;
	ofn.nMaxFile = sizeof(szFile);
	ofn.lpstrFilter = "SNES ROM files (*.smc)\0*.SMC\0";
	ofn.nFilterIndex = 1;
	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

	CFileDialog dlg(TRUE, "smc", NULL, OFN_PATHMUSTEXIST|OFN_FILEMUSTEXIST, "SNES ROM files (*.smc)\0*.SMC\0", this);

	dlg.m_ofn = ofn;


	// Display the Open dialog box
	int nResult = dlg.DoModal();

	if(nResult == IDOK) {
		if(CheckFile(szFile)) {
			LoadFile(szFile);
			theApp.AddToRecentFileList(szFile);
			theRom.FileName = szFile;
		}
	}
}

void CMainFrame::LoadFile(CString szFilename)
{
	theRom.RomFile = fopen(szFilename,"rb+");

	// enable dialog bar controls
	dlgBar->GetDlgItem(IDC_EDIT_X)->EnableWindow();
	dlgBar->GetDlgItem(IDC_EDIT_Y)->EnableWindow();
	dlgBar->GetDlgItem(IDC_COMBO_TILESET)->EnableWindow();
	dlgBar->GetDlgItem(IDC_EDIT_PALETTE)->EnableWindow();
	dlgBar->GetDlgItem(IDC_EDIT_MUSIC)->EnableWindow();

	SetMessageText("Loading...");

	theRom.Load();

	SetMessageText("Ready");

	needs_map_draw = TRUE;
	draw.render = RENDER_MAP;

	GetActiveView()->Invalidate(FALSE);

	m_wndToolboxBar.SendMessage(TM_RELOAD_ROM);

	if(m_wndToolboxBar.GetGlobalSet() == -1)
	{
		m_wndToolboxBar.SetGlobalSet(theRom.global_sets[0][0]);
		m_wndToolboxBar.SetGlobalPal(theRom.global_pal[0][0]);
	}

	m_wndToolboxBar.m_wndSpriteBar.Invalidate(FALSE);
	
	//mv->SelectSector(0,0);


	CString str;
	str.LoadString(AFX_IDS_APP_TITLE);
	str.Insert(0, " - ");
	str.Insert(0, szFilename.Right(szFilename.GetLength()-szFilename.ReverseFind('\\')-1));

	SetWindowText(str);
}

void CMainFrame::CloseFile()
{
	//fclose(RomFile);
	_fcloseall();

	theRom.is_loaded = FALSE;

	dlgBar->GetDlgItem(IDC_EDIT_X)->EnableWindow(FALSE);
	dlgBar->GetDlgItem(IDC_EDIT_Y)->EnableWindow(FALSE);
	dlgBar->GetDlgItem(IDC_COMBO_TILESET)->EnableWindow(FALSE);
	dlgBar->GetDlgItem(IDC_EDIT_PALETTE)->EnableWindow(FALSE);
	dlgBar->GetDlgItem(IDC_EDIT_MUSIC)->EnableWindow(FALSE);

	for(int i = 0; i < 320; i++) {
		for(int j = 0; j < 256; j++) {
			theRom.map_tiles[i][j] = 0;
			theRom.local_sets[i][j] = 0;
		}
	}
	for(i = 0; i < 80; i++) {
		for(int j = 0; j < 32; j++) {
			theRom.global_sets[i][j] = 0;
			theRom.global_pal[i][j] = 0;
			theRom.global_music[i][j] = 0;
		}
	}
	for(i = 0; i < 1280; i++)
	{
		theRom.door_entry[i].num_doors = 0;
		theRom.object_areas[i].num_sprites = 0;
	}
	for(i = 0; i < 160; i++)
		for(int j = 0; j < 128; j++)
			theRom.enemy_map[i][j] = 0;

	needs_map_draw = true;
	draw.render = RENDER_MAP;

	GetActiveView()->Invalidate(FALSE);

	CMapEditView* mv = (CMapEditView*)GetActiveView();
	mv->SelectSector(-1,-1);

	//toolwin.gSet = -1;
	//toolwin.Invalidate(FALSE);
	m_wndToolboxBar.SetGlobalSet(-1);

	CString str;
	str.LoadString(AFX_IDS_APP_TITLE);
	SetWindowText(str);
}

void CMapEditDoc::OnFileSave() 
{
	POSITION pos = GetFirstViewPosition();

	CMainFrame *w = ((CMainFrame*)GetNextView(pos)->GetParent());
		
	w->SetMessageText("Saving...");

	theRom.Save();

	w->SetMessageText("Ready");
}

void CMapEditDoc::OnFileSaveAs() 
{
	OnFileSave();
}

BOOL IsCloseable()
{
	CString prompt;
	BOOL retVal = FALSE;

	if(theRom.is_loaded) {
		prompt.Format("Save changes to %s?", theRom.FileName);

		int exit = AfxMessageBox(prompt, MB_YESNOCANCEL);

		if(exit == IDYES) {
			theRom.Save();
			retVal = TRUE;
		}
		if(exit == IDNO) {
			retVal = TRUE;
		}
	}
	else
		retVal = TRUE;

	return retVal;
}


BOOL CMapEditDoc::CanCloseFrame(CFrameWnd* pFrame) 
{
	return IsCloseable();//CDocument::CanCloseFrame(pFrame);
}



void CMainFrame::SwitchEditMode(int nMode, BOOL bRedraw)
{
	if(!this || m_hWnd == NULL) return;

	if(nMode != -1)
		edit_mode = nMode;

	if(dlgBar)
	{
		dlgBar->GetDlgItem(IDC_COMBO_TILESET)->EnableWindow((edit_mode == MODE_MAP) && theRom.is_loaded);
		dlgBar->GetDlgItem(IDC_EDIT_PALETTE)->EnableWindow((edit_mode == MODE_MAP) && theRom.is_loaded);
		dlgBar->GetDlgItem(IDC_EDIT_MUSIC)->EnableWindow((edit_mode == MODE_MAP) && theRom.is_loaded);
	}
	//if(bRedraw)
	{
		needs_map_draw = TRUE;
		draw.render = RENDER_MAP;

		GetActiveView()->Invalidate(FALSE);
		GetActiveView()->SendMessage(WM_PAINT);
	}
}

int CMainFrame::GetEditMode()
{
	return edit_mode;
}

BOOL CMainFrame::IsDoorSeeking()
{
	return (door_mode == DOOR_DESTSEEK);
}


void CMainFrame::OnFileClose() 
{
	if(IsCloseable())
		CloseFile();
}

void CMainFrame::OnUpdateFileClose(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(theRom.is_loaded);
}

void CMainFrame::OnUpdateFileSave(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(theRom.is_loaded);
}




void CMainFrame::OnEditCopy() 
{
	Copy();
}

void CMainFrame::OnEditPaste() 
{
	Paste();
	GetActiveView()->Invalidate(FALSE);
}


void CMainFrame::OnUpdateEditCopy(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(clipboard.isUseable && edit_mode == MODE_MAP && theRom.is_loaded);
}

void CMainFrame::OnUpdateEditPaste(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(clipboard.isValid && edit_mode == MODE_MAP && theRom.is_loaded);
}



void CMainFrame::OnEditUndo() 
{
	DoUndo();

	needs_map_draw = TRUE;
	draw.render = RENDER_MAP;
	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnUpdateEditUndo(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(undoable && edit_mode == MODE_MAP && theRom.is_loaded);
}

void CMainFrame::OnEditRepeat() 
{
	DoRedo();

	needs_map_draw = TRUE;
	draw.render = RENDER_MAP;
	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnUpdateEditRepeat(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(redoable && edit_mode == MODE_MAP && theRom.is_loaded);
}


// view stuff


void CMainFrame::OnViewToolBox() 
{
	// TODO: Add your command handler code here
	ShowControlBar(&m_wndToolboxBar, !m_wndToolboxBar.IsWindowVisible(), FALSE);
}

void CMainFrame::OnUpdateViewToolBox(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->SetCheck(m_wndToolboxBar.IsWindowVisible());
}

void CMainFrame::OnViewAttributeBar() 
{
	// TODO: Add your command handler code here
	ShowControlBar(&m_wndDlgBar, !m_wndDlgBar.IsWindowVisible(), FALSE);
}

void CMainFrame::OnUpdateViewAttributeBar(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->SetCheck(m_wndDlgBar.IsWindowVisible());
}




void CMainFrame::OnViewSpriteborders() 
{
	spriteborder = !spriteborder;
	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnUpdateViewSpriteborders(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(!sprite_outlines_only);
	pCmdUI->SetCheck(spriteborder);
}

void CMainFrame::OnViewGridlines() 
{
	gridlines = !gridlines;
	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnUpdateViewGridlines(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(gridlines);
}



void CMainFrame::OnViewSelectedInfo() 
{
	selectedinfo = !selectedinfo;
	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnUpdateViewSelectedInfo(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(selectedinfo);
}


void CMainFrame::OnViewSpriteOutlines() 
{
	sprite_outlines_only = !sprite_outlines_only;

	if(sprite_outlines_only)
		spriteborder = TRUE;

	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnUpdateViewSpriteOutlines(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(sprite_outlines_only);
	
}



void CMainFrame::OnViewDragSpriteOutline() 
{
	// TODO: Add your command handler code here
	drag_sprite_outline = !drag_sprite_outline;
}

void CMainFrame::OnUpdateViewDragSpriteOutline(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->SetCheck(drag_sprite_outline);
}

void CMainFrame::OnViewBorderFlash() 
{
	// TODO: Add your command handler code here
	color_anim = !color_anim;
}

void CMainFrame::OnUpdateViewBorderFlash(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->SetCheck(color_anim);
}




// MODE SELECTORS

void CMainFrame::OnModeMap() 
{
	SwitchEditMode(MODE_MAP, FALSE);
	m_wndToolboxBar.SwitchEditMode(MODE_MAP);
	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnUpdateModeMap(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio((edit_mode==MODE_MAP));

	// don't allow mode switch while seeking
	pCmdUI->Enable(door_mode != DOOR_DESTSEEK);
}

void CMainFrame::OnModeDoor() 
{
	SwitchEditMode(MODE_DOOR, FALSE);
	m_wndToolboxBar.SwitchEditMode(MODE_DOOR);
	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnUpdateModeDoor(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio((edit_mode==MODE_DOOR));

	// don't allow mode switch while seeking
	pCmdUI->Enable(door_mode != DOOR_DESTSEEK);
}

void CMainFrame::OnModeSprite() 
{
	SwitchEditMode(MODE_SPRITE, FALSE);
	m_wndToolboxBar.SwitchEditMode(MODE_SPRITE);
	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnUpdateModeSprite(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(edit_mode == MODE_SPRITE);

	// don't allow mode switch while seeking
	pCmdUI->Enable(door_mode != DOOR_DESTSEEK);
}


void CMainFrame::OnModeEnemy() 
{
	// TODO: Add your command handler code here
	SwitchEditMode(MODE_ENEMY, FALSE);
	m_wndToolboxBar.SwitchEditMode(MODE_ENEMY);
	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnUpdateModeEnemy(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->SetRadio(edit_mode == MODE_ENEMY);

	// don't allow mode switch while seeking
	pCmdUI->Enable(door_mode != DOOR_DESTSEEK);
}


void CMainFrame::OnToolsAddDoor() 
{
	// TODO: Add your command handler code here
	m_wndStatusBar.SetWindowText("Click a location for the new door. (Right-click to cancel)");

	door_mode = DOOR_ADD;
	door_type = DEST_DOOR;
	door_rope = 0;

	is_door_selected = FALSE;

	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnToolsAddladder() 
{
	OnToolsAddDoor();
		door_type = DEST_ROPE;
		door_rope = 0;

}

void CMainFrame::OnUpdateToolsAddladder(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(edit_mode == MODE_DOOR && door_mode != DOOR_DESTSEEK && theRom.is_loaded);
	pCmdUI->SetRadio(door_mode == DOOR_ADD);
}

void CMainFrame::OnToolsAddrope() 
{
	OnToolsAddDoor();
		door_type = DEST_ROPE;
		door_rope = 1;
}

void CMainFrame::OnUpdateToolsAddrope(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(edit_mode == MODE_DOOR && door_mode != DOOR_DESTSEEK && theRom.is_loaded);
	pCmdUI->SetRadio(door_mode == DOOR_ADD);
}

void CMainFrame::OnUpdateToolsAddDoor(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(edit_mode == MODE_DOOR && door_mode != DOOR_DESTSEEK && theRom.is_loaded);
	pCmdUI->SetRadio(door_mode == DOOR_ADD);
}

void CMainFrame::OnToolsAddSprite() 
{
	sprite_mode = SPRITE_ADD;

	m_wndStatusBar.SetWindowText("Click a location for the new sprite. (Right-click to cancel)");

	GetActiveView()->Invalidate(FALSE);
}

void CMainFrame::OnUpdateToolsAddSprite(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(edit_mode == MODE_SPRITE);
}


void CMainFrame::OnToolsRemoveObject() 
{
	// TODO: Add your command handler code here
	if(edit_mode == MODE_DOOR)
	{
		theRom.door_entry[sel_area].RemDoor(sel_door);
		//theRom.RemDoor(sel_area, sel_door);
		is_door_selected = FALSE;
	}
	else if(edit_mode == MODE_SPRITE)
	{
		theRom.RemSprite(sel_area, sel_sprite);
		is_sprite_selected = FALSE;
	}
	GetActiveView()->Invalidate(FALSE);

}

void CMainFrame::OnUpdateToolsRemoveObject(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(theRom.is_loaded && ((edit_mode == MODE_SPRITE && is_sprite_selected) || (edit_mode == MODE_DOOR && is_door_selected)));
}



void CMainFrame::OnToolsDestination() 
{
	OpenDestinationDlg(0);
}

void CMainFrame::OnUpdateToolsDestinationEdit(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(theRom.is_loaded && edit_mode == MODE_DOOR);
}


void CMainFrame::OnToolsTPTEditor() 
{
	// Show TPT editor window
	
}

void CMainFrame::OnUpdateToolsTPTEditor(CCmdUI* pCmdUI) 
{
	
}



void CMainFrame::OnToolsMusicEditor() 
{
	// TODO: Add your command handler code here
	int entry = theRom.global_music[sectorlist[0].y][sectorlist[0].x];
	CMapMusicDlg dlg(&theRom, entry);

	dlg.DoModal();

	//for(int i = 0; i < 164; i++)
	//	musicwin.entries[i] = theRom.musictable[i];

	//musicwin.cur_entry = 

	//int nResult = musicwin.DoModal();

	//if(nResult == IDOK)
	//	for(i = 0; i < 164; i++)
	//		theRom.musictable[i] = musicwin.entries[i];
}

void CMainFrame::OnUpdateToolsMusicEditor(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(theRom.is_loaded);
}



void CMainFrame::OnToolsPaletteeditor() 
{
	// TODO: Add your command handler code here
	CEditorPaletteDlg dlg;
	dlg.Rom = &theRom;
	if(sel_sector)
	{
		dlg.tileset = theRom.global_sets[sectorlist[0].y][sectorlist[0].x];
		dlg.palette = theRom.global_pal[sectorlist[0].y][sectorlist[0].x];
	}
	dlg.DoModal();
}

void CMainFrame::OnUpdateToolsPaletteeditor(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(theRom.is_loaded);
}














/////////////////////////////////////////////////////////////////////////////
// CMapEditView

IMPLEMENT_DYNCREATE(CMapEditView, CScrollView)

BEGIN_MESSAGE_MAP(CMapEditView, CScrollView)
	//{{AFX_MSG_MAP(CMapEditView)
	ON_WM_KEYDOWN()
	ON_WM_VSCROLL()
	ON_WM_HSCROLL()
	ON_WM_MOUSEWHEEL()
	ON_WM_LBUTTONDOWN()
	ON_WM_RBUTTONDOWN()
	ON_WM_SETCURSOR()
	ON_WM_MOUSEMOVE()
	ON_WM_LBUTTONUP()
	ON_WM_LBUTTONDBLCLK()
	ON_WM_RBUTTONDBLCLK()
	ON_WM_KEYUP()
	ON_WM_CREATE()
	ON_WM_ERASEBKGND()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CMapEditView construction/destruction

CMapEditView::CMapEditView()
{
	// TODO: add construction code here

}

CMapEditView::~CMapEditView()
{
}

BOOL CMapEditView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CScrollView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CMapEditView drawing

void CMapEditView::OnDraw(CDC* pDC)
{
	CRect rect;
	GetClientRect(rect);

	if(!theRom.is_loaded)
	{
		pDC->SetViewportOrg(0,0);
		pDC->FillSolidRect(rect, 0);//GetSysColor(COLOR_APPWORKSPACE));
		return;
	}



	int nx = (int)(rect.Width()/(32*zoomlevel))+1;
	int ny = (int)(rect.Height()/(32*zoomlevel))+1;

	if(nx != draw.cx || ny != draw.cy)
		draw.render = RENDER_MAP;

	draw.cx = nx;
	draw.cy = ny;

	// draw the actual tilemap
	if(draw.render == RENDER_MAP)
	{
		int cx = 0;		// screen-relative offset for drawing
		int cy = 0;

		int w = nx;		// width/height of section to be rendered
		int h = ny;

		// find the part that needs re-rendering


		// cx should be close to the right edge when scrolling right,
		// and zero when scrolling left

		int hoff = xOffset - buf_x;
		int voff = yOffset - buf_y;

		// If we've scrolled a lot, make sure no drawing of areas
		// that are now offscreen occurs.
		if(abs(hoff) > nx) hoff = (int)_copysign(nx, hoff);
		if(abs(voff) > ny) voff = (int)_copysign(ny, voff);

		// "width" might be negative here

		if(hoff > 0)
		{
			// if positive w, scrolling right
			cx = nx - hoff;
			w = hoff;
		}
		else if(hoff < 0)
		{
			// otherwise, scrolling left
			cx = 0;
			w = abs(hoff);
		}


		if(voff > 0)
		{
			// if positive w, scrolling down
			cy = ny - voff;
			h = voff;
		}
		else if(voff < 0)
		{
			// otherwise, scrolling up
			cy = 0;
			h = abs(voff);
		}

		// shift old buffer
		mapbuf.BitBlt(0,0,nx*32,ny*32,&mapbuf,hoff*32,voff*32,SRCCOPY);


		if(edit_mode == MODE_MAP)
		{
			if(hoff&&voff)
			{
				theRom.DrawMain(&mapbuf, 0, 0, nx, ny, xOffset, yOffset, sel_sector, sectorlist);
			}
			else
				theRom.DrawMain(&mapbuf, cx, cy, w, h, xOffset+cx, yOffset+cy, sel_sector, sectorlist);
		}
		else
		{
			if(hoff&&voff)
			{
				theRom.DrawMain(&mapbuf, 0, 0, nx, ny, xOffset, yOffset);
			}
			else
				theRom.DrawMain(&mapbuf, cx, cy, w, h, xOffset+cx, yOffset+cy);

		}
		//theRom.DrawMain(&mapbuf, cx, cy, w, h, xOffset+cx, yOffset+cy);
		//theRom.DrawMain(&mapbuf, 0, 0, nx, ny, xOffset, yOffset);


		buf_x = xOffset;
		buf_y = yOffset;

		needs_map_draw = FALSE;
		draw.render = RENDER_NONE;
	}
	else if(draw.render == RENDER_TILE)
	{
		int x = draw.tx - xOffset;
		int y = draw.ty - yOffset;

		if(x < nx && y < ny)
		{
			theRom.DrawTile(&mapbuf, x, y, draw.tx, draw.ty);
		}
	}

	draw.render = 0;


	// draw the map from the map buffer to the other buffer
	buffer.BitBlt(0, 0, nx*32, ny*32, &mapbuf, 0, 0, SRCCOPY);


	// GRIDLINES
	for(int y = 0; y < ny; y++) {
		for(int x = 0; x < nx; x++) {
			if(x+xOffset >= 0 && y+yOffset >= 0 && x+xOffset < 256 && y+yOffset < 320) {
				if(gridlines) {
					int _w = 32;
					int _x = 0;
					int _y = 0;

					if(edit_mode==MODE_ENEMY)
					{
						// In enemy placement mode, the atomic unit of data is
						// a two-tile square block, so we want to offset our
						// gridlines a bit when necessary
						_w = 64;
						_x = (xOffset%2)*32;
						_y = (yOffset%2)*32;
					}

					buffer.MoveTo(x*_w-_x,y*_w-_y);
					buffer.LineTo(x*_w-_x,y*_w+_w-_y);
					buffer.MoveTo(x*_w-_x,y*_w-_y);
					buffer.LineTo(x*_w+_w-_x,y*_w-_y);
				}
			}
		}
	}

switch(edit_mode)
{

	case MODE_MAP:
	{
		CString str;

		for(int i = 0; i < sel_sector; i++)
		{
			int x = sectorlist[i].x;
			int y = sectorlist[i].y;

			if(selectedinfo)
			{
				str.Format("Palette: %d Music: %d",theRom.global_pal[y][x],theRom.global_music[y][x]);
				buffer.TextOut(x*8*32-(xOffset*32)+5, y*4*32-(yOffset*32)+5,str);
			}

			// only draw border on a side if it borders an unselected sector
			bool l,r,u,d;
			l = r = u = d = false;
			for(int j = 0; j < sel_sector; j++)
			{
				int x2 = sectorlist[j].x;
				int y2 = sectorlist[j].y;
				if(x2 == x)
				{
					if(y2 == y-1)
						u = true;
					if(y2 == y+1)
						d = true;
				}
				else if(y2 == y)
				{
					if(x2 == x-1)
						l = true;
					if(x2 == x+1)
						r = true;	
				}
			}

			int _x = x*8*32-(xOffset*32);
			int _y = y*4*32-(yOffset*32);
			int _w = 8*32+1;
			int _h = 4*32+1;

			if(!u) buffer.Draw3dRect(_x, _y-1, _w, 2, RGB(0,0,255),RGB(0,0,255));
			if(!d) buffer.Draw3dRect(_x, _y+_h-1, _w, 2, RGB(0,0,255),RGB(0,0,255));
			if(!l) buffer.Draw3dRect(_x-1, _y, 2, _h, RGB(0,0,255),RGB(0,0,255));
			if(!r) buffer.Draw3dRect(_x+_w-1, _y, 2, _h, RGB(0,0,255),RGB(0,0,255));
		}
	}
	break;


	// DRAW DOORS
	case MODE_DOOR:
	{
		int areax = GetScrollPos(SB_HORZ) / int(32*zoomlevel) / 8;
		int areay = GetScrollPos(SB_VERT) / int(32*zoomlevel) / 8;

		if(timer_update) {
			timer_update = 0;
			cur_frame++;
			if(cur_frame == num_frames) cur_frame = 0;
		}

		COLORREF clr = anim_clrs[cur_frame];//RGB(0,0,255);


		int sx, sy, drawsel = 0;
		for(int i = areax; i < areax+5; i++) {
			for(int j = areay; j < areay+5; j++) {
				int cur = j*32 + i;

				int row = cur/32;
				int col = cur-row*32;

				if(cur < 1280) {
					for(int k = 0; k < theRom.door_entry[cur].num_doors; k++) {
						int x = col*256+theRom.door_entry[cur][k].x*8-int(GetScrollPos(SB_HORZ)/zoomlevel);
						int y = row*256+theRom.door_entry[cur][k].y*8-int(GetScrollPos(SB_VERT)/zoomlevel);

						int w = 8;
						int h = 8;

						if(drag_area == cur && drag_door == k && is_door_dragging)
						{
							buffer.Draw3dRect(x, y, w, h, clr, clr);
							buffer.InvertRect(CRect(x, y, x+w, y+h));
						}
						else if(sel_area == cur && sel_door == k && is_door_selected)
						{
							sx = x;
							sy = y;
							drawsel = 1;
						}
						else
							buffer.Draw3dRect(x, y, w, h, clr, clr);
					}
				}
			}
		}
		if(drawsel)
		{
			buffer.Draw3dRect(sx-1, sy-1, 8+2, 8+2, clr/RGB(2,2,2), clr/RGB(2,2,2));
			buffer.InvertRect(CRect(sx-1, sy-1, sx+8+1, sy+8+1));
		}

		if(door_mode == DOOR_ADD) {
			int x = int(mouse_x/zoomlevel/8)*8;
			int y = int(mouse_y/zoomlevel/8)*8;

			CPoint point;
			CRect rect;
			GetCursorPos(&point);
			GetWindowRect(rect);

			if(rect.PtInRect(point))
				buffer.Draw3dRect(x, y, 8, 8, clr, clr);
		}

		if(door_mode == DOOR_DESTSEEK) {
			int x = int(mouse_x/zoomlevel/8)*8;
			int y = int(mouse_y/zoomlevel/8)*8;

			buffer.InvertRect(CRect(0, y+3, 1280, y+4));
			buffer.InvertRect(CRect(x+3, 0, x+4, 1280));

			buffer.BitBlt(x-60, y-28, 128, 64, &crosshair, 0, 0, SRCAND);
			buffer.Draw3dRect(x, y, 8, 8, RGB(255,0,0), RGB(255,0,0));
		}
	}
	break;


	// DRAW SPRITES
	case MODE_SPRITE:
	{
		CSize s = GetScrollOffset();
		int areax = s.cx/256;//GetScrollPos(SB_HORZ) / int(32*zoomlevel) / 8;
		int areay = s.cy/256;//GetScrollPos(SB_VERT) / int(32*zoomlevel) / 8;

		if(timer_update)
		{
			timer_update = 0;
			cur_frame++;
			if(cur_frame == num_frames) cur_frame = 0;
		}

		COLORREF clr = anim_clrs[cur_frame];

		int spr_count = 0;

		for(int i = areax; i < areax+(nx/8)+2; i++)
		{
			for(int j = areay; j < areay+(ny/8)+2; j++)
			{

				int cur = j*32 + i;

				int row = cur/32;
				int col = cur-row*32;

				if(cur < 1280)
				{
					for(int k = 0; k < theRom.object_areas[cur].num_sprites; k++)
					{
						int x2 = theRom.object_areas[cur].sprites[k].x;
						int y2 = theRom.object_areas[cur].sprites[k].y;

						int x = col*256+x2-int(GetScrollPos(SB_HORZ)/zoomlevel);
						int y = row*256+y2-int(GetScrollPos(SB_VERT)/zoomlevel);

						int tptnum = theRom.object_areas[cur].sprites[k].tptNum;

						int spr = theRom.tpt[tptnum].sprite;

						int w = theRom.sprites[spr].width*8;
						int h = theRom.sprites[spr].height*8;
						int dir = (theRom.tpt[tptnum].dir*2)|FIRST;

						x -= w/2;
						y -= h-8;


						bool bDrawSprite = false;
						if(is_sprite_dragging && drag_area == cur && drag_sprite == k)
						{
							if(spriteborder || drag_sprite_outline)
							{
								buffer.Draw3dRect(x-1, y-1, w+2, h+2, clr, clr);
								buffer.InvertRect(CRect(x-1, y-1, x+w+1, y+h+1));
								buffer.InvertRect(CRect(x, y, x+w, y+h));
							}

							if(!sprite_outlines_only && !drag_sprite_outline)
								bDrawSprite = true;
						}

						else if(is_sprite_selected && sel_area == cur && sel_sprite == k)
						{
							buffer.Draw3dRect(x-1, y-1, w+2, h+2, clr/RGB(2,2,2), clr/RGB(2,2,2));
							buffer.InvertRect(CRect(x-1, y-1, x+w+1, y+h+1));

							if(!sprite_outlines_only && !drag_sprite_outline)
								bDrawSprite = true;
						}

						else
						{
							if(spriteborder)
								buffer.Draw3dRect(x-1, y-1, w+2, h+2, clr, clr);

							if(!sprite_outlines_only)
								bDrawSprite = true;
						}

						// If the sprite should be drawn, draw it
						if(bDrawSprite)
							theRom.sprites[spr].Draw(&buffer, x, y, dir, cur, k, &theRom);

						disp_spr[spr_count].area = cur;
						disp_spr[spr_count].spr = k;
						disp_spr[spr_count].x = x;
						disp_spr[spr_count].y = y;
						disp_spr[spr_count].w = w;
						disp_spr[spr_count++].h = h;
					}
				}
			}
		}
		n_disp_spr = spr_count;

		if(sprite_mode == SPRITE_ADD)
		{
			int x = int(mouse_x/zoomlevel);
			int y = int(mouse_y/zoomlevel);

			CPoint point;
			CRect rect;
			GetCursorPos(&point);
			GetWindowRect(rect);

			if(rect.PtInRect(point))
				buffer.Draw3dRect(x-9, y-17, 18, 26, clr, clr);
		}
	}
	break;


	// ENEMY PLACEMENT
	case MODE_ENEMY:
	{
		for(int x = 0; x < nx/2; x++)
		{
			for(int y = 0; y < ny/2; y++)
			{
				int group = theRom.enemy_map[y+yOffset/2][x+xOffset/2];

				if(!group) continue;

				CString s;
				s.Format("#%d", group);

				int _x = (xOffset%2)*32;
				int _y = (yOffset%2)*32;

				buffer.SetBkMode(OPAQUE);
				buffer.SetBkColor(0);
				buffer.TextOut(x*64-_x+5, y*64-_y+5, s);
				buffer.SetBkMode(TRANSPARENT);
			}
		}
	}
	break;

}



	//pDC->BitBlt(GetScrollPos(SB_HORZ),GetScrollPos(SB_VERT),nx*32,ny*32,&buffer,0,0,SRCCOPY);
	//pDC->StretchBlt(GetScrollPos(SB_HORZ),GetScrollPos(SB_VERT),int(nx*32*zoomlevel),int(ny*32*zoomlevel),&buffer,0,0,nx*32/2,ny*32/2,SRCCOPY);
	//CPoint pt = pDC->GetViewportOrg();



	pDC->SetViewportOrg(0, 0);

	if(zoomlevel!=1)
	{
		// Crappy temporary handler code
		CDC crap;		// crap
		CBitmap bmp;	// bitmap for crap

		crap.CreateCompatibleDC(pDC);	// Create crap
		bmp.CreateCompatibleBitmap(pDC, nx*32, ny*32);

		CBitmap *pBmpOld = crap.SelectObject(&bmp);

		crap.BitBlt(0,0,nx*32,ny*32,&buffer,0,0,SRCCOPY);
		pDC->StretchBlt( 0,0,int(nx*32*zoomlevel), int(ny*32*zoomlevel), &crap, 0, 0, nx*32, ny*32, SRCCOPY);

		// Release crap
		crap.SelectObject(pBmpOld);
	}
	else
		pDC->BitBlt(0,0,nx*32,ny*32,&buffer,0,0,SRCCOPY);

	//pDC->StretchBlt(0,0,1024,1024,pDC,0,0,512,512,SRCCOPY);

}


void CMapEditView::DrawTile(HDC dc, int cx, int cy, int set, int tile, int palette)
{
/*	BITMAPINFO bmi = { { 40, 528, -132, 1, 32, BI_RGB, 0, 0, 0, 0, 0 } };
	HBITMAP bm;
	int *data;
	PAINTSTRUCT ps;
	//HDC hdc = BeginPaint(hWnd, &ps);
	int xx, yy, x, y, tx, ty;
	bm = CreateDIBSection(dc, &bmi, DIB_RGB_COLORS, (void **)&data, NULL, 0);

	// draw all 16 minitiles in a specific map tile
	for(y = 0; y < 4; y++) for(x = 0; x < 4; x++) {
		int b = arranges[set][tile][y*4+x];//[(xx + 0) * 4 + yy][y * 4 + x];
		//b = x;
		// draw all pixels in a minitile
		for(ty = 0; ty < 8; ty++) for(tx = 0; tx < 8; tx++) {
			// x = minitile coordinate, so multiplied by 8 to draw in correct minitile
			data[33*0 + 8*x + (b & 0x4000 ? 7 - tx : tx) + 528 * 
				(33*0 + 8*y + (b & 0x8000 ? 7 - ty : ty))] =
				mpal[((b & 0x1C00) >> 10) - 2][
				tilecolor(_tilesets[set][b & 0x1FF], tx, ty)];
		}
	}
	SetDIBitsToDevice(dc, cx, cy, 32, 132, 0, 100, 0, 132, data, &bmi, DIB_RGB_COLORS);
	DeleteObject(bm);*/
}


void CMapEditView::OnInitialUpdate()
{
	CScrollView::OnInitialUpdate();

	ResizeScrollbars();
}

/////////////////////////////////////////////////////////////////////////////
// CMapEditView diagnostics

#ifdef _DEBUG
void CMapEditView::AssertValid() const
{
	CScrollView::AssertValid();
}

void CMapEditView::Dump(CDumpContext& dc) const
{
	CScrollView::Dump(dc);
}

CMapEditDoc* CMapEditView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CMapEditDoc)));
	return (CMapEditDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CMapEditView message handlers


int CMapEditView::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CScrollView::OnCreate(lpCreateStruct) == -1)
		return -1;


	// Create some cursors... ?
	m_hcrArrow = LoadCursor(NULL, IDC_ARROW);

	m_hcrDropper = LoadCursor(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDC_POINTER_DROP));
	m_hcrQuestion = LoadCursor(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDC_POINTER_INFO));
	m_hcrAdd = LoadCursor(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDC_POINTER_ADD));
	m_hcrCopy = LoadCursor(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDC_POINTER_COPY));

	m_hcrHand = LoadCursor(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDC_POINTER_HAND));
	m_hcrNohand = LoadCursor(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDC_POINTER_NOHAND));
	m_hcrNono = LoadCursor(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDC_POINTER_NONO));

	
	return 0;
}



void CMapEditView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// TODO: Add your message handler code here and/or call default

	switch(nChar) {
	case 16:	// SHIFT
		input.keys |= input.shift;
		break;
	case 17:	// CTRL
		input.keys |= input.ctrl;
		break;
	case 37:	// LEFT
		KEYSCROLL_L = TRUE;
		break;
	case 38:	// UP
		KEYSCROLL_U = TRUE;
		break;
	case 39:	// RIGHT
		KEYSCROLL_R = TRUE;
		break;
	case 40:	// DOWN
		KEYSCROLL_D = TRUE;
		break;
	}

	if(KEYSCROLL_L) xOffset--;
	if(KEYSCROLL_R) xOffset++;

	if(KEYSCROLL_U) yOffset--;
	if(KEYSCROLL_D) yOffset++;

	ScrollTo(xOffset,yOffset);

	SendMessage(WM_SETCURSOR, 0, (LPARAM)HTCLIENT);

	((CMainFrame*)GetParent())->UpdateStatusBar();

	CScrollView::OnKeyDown(nChar, nRepCnt, nFlags);
}


void CMapEditView::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// TODO: Add your message handler code here and/or call default
	switch(nChar) {
	case 16:	// SHIFT
		input.keys ^= input.shift;
		break;
	case 17:	// CTRL
		input.keys ^= input.ctrl;
		break;
	case 37:	// LEFT
		KEYSCROLL_L = FALSE;
		break;
	case 38:	// UP
		KEYSCROLL_U = FALSE;
		break;
	case 39:	// RIGHT
		KEYSCROLL_R = FALSE;
		break;
	case 40:	// DOWN
		KEYSCROLL_D = FALSE;
		break;
	}

	SendMessage(WM_SETCURSOR, 0, (LPARAM)HTCLIENT);

	((CMainFrame*)GetParent())->UpdateStatusBar();

	CScrollView::OnKeyUp(nChar, nRepCnt, nFlags);
}



void CMapEditView::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	int y = yOffset;

	switch(nSBCode) {
	case SB_THUMBTRACK:
	case SB_THUMBPOSITION:
		y = nPos/int(32*zoomlevel);
		break;

	case SB_LINEUP:
		y--;
		break;
	case SB_LINEDOWN:
		y++;
		break;

	case SB_PAGEUP:
		y-=8;
		break;
	case SB_PAGEDOWN:
		y+=8;
		break;
	}

	// Only do the scrolling if the view has actually changed
	if(y!=yOffset)
		ScrollTo(xOffset, y);

}

void CMapEditView::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	int x = xOffset;

	switch(nSBCode) {
	case SB_THUMBTRACK:
	case SB_THUMBPOSITION:
		x = nPos/int(32*zoomlevel);
		break;

	case SB_LINELEFT:
		x--;
		break;
	case SB_LINERIGHT:
		x++;
		break;

	case SB_PAGELEFT:
		x-=8;
		break;
	case SB_PAGERIGHT:
		x+=8;
		break;
	}

	if(x != xOffset)
		ScrollTo(x, yOffset);

}


BOOL CMapEditView::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt) 
{

	CRect rect;

	CWnd *wnd;	// mode window to forward message to
	switch(edit_mode)
	{
	case MODE_MAP: wnd = &((CMainFrame*)GetParent())->m_wndToolboxBar.m_wndTilePal;
		break;
	case MODE_DOOR: wnd = &((CMainFrame*)GetParent())->m_wndToolboxBar.m_wndDoorBar;
		break;
	case MODE_SPRITE: wnd = &((CMainFrame*)GetParent())->m_wndToolboxBar.m_wndSpriteBar;
		break;
	case MODE_ENEMY: wnd = &((CMainFrame*)GetParent())->m_wndToolboxBar.m_wndEnemyBar;
	}

	if(wnd)
	{
		wnd->GetWindowRect(rect);

		if(rect.PtInRect(pt) && pt.x < rect.left+160)
		{
			wnd->SendMessage(WM_MOUSEWHEEL, WPARAM((short)nFlags | (zDelta<<16)), LPARAM(pt.x | (pt.y << 16)));
			return FALSE;
		}
	}


	// Override cancels scrolling messages
	if(zDelta < 0) yOffset+=3;
	else if(zDelta > 0) yOffset-=3;

	ScrollTo(xOffset, yOffset);



	return TRUE;
}



void CMapEditView::ScrollTo(int x, int y)
{
	CRect rect;
	GetClientRect(rect);

	xOffset = x;
	yOffset = y;

	int xlim = int(256 - rect.Width()/(32*zoomlevel)+1);
	int ylim = int(320 - rect.Height()/(32*zoomlevel)+1);

	if(x > xlim) x = xlim;
	if(x < 0) x = 0;
	if(y > ylim) y = ylim;
	if(y < 0) y = 0;

	dlgBar->SetDlgItemInt(IDC_EDIT_X, x);
	dlgBar->SetDlgItemInt(IDC_EDIT_Y, y);

	needs_map_draw = TRUE;
	draw.render = RENDER_MAP;

	Invalidate(FALSE);

	SetScrollPos(SB_VERT,int(y*32*zoomlevel),TRUE);
	SetScrollPos(SB_HORZ,int(x*32*zoomlevel),TRUE);
}




void CMapEditView::PlaceTile(int x, int y, int tile)
{
	place_set = tile/256;
	place_tile = tile%256;

	if(theRom.local_sets[y][x] != place_set ||
	   theRom.map_tiles[y][x] != place_tile)
	{

		AddUndoState(x,y,tile);

		theRom.local_sets[y][x] = place_set;
		theRom.map_tiles[y][x] = place_tile;
	}

	//needs_map_draw = TRUE;
	draw.render = RENDER_TILE;
	draw.tx = x;
	draw.ty = y;

	//int mx = (x-xOffset)*32;
	//int my = (y-yOffset)*32;
	//CRect rect(mx, my, mx+32, my+32);

	Invalidate(FALSE);
}



void CMapEditView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	int m_x = int(point.x/zoomlevel);
	int m_y = int(point.y/zoomlevel);

	mouse.b |= 1;	// Set lbutton flag

	SetCapture();	// Set mouse capture

	if(!theRom.is_loaded) return;


	if(tool_mode != TOOL_NONE)
	{
		// Do toolmode lbuttondown processing
		switch(tool_mode)
		{
		case TOOL_MAGNIFY:
			// A left-click in magnify mode increases the mag level
			zoomlevel+=.25;
			Invalidate(FALSE);
			break;
		}
	}

	else switch(edit_mode)
	{
	case MODE_MAP:
	{
		CPoint pt = GetMousePos();

		int xTile = pt.x;
		int yTile = pt.y;

		if(infomode)
		{
			infomode = FALSE;
			tilemode = -1;
		}

		switch(tilemode)
		{
		case TILE_PLACE:
			// Place the current tile
			PlaceTile(xTile, yTile, m_pToolbox->GetSelTile());

			break;
		case TILE_PICK:
			place_set = theRom.local_sets[yTile][xTile];
			place_tile = theRom.map_tiles[yTile][xTile];

			m_pToolbox->SetGlobalPal(theRom.global_pal[yTile/4][xTile/8]);
			m_pToolbox->SetSelTile(theRom.global_sets[yTile/4][xTile/8], place_tile+place_set*256);
			m_pToolbox->m_wndTilePal.SendMessage(WM_PAINT);
			break;
		case TILE_INFO:
			infomode = TRUE;
			infoX = xTile;
			infoY = yTile;
			ShowInfo();
			break;
		case TILE_SELECT:
			break;
		default:
			tilemode = TILE_PLACE;
			break;
		}
	}
	break;


	case MODE_DOOR:
	{
		CPoint pt = GetMousePos();		// Mouse position (in 8x8 units)

		if(door_mode == DOOR_NORMAL)
		{
			if(nFlags & MK_CONTROL)
			{

				DoorEntry* d = theRom.GetDoorAt(pt.x, pt.y);

				if(d != NULL && d->flags == DEST_DOOR)			
					ScrollToDestination(theRom.dest.entries[d->destindex]);

				return;
			}


			drag_area = int(pt.y/32)*32 + pt.x/32;
			drag_door = -1;

			int x = pt.x%32;
			int y = pt.y%32;

			for(int i = 0; i < theRom.door_entry[drag_area].num_doors; i++) {
				if(theRom.door_entry[drag_area][i].x == x && theRom.door_entry[drag_area][i].y == y) {
					drag_door = i;
					is_door_dragging = TRUE;
				}
			}

			if(is_door_dragging)
			{
				sel_area = drag_area;
				sel_door = drag_door;
				is_door_selected = TRUE;
			}
			else
				is_door_selected = FALSE;
		}

		
	}
	break;

	case MODE_SPRITE:
	{
		// Note:
		// Sprite/mouse collision cannot perform normal searching of the
		// area in which the mouse is located, as sprites can often extend
		// the boundaries of their area a little. Instead, this uses a list
		// of information about displayed sprites to locate the correct
		// object when the user clicks a sprite.

		int x = m_x;
		int y = m_y;

		// Search displayed sprite information for a coordinate match
		bool bFound = false;
		for(int i = 0; i < n_disp_spr; i++) {
			int sx = disp_spr[i].x;
			int sy = disp_spr[i].y;
			int sw = disp_spr[i].w;
			int sh = disp_spr[i].h;

			if(x >= sx && x <= sx+sw && y >= sy && y <= sy+sh) {
				drag_area = disp_spr[i].area;
				drag_sprite = disp_spr[i].spr;

				spr_offx = m_x - sx-(sw/2);
				spr_offy = m_y - sy-(sh-8);

				bFound = TRUE;
			}
		}


		if(bFound)
		{
			CString str;
			str.Format("a:%d s:%d (%d, %d)", drag_area, drag_sprite, spr_offx, spr_offy);
			((CMainFrame*)GetParent())->m_wndStatusBar.SetPaneText(2, str);

			sel_area = drag_area;
			sel_sprite = drag_sprite;
			is_sprite_selected = TRUE;
			is_sprite_dragging = TRUE;
		}
		else
		{
			is_sprite_selected = FALSE;
			is_sprite_dragging = FALSE;
		}
	}
	break;

	case MODE_ENEMY:
	{
		CPoint pt = GetMousePos();

		int x = pt.x;//((m_x/32) + xOffset)/2;
		int y = pt.y;//((m_y/32) + yOffset)/2;
		theRom.enemy_map[y][x] = ((CMainFrame*)GetParent())->m_wndToolboxBar.GetSelectedEnemyCluster();
	}
	break;
	}

	Invalidate(FALSE);

	CScrollView::OnLButtonDown(nFlags, point);
}



void CMapEditView::OnMouseMove(UINT nFlags, CPoint point) 
{
	mouse.x = point.x;
	mouse.y = point.y;

	int m_x = int(point.x/zoomlevel);
	int m_y = int(point.y/zoomlevel);

	if(!theRom.is_loaded) return;

	// Get mouse position
	CPoint pt = GetMousePos();

	switch(edit_mode)
	{
		case MODE_MAP:
		{
			if(mouse.b & 1)
			{
				int xTile = pt.x;//m_x/32 + xOffset;
				int yTile = pt.y;//m_y/32 + yOffset;

				if(tilemode == TILE_PLACE)
				{
					// Place selected tile on map
					PlaceTile(xTile, yTile, m_pToolbox->GetSelTile());
				}
			}
		}
		break;


		case MODE_DOOR:
		{
			// If a door is being dragged around
			if(is_door_dragging) {

				int ar_y = drag_area / 32;
				int ar_x = drag_area - (ar_y*32);

				int ax = pt.x;			// absolute coordinates
				int ay = pt.y;

				int x = ax-(ar_x*32);		// relative coordinates
				int y = ay-(ar_y*32);

				bool move_ok = true;

				ar_x = ax/32;
				ar_y = ay/32;

				int new_area = ar_y*32 + ar_x;

				for(int i = 0; i < theRom.door_entry[new_area].num_doors; i++) {
					if(theRom.door_entry[new_area][i].x == ax%32 && theRom.door_entry[new_area][i].y == ay%32)
						move_ok = false;
				}

				if(move_ok) {
					theRom.door_entry[drag_area][drag_door].x = x;
					theRom.door_entry[drag_area][drag_door].y = y;
				}

				Invalidate(FALSE);
			}

			if(door_mode == DOOR_ADD || door_mode == DOOR_DESTSEEK) {
				if(point.y/8 != mouse_y/8 || point.x/8 != mouse_x/8)
					Invalidate(FALSE);

				CString str;
				str.Format("%d, %d", pt.x, pt.y);

				((CMainFrame*)GetParent())->m_wndStatusBar.SetPaneText(2, str);
			}
		}
		break;

		case MODE_SPRITE:
		{

			// If a sprite is being dragged around
			if(is_sprite_dragging)
			{
				int ar_y = drag_area / 32;
				int ar_x = drag_area - (ar_y*32);

				int ax = pt.x;		// absolute coordinates
				int ay = pt.y;

				int x = ax-ar_x*256;									// relative coordinates
				int y = ay-ar_y*256;


				theRom.object_areas[drag_area].sprites[drag_sprite].x = x - spr_offx;
				theRom.object_areas[drag_area].sprites[drag_sprite].y = y - spr_offy;

				
				if(zoomlevel == 1)
					InvalidateRect(CRect(point.x-70, point.y-70, point.x+70, point.y+70), FALSE);
				else
					Invalidate(FALSE);

				CString str;
				str.Format("a:%d s:%d (%d, %d)", drag_area, drag_sprite, x - spr_offx, y - spr_offy);
				((CMainFrame*)GetParent())->m_wndStatusBar.SetPaneText(2, str);

			}
			if(sprite_mode == SPRITE_ADD)
			{
				InvalidateRect(CRect(point.x-80, point.y-80, point.x+80, point.y+80), FALSE);

				CString str;
				str.Format("%d, %d", pt.x, pt.y);

				((CMainFrame*)GetParent())->m_wndStatusBar.SetPaneText(2, str);
			}
			else
			{
				if(((CMainFrame*)GetParent())->m_wndToolboxBar.IsDraggingSprite(TRUE) && nFlags&MK_LBUTTON)
				{
					sprite_mode = SPRITE_ADD;
					sprite_place = ((CMainFrame*)GetParent())->m_wndToolboxBar.GetSelectedSprite();
					Invalidate(FALSE);
				}
			}
		}
		break;

		case MODE_ENEMY:
		{
			if(nFlags&MK_LBUTTON)
			{
				theRom.enemy_map[pt.y][pt.x] = ((CMainFrame*)GetParent())->m_wndToolboxBar.GetSelectedEnemyCluster();
				Invalidate(FALSE);
			}
			else if(nFlags&MK_RBUTTON)
			{
				theRom.enemy_map[pt.y][pt.x] = 0;
				Invalidate(FALSE);
			}
		}
		break;
	}

	mouse_x = point.x;
	mouse_y = point.y;
	
	CScrollView::OnMouseMove(nFlags, point);
}


void CMapEditView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	int m_x = int(point.x/zoomlevel);
	int m_y = int(point.y/zoomlevel);

	mouse.b ^= mouse.b&1;	// Unset lbutton flag

	ReleaseCapture();

	CPoint pt = GetMousePos();	// Get relative mouse position

	// TODO: Add your message handler code here and/or call default
	switch(edit_mode)
	{
		case MODE_DOOR:
		{
			if(door_mode == DOOR_NORMAL)
			{
				if(is_door_dragging)
				{
					// move some stuff around
					int ar_y = drag_area / 32;
					int ar_x = drag_area - (ar_y*32);

					int ax = pt.x;		// absolute coordinates
					int ay = pt.y;

					int x = ax-ar_x*32;									// relative coordinates
					int y = ay-ar_y*32;

					int dest_area = (ay/32)*32 + (ax/32);

					// compare the two areas; if they're the same, we don't need to do anything
					if(dest_area != drag_area)
					{
						// first, add a new door (yay :D)
						DoorEntry a = theRom.door_entry[drag_area][drag_door];

						// mod the door's coordinates, as they go way out of range
						// when the door is being dragged around
						a.x = ax%32;
						a.y = ay%32;

						// add door to destination area
						theRom.door_entry[dest_area].AddDoor(a);

						// remove old door from drag source area
						theRom.door_entry[drag_area].RemDoor(drag_door);

						// if the door has moved, we also need to redo the selection
						sel_area = dest_area;
						sel_door = theRom.door_entry[dest_area].num_doors - 1;
						is_door_selected = TRUE;
					}

					is_door_dragging = FALSE;
					Invalidate(FALSE);
				}
			}

			if(door_mode == DOOR_DESTSEEK)
			{
				int x = pt.x;
				int y = pt.y;


				((CMainFrame*)GetParent())->m_wndStatusBar.SetWindowText("Ready");
				((CMainFrame*)GetParent())->m_wndStatusBar.SetPaneText(2, "");


				((CMainFrame*)GetParent())->SeekProc(false, NULL, TRUE, -1, x, y);

				Invalidate(FALSE);
			}

			if(door_mode == DOOR_ADD)
			{
				int ax = pt.x;
				int ay = pt.y;
				int cur_area = (ay/32)*32 + (ax/32);

				int x = pt.x % 32;
				int y = pt.y % 32;

				DoorEntry door;
				door.address = 0;
				door.x = x;
				door.y = y;
				door.destination = 0;
				door.destindex = 0;
				door.direction = 0;
				door.is_esc_end = false;
				door.flags = DEST_DOOR;
				door.ropetype = 0;

				theRom.door_entry[cur_area].AddDoor(door);


				door_mode = DOOR_NORMAL;
				((CMainFrame*)GetParent())->m_wndStatusBar.SetWindowText("Ready");
				((CMainFrame*)GetParent())->m_wndStatusBar.SetPaneText(2, "");
				Invalidate(FALSE);
			}
		}
		break;

		case MODE_SPRITE:
		{
			if(is_sprite_dragging) {
				int ar_y = drag_area / 32;
				int ar_x = drag_area - (ar_y*32);

				int ax = pt.x;		// absolute coordinates
				int ay = pt.y;

				int x = ax-ar_x*256;									// relative coordinates
				int y = ay-ar_y*256;

				int dest_area = (ay/256)*32 + (ax/256);

				if(dest_area != drag_area)
				{
					int sx = (ax%256) - spr_offx;
					int sy = (ay%256) - spr_offy;
					int stptNum = theRom.object_areas[drag_area].sprites[drag_sprite].tptNum;
					theRom.AddSprite(dest_area, sx, sy, stptNum);
					theRom.RemSprite(drag_area, drag_sprite);

					sel_area = dest_area;
					sel_sprite = theRom.object_areas[dest_area].num_sprites - 1;
					is_sprite_selected = TRUE;
				}

				is_sprite_dragging = FALSE;
				Invalidate(FALSE);
			}

			if(sprite_mode == SPRITE_ADD)
			{
				int ax = pt.x;		// absolute coordinates
				int ay = pt.y;


				int cur_area = (ay/256)*32 + (ax/256);


				int x = pt.x%256;
				int y = pt.y%256;

				theRom.AddSprite(cur_area, x, y, sprite_place);

				sprite_mode = SPRITE_NORMAL;
				((CMainFrame*)GetParent())->m_wndStatusBar.SetWindowText("Ready");
				((CMainFrame*)GetParent())->m_wndStatusBar.SetPaneText(2, "");
				Invalidate(FALSE);
			}
		}
		break;
	}


	CScrollView::OnLButtonUp(nFlags, point);
}

void CMapEditView::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	int m_x = int(point.x/zoomlevel);
	int m_y = int(point.y/zoomlevel);

	CPoint pt = GetMousePos();

	if(edit_mode == MODE_DOOR) {
		int x = pt.x;
		int y = pt.y;

		int area, door;

		theRom.GetDoorAt(x,y,area,door);

		if(area != -1 && door != -1)
			((CMainFrame*)GetParent())->OpenDoorDlg(area, door);
	}


	if(edit_mode == MODE_SPRITE)
	{
		int x = m_x; //((point.x + GetScrollPos(SB_HORZ))%256);
		int y = m_y; //((point.y + GetScrollPos(SB_VERT))%256);

		// Search the displayed sprites for a match at the current coordinates
		for(int i = 0; i < n_disp_spr; i++) {
			int sx = disp_spr[i].x;
			int sy = disp_spr[i].y;
			int sw = disp_spr[i].w;
			int sh = disp_spr[i].h;

			if(x >= sx && x <= sx+sw && y >= sy && y <= sy+sh) {
				int area = disp_spr[i].area;
				int spr = disp_spr[i].spr;

				CSpriteDlg dlg(area, spr);
				dlg.Rom = &theRom;
				dlg.DoModal();

				Invalidate(FALSE);
			}
		}
	}

	CScrollView::OnLButtonDblClk(nFlags, point);
}



void CMapEditView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	int m_x = int(point.x/zoomlevel);
	int m_y = int(point.y/zoomlevel);

	CPoint pt = GetMousePos();

	if(!theRom.is_loaded) return;

	switch(edit_mode)
	{
		case MODE_MAP:
		{
			if(tilemode == TILE_PICK)	// shift key is up
				SelectSector(pt.x/8, pt.y/4);

			else if(tilemode == TILE_INFO)
			{
				sel_sector = 0;
				SelectAdjacentSectors(pt.x/8, pt.y/4);
			}
			else
				SelectSector(pt.x/8, pt.y/4, TRUE);

			needs_map_draw = TRUE;
			draw.render = RENDER_MAP;
			Invalidate(FALSE);
		}
		break;

		case MODE_DOOR:
		{
			if(door_mode == DOOR_DESTSEEK)
			{
				((CMainFrame*)GetParent())->m_wndStatusBar.SetWindowText("Ready");
				SendMessage(WM_SETCURSOR);

				((CMainFrame*)GetParent())->SeekProc(FALSE, NULL, FALSE, -1, -1, -1);
			}
			else if(door_mode == DOOR_ADD)
			{
				door_mode = DOOR_NORMAL;
				((CMainFrame*)GetParent())->m_wndStatusBar.SetWindowText("Ready");
			}
		}
		break;

		case MODE_SPRITE:
		{
			if(sprite_mode == SPRITE_NORMAL)
			{
				for(int i = 0; i < n_disp_spr; i++)
				{
					int sx = disp_spr[i].x;
					int sy = disp_spr[i].y;
					int sw = disp_spr[i].w;
					int sh = disp_spr[i].h;

					if(m_x >= sx && m_x <= sx+sw &&m_y >= sy && m_y <= sy+sh)
					{
						CMenu menu;
						CMenu *popup;
						CPoint pt = point;

						ClientToScreen(&pt);

						menu.LoadMenu(IDR_SPRITEPOPUP);
						popup = menu.GetSubMenu(0);

						sel_area = disp_spr[i].area;
						sel_sprite = disp_spr[i].spr;
						is_sprite_selected = TRUE;

						Invalidate(FALSE);

						TrackPopupMenu((HMENU)*popup, TPM_LEFTALIGN | TPM_RIGHTBUTTON, pt.x, pt.y, 0, (HWND)*this, NULL);

						break;
					}
				}
			}
			else if(sprite_mode == SPRITE_ADD)
			{
				((CMainFrame*)GetParent())->m_wndStatusBar.SetWindowText("Ready");
				SendMessage(WM_SETCURSOR);

				sprite_mode = SPRITE_NORMAL;
			}
		}
		break;

		case MODE_ENEMY:
		{
			theRom.enemy_map[pt.y][pt.x] = 0;
			Invalidate(FALSE);
		}

	}


	CScrollView::OnRButtonDown(nFlags, point);
}



void CMapEditView::OnRButtonDblClk(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	int m_x = int(point.x/zoomlevel);
	int m_y = int(point.y/zoomlevel);

	CPoint pt = GetMousePos();

	if(theRom.is_loaded) {
		if(edit_mode == MODE_MAP) {
			sel_sector = 0;
			SelectAdjacentSectors(pt.x/8, pt.y/4);
		}
	}
	CScrollView::OnRButtonDblClk(nFlags, point);
}



BOOL CMapEditView::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message) 
{
	if(nHitTest == HTHSCROLL || nHitTest == HTVSCROLL)
		return CScrollView::OnSetCursor(pWnd, nHitTest, message);

	switch(edit_mode)
	{
	case MODE_MAP:
	{
		if(nHitTest != HTCLIENT)
			return CScrollView::OnSetCursor(pWnd, nHitTest, message);

		switch(tilemode) {
		case TILE_PLACE:
			SetCursor(m_hcrArrow);
			break;
		case TILE_PICK:
			SetCursor(m_hcrDropper);
			break;
		case TILE_INFO:
			SetCursor(m_hcrQuestion);
			break;
		}
		break;
	}

	case MODE_DOOR:
	{
		if(nHitTest != HTCLIENT)
			return CScrollView::OnSetCursor(pWnd, nHitTest, message);

		
		switch(door_mode) {
		case DOOR_NORMAL:
			if(input.keys & input.ctrl)
			{
				CPoint pt = GetMousePos();
				DoorEntry *d = theRom.GetDoorAt(pt.x, pt.y);

				if(d != NULL)
				{
					if(d->flags == DEST_DOOR)
						SetCursor(m_hcrHand);
					else
						SetCursor(m_hcrNohand);
				}
				else
					SetCursor(m_hcrNono);
			}
			else
				SetCursor(m_hcrArrow);
			break;
		case DOOR_ADD:
			SetCursor(m_hcrCopy);
			break;
		case DOOR_DESTSEEK:
			SetCursor(0);
			break;
		}
	}
	break;

	case MODE_SPRITE:
	{
		switch(sprite_mode)
		{
		case SPRITE_NORMAL:
			SetCursor(LoadCursor(NULL, IDC_ARROW));
			break;
		case SPRITE_ADD:
			SetCursor(LoadCursor(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDC_POINTER_ADD)));
			break;
		}
	}
	break;


	default:
		return CScrollView::OnSetCursor(pWnd, nHitTest, message);
	}

	return FALSE;
}





void CMapEditView::ShowInfo()
{
	CInfoDlg dlg;
	dlg.DoModal();
}

void CMapEditView::SelectSector(int x, int y, bool newSelection, bool redraw)
{
	//sector_x = x;
	//sector_y = y;

	bool isOkay = true;

	for(int i = 0; i < sel_sector; i++)
	{
		// the selected sector is already in the selection
		if(sectorlist[i].x == x && sectorlist[i].y == y)
		{
			// remove it
			sel_sector--;
			for(int j = i; j < sel_sector; j++)
				sectorlist[j] = sectorlist[j+1];

			Invalidate(FALSE);

			if(!newSelection) return;
			else if(sel_sector < 1) isOkay = false;
		}
	}

	if(newSelection)
		sel_sector = 0;


	if(isOkay)
	{
		sectorlist[sel_sector].x = x;
		sectorlist[sel_sector].y = y;
		sel_sector++;
	}

	if(redraw)
	{
		Invalidate(FALSE);
	}

	if(sel_sector == 1)
	{
		CComboBox* CBTileset = (CComboBox*)dlgBar->GetDlgItem(IDC_COMBO_TILESET);

		CBTileset->SetCurSel(theRom.global_sets[y][x]);

		dlgBar->SetDlgItemInt(IDC_EDIT_X2, x);
		dlgBar->SetDlgItemInt(IDC_EDIT_Y2, y);
		dlgBar->SetDlgItemInt(IDC_EDIT_PALETTE, theRom.global_pal[y][x]);
		dlgBar->SetDlgItemInt(IDC_EDIT_MUSIC, theRom.global_music[y][x]);

		m_pToolbox->SetGlobalSet(theRom.global_sets[y][x]);
		m_pToolbox->SetGlobalPal(theRom.global_pal[y][x]);

		/*((CMainFrame*)GetParent())->m_wndToolboxBar.gSet = theRom.global_sets[y][x];
		((CMainFrame*)GetParent())->m_wndToolboxBar.gPal = theRom.global_pal[y][x];
		((CMainFrame*)GetParent())->m_wndToolboxBar.Invalidate(FALSE);*/

		clipboard.isUseable = TRUE;

		
		if(muzak)
		{
			// play music =D
			int music = theRom.global_music[y][x];
			int item = theRom.musictable[music].num_events-1;
			int track = theRom.musictable[music].events[item].music;

			if(track != theRom.music_playing)
				theRom.PlayMusic(track);
		}
	}
	else if(sel_sector == 0)
	{
		theRom.StopMusic();
	}

	if(newSelection)
		m_pToolbox->SendMessage(TM_SELSECTOR, (WPARAM)sel_sector, (LPARAM)sectorlist);
}


void CMapEditView::SelectAdjacentSectors(int x, int y)
{
	int set = theRom.global_sets[y][x];	// set to match against

	CPoint pt[1000];	// point queue

	int len = 0;		// length of queue
	int cur = 0;		// current point being checked

	// add first point to queue
	pt[len++] = CPoint(x,y);

	do
	{
		if(theRom.global_sets[pt[cur].y][pt[cur].x] == set)
		{
			CPoint t[4] = {
				CPoint(pt[cur].x+1, pt[cur].y),
				CPoint(pt[cur].x-1, pt[cur].y),
				CPoint(pt[cur].x, pt[cur].y+1),
				CPoint(pt[cur].x, pt[cur].y-1)
			};

			for(int i = 0; i < 4; i++)
			{
				BOOL okay = TRUE;
				for(int j = 0; j < len; j++)
				{
					if(t[i] == pt[j])
						okay = FALSE;
				}

				if(t[i].x < 0 || t[i].x > 255 || t[i].y < 0 || t[i].y > 319)
					okay = FALSE;

				if(okay)
					pt[len++] = t[i];
			}

			SelectSector(pt[cur].x, pt[cur].y, FALSE, FALSE);
		}cur++;
	}while(cur != len); // when current is last, quit

	needs_map_draw = TRUE;
	draw.render = RENDER_MAP;
	Invalidate(FALSE);
}

BOOL CMapEditView::CheckPt(CPoint pt, int set, CPoint pt_list[], int& len)
{
	if(theRom.global_sets[pt.y][pt.x] == set)
	{
		// Add surrounding points
		len = AddPt(CPoint(pt.x+1, pt.y), pt_list, len);
		len = AddPt(CPoint(pt.x-1, pt.y), pt_list, len);
		len = AddPt(CPoint(pt.x, pt.y+1), pt_list, len);
		len = AddPt(CPoint(pt.x, pt.y-1), pt_list, len);
		return TRUE;
	}
	return FALSE;
}

int CMapEditView::AddPt(CPoint pt, CPoint pt_list[], int len)
{
	BOOL okay = TRUE;
	for(int i = 0; i < len; i++)
	{
		if(pt == pt_list[i])
			okay = FALSE;
	}

	if(pt.x < 0 || pt.x > 255 || pt.y < 0 || pt.y > 319)
		okay = FALSE;

	if(okay)
		pt_list[len++] = pt;

	//return val: new length of queue
	return len;
}


/*int CMapEditView::AddCheckPt(int x, int y, int cur, CPoint pt[])
{
	BOOL okay = TRUE;
	for(int i = 0; i < 10000; i++)
	{
		if(pt[i].x == x && pt[i].y == y)
			okay = FALSE;
	}

	if(okay)
	{
		pt[cur] = CPoint(x,y);
		cur++;
	}
	return cur;
}*/



// CheckPt:
//  Checks a given point; if it matches the global set, the surrounding
//  points are added to the queue.

// AddPt:
//  Adds a given point to the queue for checking, but only if it is not
//  already in the queue.

// Loop:
//  Manages queue; calls CheckPt on each point the in queue (after each
//  call, there might be additional points, so it keeps going until there
//  aren't any more points). If a point checks out, it adds the point to
//  the selection list.







void CMainFrame::OnSelectTileset()
{
	CComboBox* CBTileset = (CComboBox*)m_wndDlgBar.GetDlgItem(IDC_COMBO_TILESET);

	int set = CBTileset->GetCurSel();

	for(int i = 0; i < sel_sector; i++)
		theRom.global_sets[sectorlist[i].y][sectorlist[i].x] = set;

	needs_map_draw = TRUE;
	draw.render = RENDER_MAP;
	GetActiveView()->Invalidate(FALSE);
	m_wndToolboxBar.SetGlobalSet(set);
}

void CMainFrame::OnEditX()
{
	if(auto_update) {
		auto_update = FALSE;
		xOffset = dlgBar->GetDlgItemInt(IDC_EDIT_X);

		needs_map_draw = TRUE;
		draw.render = RENDER_MAP;

		GetActiveView()->Invalidate(FALSE);
		GetActiveView()->SetScrollPos(SB_HORZ,xOffset*32);

		auto_update = TRUE;
	}
}

void CMainFrame::OnEditY()
{
	if(auto_update) {
		auto_update = FALSE;
		yOffset = dlgBar->GetDlgItemInt(IDC_EDIT_Y);

		needs_map_draw = TRUE;
		draw.render = RENDER_MAP;

		GetActiveView()->Invalidate(FALSE);
		GetActiveView()->SetScrollPos(SB_VERT,yOffset*32);

		auto_update = TRUE;
	}
}

void CMainFrame::OnEditPalette()
{
	int pal = dlgBar->GetDlgItemInt(IDC_EDIT_PALETTE);
	if(pal <= 7)
	{
		for(int i = 0; i < sel_sector; i++)
			theRom.global_pal[sectorlist[i].y][sectorlist[i].x] = dlgBar->GetDlgItemInt(IDC_EDIT_PALETTE);

		needs_map_draw = TRUE;
		draw.render = RENDER_MAP;

		GetActiveView()->Invalidate(FALSE);

		m_wndToolboxBar.SetGlobalPal(pal);
	}
}


void CMainFrame::OnEditMusic()
{
	for(int i = 0; i < sel_sector; i++)
		theRom.global_music[sectorlist[i].y][sectorlist[i].x] = dlgBar->GetDlgItemInt(IDC_EDIT_MUSIC);

	GetActiveView()->Invalidate(FALSE);
}















void CMainFrame::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if(nChar == 16 && (tilemode != TILE_PICK)) {
		tilemode = TILE_PICK;
		GetActiveView()->SendMessage(WM_SETCURSOR, 0, (LPARAM)HTCLIENT);
	}
	if(nChar == 17 && (tilemode != TILE_INFO)) {
		tilemode = TILE_INFO;
		GetActiveView()->SendMessage(WM_SETCURSOR, 0, (LPARAM)HTCLIENT);
	}

	CFrameWnd::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CMainFrame::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if(nChar == 16 || nChar == 17) {
		tilemode = TILE_PLACE;
		GetActiveView()->SendMessage(WM_SETCURSOR, 0, (LPARAM)HTCLIENT);
	}

	CFrameWnd::OnKeyUp(nChar, nRepCnt, nFlags);
}

BOOL CMapEditApp::PreTranslateMessage(MSG* pMsg) 
{
	// Send all key messages to our frame window
	if(pMsg->message == WM_KEYDOWN || pMsg->message == WM_KEYUP || pMsg->message == WM_CHAR) {
		m_pMainWnd->SendMessage(pMsg->message, pMsg->wParam, pMsg->lParam);
	}

	return CWinApp::PreTranslateMessage(pMsg);
}





/////////////////////////////////////////////////////////////////////////////
// CInfoDlg dialog


CInfoDlg::CInfoDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CInfoDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CInfoDlg)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
}


void CInfoDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CInfoDlg)
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CInfoDlg, CDialog)
	//{{AFX_MSG_MAP(CInfoDlg)
	ON_WM_LBUTTONDOWN()
	ON_WM_PAINT()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CInfoDlg message handlers

BOOL CInfoDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	SetCapture();

	SetDlgItemInt(IDC_TILE, theRom.map_tiles[infoY][infoX]);
	SetDlgItemInt(IDC_LSET, theRom.local_sets[infoY][infoX]);
	SetDlgItemInt(IDC_GSET, theRom.global_sets[infoY/4][infoX/8]);
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CInfoDlg::OnLButtonDown(UINT nFlags, CPoint point) 
{
	ReleaseCapture();
	infomode = FALSE;
	OnCancel();
	CDialog::OnLButtonDown(nFlags, point);
}

void CInfoDlg::OnPaint() 
{
	CPaintDC dc(this); // device context for painting

	dc.FillSolidRect(2,2,68,68,RGB(0,0,0));
	//dc.StretchBlt(4,4,64,64,&tilesets[global_sets[infoY/4][infoX/8]], map_tiles[infoY][infoX]*32, local_sets[infoY][infoX]*32, 32, 32, SRCCOPY);
}


























































































void CMainFrame::RecalcLayout(BOOL bNotify) 
{
	// TODO: Add your specialized code here and/or call the base class
	
	// 
	m_wndToolboxBar.CalcLayout();

	CFrameWnd::RecalcLayout(bNotify);
}

void CMainFrame::OnHelpTest() 
{
	// TODO: Add your command handler code here
}




void CMainFrame::OnToolsEnemycluster() 
{
	// TODO: Add your command handler code here
	CEnemyClusterDlg dlg;
	dlg.Rom = &theRom;

	dlg.DoModal();
}

void CMainFrame::OnUpdateToolsEnemycluster(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(theRom.is_loaded);
}









void CMainFrame::OpenDoorDlg(int nArea, int nDoor)
{
	static int last_area, last_door;

	if(nArea == -1) nArea = last_area;
	if(nDoor == -1) nDoor = last_door;

	last_area = nArea;
	last_door = nDoor;

 	CDoorDlg dlg;
 
 	dlg.ebRom = &theRom;
 	dlg.n_area = nArea;
	dlg.n_door = nDoor;

	int nResult = dlg.DoModal();

	if(nResult == 42)
	{
		// The dialog's internal code should apply the changes here
		OpenDestinationDlg(dlg.door.destindex, TRUE);
	}
}



void CMainFrame::OpenDestinationDlg(int nDest, bool from_doordlg, bool set_val, int x, int y)
{
	static int last_dest;

	if(nDest == -1)
		nDest = last_dest;

	last_dest = nDest;

	CDestinationDlg dlg(&theRom, nDest);

	if(set_val && x != -1 && y != -1)
	{
		// Copy new x and y to the temporary object in the dialog
		dlg.dest.x = x;
		dlg.dest.y = y;
	}

	int nResult = dlg.DoModal();

	if(nResult == 42)
	{
		last_dest = dlg.sel_dest;	// Selection may have changed
		SeekProc(TRUE, from_doordlg, NULL, dlg.sel_dest, -1, -1);
	}
	else
	{
		if(from_doordlg)
		{
			OpenDoorDlg(-1,-1);
		}
	}
}


void CMainFrame::SeekProc(bool is_starting, bool from_doordlg, bool set_val, int destindex, int x, int y)
{
	static bool open_doordlg;

	if(is_starting)
	{
		open_doordlg = from_doordlg;
		door_mode = DOOR_DESTSEEK;

		((CMapEditView*)GetActiveView())->ScrollToDestination(theRom.dest.entries[destindex]);

		GetActiveView()->Invalidate(FALSE);
	}

	else
	{
		door_mode = DOOR_NORMAL;
		GetActiveView()->Invalidate(FALSE);

		// If the seek is ending, this ignores the from_doordlg parameter
		// and uses the statically assigned open_doordlg that was set in
		// the starting call to this function
		OpenDestinationDlg(-1, open_doordlg, set_val, x, y);
	}
}





void CMainFrame::ReturnFromDestSeek()
{
	// For now, use some global information about the seek
}

void CMapEditView::ScrollToDestination(DestinationEntry dest)
{
	int w = GetTileWidth()-1;	// Tile dimensions
	int h = GetTileHeight()-1;

	int x = (dest.x / 4) - w/2;	// tile coordinates
	int y = (dest.y / 4) - h/2;	// (Offset for center)

	ScrollTo(x, y);	// Center target, more or less

	x = xOffset;	// Re-check offsets, in case the scrolling operation
	y = yOffset;	// was not able to fully scroll to the specified position

	int ox = dest.x - (x*4); // Offset coordinates
	int oy = dest.y - (y*4);

	int px = int((ox*8+5)*zoomlevel); // Offset pixel coordinates
	int py = int((oy*8+5)*zoomlevel);

	CRect rect;
	GetWindowRect(rect);
	SetCursorPos(px + rect.left, py + rect.top);

	Invalidate(FALSE);
}




void CMainFrame::UpdateStatusBar()
{
	// Updates status pane text based on stuff that's going on

	switch(edit_mode)
	{
	case MODE_DOOR:
	{
		if(input.keys & input.ctrl)
		{
			m_wndStatusBar.SetPaneText(0, "Click a door entry on the map to seek to its destination.");
		}
		else
		{
			m_wndStatusBar.SetPaneText(0, "Ready");
		}
	}
	break;

	}
}






int CMapEditView::GetHorzOffset()
{
	int retval = int(GetScrollPos(SB_HORZ)/zoomlevel);
	return retval;
}

int CMapEditView::GetVertOffset()
{
	return int(GetScrollPos(SB_VERT)/zoomlevel);
}




/*
 * Returns the position of the mouse cursor, relative to the origin of
 * the map, in the units relevant to the current editing mode.
 * Accounts for zoom level.
 */
CPoint CMapEditView::GetMousePos()
{
	// pixel coordinates
	int px = int((mouse.x + GetScrollPos(SB_HORZ))/zoomlevel);
	int py = int((mouse.y + GetScrollPos(SB_VERT))/zoomlevel);

	switch(edit_mode)
	{
	case MODE_MAP:
		// Units are 32 pixel tiles
		px/=32;
		py/=32;
		break;

	case MODE_DOOR:
		// Units are 8 pixel minitiles
		px/=8;
		py/=8;
		break;

	case MODE_SPRITE:
		// Units are single-pixel, no adjustment necessary
		break;

	case MODE_ENEMY:
		// Units are 64-pixel blocks
		px/=64;
		py/=64;
		break;
	}

	return CPoint(px,py);
}

CSize CMapEditView::GetScrollOffset()
{
	int px = int(GetScrollPos(SB_HORZ)/zoomlevel);
	int py = int(GetScrollPos(SB_VERT)/zoomlevel);

	switch(edit_mode)
	{
	case MODE_MAP:
		px/=32;
		py/=32;
		break;
	case MODE_DOOR:
		px/=8;
		py/=8;
		break;
	case MODE_SPRITE:
		break;
	case MODE_ENEMY:
		px/=64;
		py/=64;
		break;
	}
	return CSize(px,py);
}

int CMapEditView::GetTileWidth()
{
	CRect rect;
	GetWindowRect(rect);

	return int(rect.Width()/(32*zoomlevel)) + 1;
}

int CMapEditView::GetTileHeight()
{
	CRect rect;
	GetWindowRect(rect);

	return int(rect.Height()/(32*zoomlevel)) + 1;
}

void CMapEditView::SetZoomLevel(float zoom)
{
	if(zoom >= 1 && zoom <= 5)
		zoomlevel = zoom;

	needs_map_draw = true;
	draw.render = RENDER_MAP;

	ResizeScrollbars();

	Invalidate(FALSE);
}

float CMapEditView::GetZoomLevel()
{
	return zoomlevel;
}

void CMapEditView::ResizeScrollbars()
{
	CSize sizeTotal;

	sizeTotal.cx = int((8192+32)*zoomlevel);
	sizeTotal.cy = int((10240+32)*zoomlevel);

	SetScrollSizes(MM_TEXT, sizeTotal, CSize(0,0), CSize(0,0));
	SetScrollRange(SB_VERT,0,320);
	SetScrollRange(SB_HORZ,0,256);

	SetScrollPos(SB_HORZ, int(xOffset*32*zoomlevel));
	SetScrollPos(SB_VERT, int(yOffset*32*zoomlevel));
}




// Zoom menu commands

void CMainFrame::OnViewZoom100() 
{
	((CMapEditView*)GetActiveView())->SetZoomLevel(1);
}

void CMainFrame::OnUpdateViewZoom100(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(zoomlevel == 1);
}

void CMainFrame::OnViewZoom150() 
{
	((CMapEditView*)GetActiveView())->SetZoomLevel(1.5);
}

void CMainFrame::OnUpdateViewZoom150(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(zoomlevel == 1.5);
}

void CMainFrame::OnViewZoom200() 
{
	((CMapEditView*)GetActiveView())->SetZoomLevel(2);
}

void CMainFrame::OnUpdateViewZoom200(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(zoomlevel == 2);
}

void CMainFrame::OnViewZoom300() 
{
	((CMapEditView*)GetActiveView())->SetZoomLevel(3);
}

void CMainFrame::OnUpdateViewZoom300(CCmdUI* pCmdUI) 
{
	pCmdUI->SetRadio(zoomlevel == 3);
}



void CMainFrame::OnToolsEnemygroupeditor() 
{
	// TODO: Add your command handler code here
	CEditorEnemyGroupDlg dlg(&theRom, 0);
	dlg.DoModal();
}

void CMainFrame::OnUpdateToolsEnemygroupeditor(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
	pCmdUI->Enable(theRom.is_loaded);
}

void CMainFrame::OnFunstuffMuzak() 
{
	// TODO: Add your command handler code here
	muzak = !muzak;

	if(!muzak)
	{
		theRom.spc.Stop();
		theRom.music_playing = -1;
	}
}

void CMainFrame::OnUpdateFunstuffMuzak(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(muzak);
}

void CMainFrame::OnMouseMove(UINT nFlags, CPoint point) 
{
	// TODO: Add your message handler code here and/or call default
	
	CFrameWnd::OnMouseMove(nFlags, point);
}




BOOL CMapEditView::OnEraseBkgnd(CDC* pDC) 
{
	// Do not erase the background
	return TRUE;
}
