// Romclasses.cpp: Implementation for various new ROM classes
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MapEdit.h"
#include "Rom.h"
#include "classes.h"


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif





//////////////////////////////////////////////////////////////////////
// CSprite class
//////////////////////////////////////////////////////////////////////

void CSprite::ReadObject(FILE* file, int index, DWORD data)
{
	// Set object properties
	this->index = index;
	this->address = SNESToHex(GetPointer(file, ptr_address + (index*ptr_size)));

	// Seek to address and load data
	fseek(file, address, SEEK_SET);

	height = fgetc(file) & 0xF;
	width = (fgetc(file) >> 4) & 0xF;

	yy = fgetc(file);

	flags = fgetc(file);

	palette = (flags & 0xE) >> 1;

	uu[0] = fgetc(file);
	uu[1] = fgetc(file);
	uu[2] = fgetc(file);
	uu[3] = fgetc(file);

	bank = fgetc(file);


	// Check for extended frames
	int nextadr = SNESToHex(GetPointer(file, ptr_address + ((index+1)*ptr_size)));

	int diff = nextadr - this->address;

	switch(diff)
	{
	case 0x19:		// Standard eight frames
		num_frames = FRAME_STD;
		break;
	case 0x1B:		// Standard frame set with extra resting frame
		num_frames = FRAME_EXTRA;
		break;
	case 0x29:		// Full sixteen frames
		num_frames = FRAME_FULL;
		break;
	default:
		num_frames = FRAME_STD;
		break;
	}

	exframes = true;
	if(nextadr - this->address < 0x29)	// A sprite w/extra frames uses 29 bytes
		exframes = false;


	for(int i = 0; i < num_frames; i++)
	{
		unsigned char ch = fgetc(file);

		frame_ptr[i] = ch;
		frame_ptr[i]+= fgetc(file) << 8;

		flip[i] = frame_ptr[i] & 1;

		// Remove the last four bits
		frame_ptr[i] &= (0xFFF0);

		frames[i] = -1;
	}



	// Super-special fun happy stuff: load sprites

	CRom *pRom = (CRom*)data;

	static int gcount;	// Static counter: total sprites loaded by this class

	for(i = 0; i < num_frames; i++)
	{
		unsigned int adr = SNESToHex(frame_ptr[i] + (bank<<16));


		// search existing sprites
		bool bExists = false;
		for(int j = 0; j < gcount; j++)
		{
			if(adr == pRom->gfx2[j].address)
			{
				frames[i] = j;	// Set frame index
				bExists = true;
			}
		}

		// If a sprite from this address does not exist yet, load one
		if(!bExists)
		{
			pRom->gfx2[gcount].address = adr;
			pRom->gfx2[gcount].height = height;
			pRom->gfx2[gcount].width = width;
			pRom->gfx2[gcount].palette = palette;

			GetSprite(file, &pRom->gfx2[gcount]);

			frames[i] = gcount;	// Set frame index

			gcount++;
		}
	}

	pRom->gcount = gcount;
}

void CSprite::WritePtr(FILE* file, unsigned long address)
{

}

void CSprite::WriteObject(FILE* file, unsigned long address)
{

}


int CSprite::GetAddress()
{
	return address;
}

int CSprite::GetSize()
{
	return -1;
}


void CSprite::Draw(CDC* dc, int x, int y, int frame, CRom* pRom)
{
	int spr = frames[frame];

	Palette pal = pRom->pal[palette];

	pRom->gfx2[spr].Draw(dc, x, y, pal, flip[frame]);

}

void CSprite::Draw(CDC* dc, int x, int y, int frame, int area, int entry, CRom* pRom)
{
	Palette pal;

	// palette 4 is used to indicate that a sprite should use map palettes
	if(palette == 4)
	{
		int yy = pRom->object_areas[area].sprites[entry].y;

		// compute the map sector in which the object resides
		int sx = area % 32;
		int sy = area / 32; sy*=2;
		if(yy > 128) sy += 1;

		int set = pRom->global_sets[sy][sx];
		int p = pRom->global_pal[sy][sx];

		// the palette to use for sprites is stored in the first
		// color entry of the second subpalette of the tileset
		int sub = pRom->map_pal[set][p].spr_pal;// mpal_snes[set][p][2][0];

		if(sub == 0)
			pal = pRom->pal[4];
		else
		for(int i = 0; i < 16; i++)
		{
			pal[i].b = (pRom->mpal[set][p][sub-2][i]) & 255;
			pal[i].g = (pRom->mpal[set][p][sub-2][i] >> 8) & 255;
			pal[i].r = (pRom->mpal[set][p][sub-2][i] >>16) & 255;
		}
	}
	else
		pal = pRom->pal[palette];

	int spr = frames[frame];


	pRom->gfx2[spr].Draw(dc, x, y, pal, flip[frame]);
}






//////////////////////////////////////////////////////////////////////
// CMapPalGroup class
//////////////////////////////////////////////////////////////////////


void CMapPalGroup::ReadObject(FILE* file, int index, DWORD data)
{
	// Set object properties
	this->index = index;
	this->address = SNESToHex(GetPointer(file, ptr_address + (index*ptr_size)));

	// Seek to address and load data
	fseek(file, address, SEEK_SET);

	// Find the number of palettes in this set
	int nextadr = SNESToHex(GetPointer(file, ptr_address + ((index+1)*ptr_size)));

	num_pal = (nextadr - address) / 192;


	// Now load the colors and miscellaneous data stored in the palette
	// (Go ahead and load eight palettes, regardless of the actual number)
	for(int i = 0; i < 8; i++)
	{
		// Load miscellaneous junk
		fseek(file, address + 192*i, SEEK_SET);
		pal[i].event = fgetc(file);
		pal[i].event+= fgetc(file) << 8;

		fseek(file, address + 192*i + 32, SEEK_SET);
		pal[i].pal_ptr = fgetc(file);
		pal[i].pal_ptr+= fgetc(file) << 8;

		fseek(file, address + 192*i + 64, SEEK_SET);
		pal[i].spr_pal = fgetc(file);
		pal[i].spr_pal+= fgetc(file) << 8;

		// Sprite palette is two greater than the actual index of the palette used
		//pal[i].spr_pal -= 2;

		fseek(file, address + 192*i + 96, SEEK_SET);
		pal[i].flash = fgetc(file);
		pal[i].flash+= fgetc(file) << 8;

		// Seek back to beginning of palette
		fseek(file, address + 192*i, SEEK_SET);

		// Load colors
		for(int j = 0; j < 6; j++)
		{
			for(int k = 0; k < 16; k++)
			{
				unsigned short c;
				c = fgetc(file);
				c += fgetc(file) << 8;

				c &= 0x7FFF;

				pal[i][j][k].b = (c) >> 10;
				pal[i][j][k].g = (c & 0x03E0) >> 5;
				pal[i][j][k].r = (c & 0x001F);
			}
		}
	}
}

int CMapPalGroup::GetAddress()
{
	return address;
}

int CMapPalGroup::GetSize()
{
	// 192 bytes per 96-color palette
	return num_pal * 192;
}

MapPalette& CMapPalGroup::operator[](int n)
{
	return pal[n];
}


/* <NYI> */
void CMapPalGroup::WritePtr(FILE* file, unsigned long address)
{
	// Seek to the location of the pointer for this object index
	fseek(file, ptr_address + (index*ptr_size), SEEK_SET);

	// Write the specified address to that location
	address = HexToSNES(address);
	fputc((address >> 0) & 255, file);
	fputc((address >> 8) & 255, file);
	fputc((address >> 16)& 255, file);
	//fputc((address >> 24)& 255, file);

	fputc(0, file);
}

void CMapPalGroup::WriteObject(FILE* file, unsigned long address)
{
	// Write object's pointer
	WritePtr(file, address);

	// Write the data
	fseek(file, address, SEEK_SET);


	for(int i = 0; i < num_pal; i++)
	{
		// Hide secret data in the transparent colors :O
		fseek(file, address + 192*i, SEEK_SET);
		fputc(pal[i].event & 255, file);
		fputc((pal[i].event >> 8) & 255, file);

		fseek(file, address + 192*i + 32, SEEK_SET);
		fputc(pal[i].pal_ptr & 255, file);
		fputc((pal[i].pal_ptr >> 8) & 255, file);

		fseek(file, address + 192*i + 64, SEEK_SET);
		// Add two to the sprite palette for some reason...
		fputc((pal[i].spr_pal) & 255, file);
		fputc(((pal[i].spr_pal) >> 8) & 255, file);

		fseek(file, address + 192*i + 96, SEEK_SET);
		fputc(pal[i].flash & 255, file);
		fputc((pal[i].flash >> 8) & 255, file);
	}

	// Don't bother with the palette colors just yet...

/*
	for(int i = 0; i < 8; i++)
	{
		// Load miscellaneous junk
		fseek(file, address + 192*i, SEEK_SET);
		pal[i].event = fgetc(file);
		pal[i].event+= fgetc(file) << 8;

		fseek(file, address + 192*i + 32, SEEK_SET);
		pal[i].pal_ptr = fgetc(file);
		pal[i].pal_ptr+= fgetc(file) << 8;

		fseek(file, address + 192*i + 64, SEEK_SET);
		pal[i].spr_pal = fgetc(file);
		pal[i].spr_pal+= fgetc(file) << 8;

		fseek(file, address + 192*i + 96, SEEK_SET);
		pal[i].flash = fgetc(file);
		pal[i].flash+= fgetc(file) << 8;

		// Seek back to beginning of palette
		fseek(file, address + 192*i, SEEK_SET);

		// Load colors
		for(int j = 0; j < 6; j++)
		{
			for(int k = 0; k < 16; k++)
			{
				unsigned short c;
				c = fgetc(file);
				c += fgetc(file) << 8;

				c &= 0x7FFF;

				pal[i][j][k].b = (c) >> 10;
				pal[i][j][k].g = (c & 0x03E0) >> 5;
				pal[i][j][k].r = (c & 0x001F);
			}
		}
	}
	*/

}
/* </NYI> */








//////////////////////////////////////////////////////////////////////
// CEnemyCluster class
//////////////////////////////////////////////////////////////////////




void CEnemyCluster::ReadObject(FILE* file, int index, DWORD data)
{
	// Set object properties
	this->index = index;
	this->address = SNESToHex(GetPointer(file, ptr_address + (index*ptr_size)));

	// Seek to address and load data
	fseek(file, address, SEEK_SET);

	event = fgetc(file);
	event+= fgetc(file) << 8;

	p_pri = fgetc(file);
	p_sec = fgetc(file);

	if(p_pri) has_primary = true;
	if(p_sec) has_secondary = true;

	unsigned char ptotal = 0;
	unsigned char count = 0;

	if(has_primary)
		while(ptotal < 8)
		{
			primary[count].prb	= fgetc(file);

			ptotal+=primary[count].prb;

			primary[count].group	= fgetc(file);
			primary[count].group+= fgetc(file)<<8;
			count++;
		}
	n_pri = count;

	ptotal = 0;
	count = 0;

	if(has_secondary)
		while(ptotal < 8)
		{
			secondary[count].prb   = fgetc(file);

			ptotal+=secondary[count].prb;

			secondary[count].group = fgetc(file);
			secondary[count].group+= fgetc(file)<<8;
			count++;
		}
	n_sec = count;
}

void CEnemyCluster::WritePtr(FILE* file, unsigned long address)
{
	// Seek to the location of the pointer for this object index
	fseek(file, ptr_address + (index*ptr_size), SEEK_SET);

	// Write the specified address to that location
	address = HexToSNES(address);
	fputc((address >> 0) & 255, file);
	fputc((address >> 8) & 255, file);
	fputc((address >> 16)& 255, file);
	//fputc((address >> 24)& 255, file);

	fputc(0, file);
}

void CEnemyCluster::WriteObject(FILE* file, unsigned long address)
{
	// Writes the object to the specified address.
	// Also uses class-specific behavior to write the object's
	// entry in the class's pointer table.
	// This uses the index provided in the call to ReadObject to
	// write to the appropriate location in the pointer table.

	this->address = address;

	// Write pointer table entry
	WritePtr(file, address);

	// Seek to address
	fseek(file, address, SEEK_SET);

	// Write event flag
	fputc((event) & 255, file);
	fputc((event >> 8) & 255, file);

	// Encounter rates
	fputc(p_pri, file);
	fputc(p_sec, file);

	if(this->has_primary)
		for(int i = 0; i < n_pri; i++)
		{
			fputc(primary[i].prb, file);
			fputc((primary[i].group) & 255, file);
			fputc((primary[i].group >> 8) & 255, file);
		}

	if(this->has_secondary)
		for(int i = 0; i < n_sec; i++)
		{
			fputc(secondary[i].prb, file);
			fputc((secondary[i].group) & 255, file);
			fputc((secondary[i].group >> 8) & 255, file);
		}
}

int CEnemyCluster::GetAddress()
{
	return address;
}

int CEnemyCluster::GetSize()
{
	// Returns size of entry in bytes

	// format: AA AA XX YY [(BB CC CC) (DD EE EE)]
	return 4 + (has_primary? 3*n_pri:0) + (has_secondary? 3*n_sec:0);
}


void CEnemyCluster::AddGroup(int nSubCluster, int nFreq, int nGroup)
{
	switch(nSubCluster)
	{
	case 0:
		if(n_pri >= 8) return;

		primary[n_pri].prb = nFreq;
		primary[n_pri].group = nGroup;

		n_pri++;
		break;
	case 1:
		if(n_sec >= 8) return;

		secondary[n_sec].prb = nFreq;
		secondary[n_sec].group = nGroup;

		n_sec++;
		break;
	}
}

void CEnemyCluster::RemGroup(int nSubCluster, int nItem)
{
	int i;
	switch(nSubCluster)
	{
	case 0:
		if(n_pri <= 0) return;

		for(i = nItem; i < 7; i++)
			primary[i] = primary[i+1];
		n_pri--;
		break;
	case 1:
		if(n_sec <= 0) return;

		for(i = nItem; i < 7; i++)
			secondary[i] = secondary[i+1];
		n_sec--;
		break;
	}
}





//////////////////////////////////////////////////////////////////////
// CEnemyBattleGroup class
//////////////////////////////////////////////////////////////////////

CEnemyBattleGroup::CEnemyBattleGroup()
{
	entrycount = 0;
	runflag = 0;
	flaguse = 0;
	letterbox = 0;
	bg1 = bg2 = 0;
}

CEnemyBattleGroup::~CEnemyBattleGroup()
{

}

void CEnemyBattleGroup::ReadObject(FILE* file, int index, DWORD data)
{
	this->index = index;
	this->address = SNESToHex(GetPointer(file, ptr_address + (index*ptr_size)));



	fseek(file, ptr_address + (index*ptr_size), SEEK_SET);

	// skip the pointer itself, since we already read that
	fseek(file, 4, SEEK_CUR);

	// Read the group data embedded in the pointer table
	runflag = fgetc(file);
	runflag+= fgetc(file) << 8;
	flaguse = fgetc(file);
	letterbox=fgetc(file);


	// seek to the actual entry data
	fseek(file, address, SEEK_SET);

	unsigned char ch[3] = {0};
	unsigned char count = 0;

	while(ch[0]!=0xFF)
	{
		ch[0] = fgetc(file);

		if(ch[0]!=0xFF)
		{
			ch[1] = fgetc(file);
			ch[2] = fgetc(file);
			enemies[count].num = ch[0];
			enemies[count].enemy = ch[1];
			ch[2];	// High-order byte should always be zero
			count++;
		}
	}
	entrycount = count;

	// Read battle animation info
	// (This comes from an entirely different part of the ROM)
	fseek(file, 0xBDA9A + index*4, SEEK_SET);

	bg1 = fgetc(file);
	bg1+= fgetc(file) << 8;
	bg2 = fgetc(file);
	bg2+= fgetc(file) << 8;
}

void CEnemyBattleGroup::WritePtr(FILE* file, unsigned long address)
{
	// seek to appropriate entry in pointer table
	fseek(file, ptr_address + (index*ptr_size), SEEK_SET);

	address = HexToSNES(address);

	// write the pointer
	fputc((address >> 0) & 255, file);
	fputc((address >> 8) & 255, file);
	fputc((address >> 16)& 255, file);
	//fputc((address >> 24)& 255, file);
	fputc(0, file);

	// write the additional pointer table data
	fputc((runflag >> 0) & 255, file);
	fputc((runflag >> 8) & 255, file);
	fputc(flaguse, file);
	fputc(letterbox, file);
}

void CEnemyBattleGroup::WriteObject(FILE* file, unsigned long address)
{

	this->address = address;

	// Write object's pointer
	WritePtr(file, address);

	// Write the data
	fseek(file, address, SEEK_SET);

	for(int i = 0; i < entrycount; i++)
	{
		fputc(enemies[i].num, file);
		fputc((enemies[i].enemy) & 255, file);
		fputc(0, file);
	}
	fputc(0xFF, file);

	fseek(file, 0xBDA9A + index*4, SEEK_SET);

	fputc(bg1 & 255, file);
	fputc((bg1 >> 8) & 255, file);
	fputc(bg2 & 255, file);
	fputc((bg2 >> 8) & 255, file);
}

int CEnemyBattleGroup::GetAddress()
{
	return address;
}

int CEnemyBattleGroup::GetSize()
{
	// Format: (AA BB BB) 0xFF
	return entrycount*3 + 1;
}


bool CEnemyBattleGroup::AddEnemy(int num, int enemy)
{
	if(entrycount > 15) return false;

	enemies[entrycount].num = num;
	enemies[entrycount].enemy = enemy;
	entrycount++;
		
	return true;
}


bool CEnemyBattleGroup::InsertEnemy(int pos, int num, int enemy)
{
	if(pos >= entrycount) return false;

	// First, move all the entries below pos down a notch
	for(int i = entrycount; i > pos; i--)
	{
		if(i < 16)	// Anything that would be shifted above 16 just gets erased
			enemies[i] = enemies[i-1];
	}

	enemies[pos].num = num;
	enemies[pos].enemy = enemy;

	return true;
}


bool CEnemyBattleGroup::DeleteEnemy(int pos)
{
	if(pos < 0 || pos >= entrycount) return false;

	entrycount--;

	for(int i = pos; i < entrycount; i++)
		enemies[i] = enemies[i+1];

	return true;
}


int CEnemyBattleGroup::MoveEnemy(int pos, int newpos)
{
	unsigned char n,e;
	n = enemies[pos].num;
	e = enemies[pos].enemy;

	if(DeleteEnemy(pos))
	{
		if(newpos < 0)
			newpos = 0;
		else if(newpos < entrycount)
			newpos = entrycount;

		InsertEnemy(newpos, n, e);
	}
	return 1;
}


int CEnemyBattleGroup::CreateString(char* s, CRom* pRom)
{
	// Constructs a string that lists the enemies in this battle group

	char buf[64];	// Remember kiddies, buffer overflows are a very bad thing.

	sprintf(s, "%2d: ", index);

	// Group zero is always empty (doesn't even have a PT entry)
	if(index == 0)
	{
		strcat(s, "(No enemies)");
		return 0;
	}

	for(int j = 0; j < entrycount; j++)
	{
		int enemy = enemies[j].enemy-1;
		int num	  = enemies[j].num;

		if(num == 1)
			sprintf(buf, "%s%s", pRom->enemies[enemy].name, (j<entrycount-1?", ":""));
		else
			sprintf(buf, "%d  %s%s", num, pRom->enemies[enemy].name, (j<entrycount-1?", ":""));

		strcat(s, buf);
	}

	return 1;
}


//////////////////////////////////////////////////////////////////////
// CDoorEntry class
//////////////////////////////////////////////////////////////////////


CDoorEntry::CDoorEntry()
{
	doors.SetSize(0, 2);
}

CDoorEntry::~CDoorEntry()
{
	// delete all doors
	int len = doors.GetSize();

	for(int i = 0; i < len; i++)
	{
		delete doors[i];
	}
}


void CDoorEntry::ReadObject(FILE* file, int index, DWORD data)
{
	this->index = index;
	this->address = SNESToHex(GetPointer(file, ptr_address + (ptr_size*index)));

	bool firstload = data!=0;

	// Seek to data location
	fseek(file, address, SEEK_SET);

	// Number of doors in this entry
	int num_doors = fgetc(file);
	fgetc(file);

	// Adjustment for slight inaccuracies in original data
	if(firstload) {
		switch(index) {
		case 49:
		case 81:
		case 97:
		case 98:
		case 140:
		case 300:
		case 333:
		case 395:
		case 562:
		case 604:
		case 613:
		case 681:
		case 688:
		case 752:
		case 815:
		case 816:
		case 911:
		case 1136:
			num_doors--;
			break;
		case 1073:
			num_doors-=2;
			break;
		}
	}

	this->num_doors = num_doors;

	if(num_doors < 256 && num_doors >= 0)
	{
		doors.RemoveAll();
		doors.SetSize(num_doors);
		for(int i = 0; i < num_doors; i++)
		{
			doors[i] = new DoorEntry;
		}
	}

	for(int j = 0; j < num_doors; j++) {
		DoorEntry *d = (DoorEntry*)doors.GetAt(j);

		d->address = address + 2 + (j*5);
		d->y = fgetc(file);
		d->x = fgetc(file);
		d->flags = fgetc(file);

		int ch = fgetc(file);
		ch += fgetc(file) << 8;

		d->destination =  ch;
		d->direction = (ch & 0x300) >> 8;
		d->ropetype = ch;
		d->is_esc_end = (ch == 0x8000);
	}
}


void CDoorEntry::AddDoor(DoorEntry door)
{
	int n = doors.GetSize();

	// Allocates space for a new door entry with provided data
	doors.Add(new DoorEntry(door));


	num_doors = doors.GetSize();
}

void CDoorEntry::RemDoor(int door)
{
	// Free memory associated with door
	delete doors[door];

	doors.RemoveAt(door);
	//doors.FreeExtra();

	num_doors = doors.GetSize();
}

DoorEntry& CDoorEntry::operator[](int n)
{
	return *((DoorEntry*)doors[n]);
}

int CDoorEntry::GetAddress()
{
	return address;
}

int CDoorEntry::GetSize()
{
	// Format: AA BB (CC DD EE FF FF)
	return 2 + (5*num_doors);
}

void CDoorEntry::WritePtr(FILE* file, unsigned long address)
{
	// Seek to the location of the pointer for this object index
	fseek(file, ptr_address + (index*ptr_size), SEEK_SET);

	// Write the specified address to that location
	address = HexToSNES(address);
	fputc((address >> 0) & 255, file);
	fputc((address >> 8) & 255, file);
	fputc((address >> 16)& 255, file);
	//fputc((address >> 24)& 255, file);

	fputc(0, file);
}

void CDoorEntry::WriteObject(FILE* file, unsigned long address)
{
	this->address = address;

	// Write pointer table entry
	WritePtr(file, address);

	fseek(file, address, SEEK_SET);

	fputc(num_doors, file);
	fputc(0, file);

	for(int j = 0; j < num_doors; j++) {
		DoorEntry* door = (DoorEntry*)doors[j];
		
		door->address = address+2+(j*5);

		fputc(door->y, file);
		fputc(door->x, file);
		fputc(door->flags, file);
		switch(door->flags) {
		case DEST_SWITCH:
		case DEST_DOOR:
		case DEST_OBJECT:
		case DEST_PERSON:
			fputc(door->destination & 255, file);
			fputc((door->destination >> 8) & 255, file);
			break;
		case DEST_ROPE:
			fputc(door->ropetype & 255, file);
			fputc((door->ropetype >> 8) & 255, file);
			break;
		case DEST_STAIR:
			fputc(0, file);
			fputc((door->direction) & 255, file);
			break;
		case DEST_ESCALATOR:
			fputc(0, file);
			if(door->is_esc_end)
				fputc(0x80, file);
			else
				fputc((door->direction) & 255, file);
			break;
		}
	}
}










//////////////////////////////////////////////////////////////////////
// CDataPack class
//////////////////////////////////////////////////////////////////////

CDataPack::CDataPack()
{
	m_data.SetSize(0, 1);
}

CDataPack::~CDataPack()
{
	int n = m_data.GetSize();

	// Fwahaha! What was I thinking? The objects pointed to by the
	// CPtrArray in this object are allocated on the stack frame,
	// not the heap! They should not be deleted at all! rofl @ stupid :P
	/*for(int i = 0; i < n; i++)
	{
		void* ptr = m_data[i];

		delete ptr;
	}*/

	m_data.RemoveAll();
	m_data.FreeExtra();
}


// overloaded array-access operator
CDataEntry& CDataPack::operator[](int n)
{
	return *((CDataEntry*)m_data[n]);
}


// Accepts a pointer to a data entry and adds it to the array
void CDataPack::Add(CDataEntry *data)
{
	m_data.Add((void*)data);
}


// Returns number of items in package
int CDataPack::GetCount()
{
	return m_data.GetSize();
}


// Returns the total size of the package
int CDataPack::GetSize()
{
	// iterate through all items and determine size

	int len = m_data.GetSize();
	int total = 0;

	for(int i = 0; i < len; i++)
	{
		total+= ((CDataEntry*)m_data[i])->GetSize();
	}

	return total;
}


// Returns the address of the first byte of data
int CDataPack::GetStart()
{
	if(m_data.GetSize == 0) return -1;

	CDataEntry* d = (CDataEntry*)m_data[0];

	return d->GetAddress();
}


// Returns the address of the first unused byte after the end of the pack
int CDataPack::GetEnd()
{
	if(m_data.GetSize == 0) return -1;

	int a = GetStart() + GetSize();

	return a;
}


// Writes the package to the specified address
void CDataPack::WritePack(FILE* file, unsigned long address)
{
	int offset = 0;
	int len = m_data.GetSize();

	for(int i = 0; i < len; i++)
	{
		((CDataEntry*)m_data[i])->WriteObject(file, address+offset);

		offset += ((CDataEntry*)m_data[i])->GetSize();
	}
}









//////////////////////////////////////////////////////////////////////
// CDestinationList class

CDestinationList::CDestinationList()
{
	m_data.SetSize(0, 1);
	m_nCount = 0;
}

CDestinationList::~CDestinationList()
{
	int n = m_data.GetSize();

	// Delete all data items
	for(int i = 0; i < n; i++)
	{
		delete m_data[i];
	}

	m_data.RemoveAll();
	m_data.FreeExtra();
}

void CDestinationList::AddDest(DestinationEntry d)
{
	// Add a pointer to a new destination
	m_data.Add(new DestinationEntry(d));

	// Reset entry count
	m_nCount = m_data.GetSize();
}

DestinationEntry& CDestinationList::GetDest(int n)
{
	return *(DestinationEntry*)m_data[n];
}

DestinationEntry& CDestinationList::operator[](int n)
{
	return *(DestinationEntry*)m_data[n];
}

void CDestinationList::RemDest(int n)
{
	// Free the memory used by the obsolete data
	delete m_data[n];

	// Remove the entry in the pointer array
	m_data.RemoveAt(n);
	m_data.FreeExtra();

	// Reset the entry count
	m_nCount = m_data.GetSize();
}

void CDestinationList::ReadData(FILE* file, unsigned long address, unsigned long typelist_address)
{

}







