/*********************************************************************************************
*
*	SPCPlay.CPP:
*
*
*
*********************************************************************************************/

#include "spcplay.h"



/*CSPCPlayer::CSPCPlayer()
{
	
}


CSPCPlayer::~CSPCPlayer()
{
	
}*/


//**************************************************************************************************
// Wave Out Callback
//
// Event handler called by Windows for dealing with an open wave device
//
// In:
//    hwo        = Handle to output device
//    uMsg       = Message being sent
//    dwInstance = User defined data (not used in here)
//    pWHdr     -> Current wave header (if applicable)
//    dwParam2   = not used

void __stdcall WaveOutDone(HWAVEOUT hwo, u32 umsg, u32 dwInstance, WAVEHDR *wHdr, u32 dwParam2)
{
	CSPCPlayer* c = (CSPCPlayer*)dwInstance;

	if (umsg != WOM_DONE || !c->is_open) return;

	//char t[5];
	//itoa(dwInstance, t, 16);
	//allegro_message(t);

	c->apu.EmuAPU(wHdr->lpData, c->bufSize / c->wav.fmt.nBlockAlign, true);
	waveOutWrite(c->wav.handle, wHdr, sizeof(WAVEHDR));
}


//**************************************************************************************************
void CSPCPlayer::WavInit()
{
	wav.fmt.wFormatTag		= WAVE_FORMAT_PCM;
	wav.fmt.nChannels		= 2;
	wav.fmt.nSamplesPerSec	= 32000;
	wav.fmt.nAvgBytesPerSec	= 2 * (16 >> 3) * 32000;
	wav.fmt.nBlockAlign		= 2 * (16 >> 3);
	wav.fmt.wBitsPerSample	= 16;
	wav.fmt.cbSize			= 0;

	bufSize	= (100 * 32000 * 2 * 2) / 1000;		//100ms at 32kHz 16-bit stereo
	pBuf	= malloc(bufSize * 2);
}


//**************************************************************************************************
void CSPCPlayer::WavQuit()
{
	if (is_open) WavClose();					//Close device is it's open
	if (pBuf) free(pBuf);						//Free up allocated memory
}


//**************************************************************************************************
b8 CSPCPlayer::WavOpen()
{
	s32	error;

	//Open audio device ------------------------
	error = waveOutOpen(&wav.handle, WAVE_MAPPER , &wav.fmt, (u32)WaveOutDone, (u32)this, CALLBACK_FUNCTION | WAVE_ALLOWSYNC);

	if (error == MMSYSERR_NOERROR)
	{
		is_open = true;
		is_playing = true;

		//Initialize wave blocks ---------------
		memset(&wav.hdr[0], 0, sizeof(WAVEHDR));
		wav.hdr[0].dwBufferLength = bufSize;
		wav.hdr[0].lpData = (s8*)pBuf;
		waveOutPrepareHeader(wav.handle, &wav.hdr[0], sizeof(WAVEHDR));
		wav.hdr[0].dwUser = 0;

		memset(&wav.hdr[1], 0, sizeof(WAVEHDR));
		wav.hdr[1].dwBufferLength = bufSize;
		wav.hdr[1].lpData = (s8*)((u32)pBuf + bufSize);
		waveOutPrepareHeader(wav.handle, &wav.hdr[1], sizeof(WAVEHDR));
		wav.hdr[1].dwUser = 1;

		//Start output -------------------------
		WaveOutDone(wav.handle, WOM_DONE, (u32)this, &wav.hdr[0], 0);
		WaveOutDone(wav.handle, WOM_DONE, (u32)this, &wav.hdr[1], 0);

		return 1;
	}

	return 0;
}


//**************************************************************************************************
void CSPCPlayer::WavClose()
{
	is_open = false;

	waveOutReset(wav.handle);					//Tell Windows to reset the output device
	waveOutUnprepareHeader(wav.handle, &wav.hdr[0], sizeof(WAVEHDR));
	waveOutUnprepareHeader(wav.handle, &wav.hdr[1], sizeof(WAVEHDR));
	waveOutClose(wav.handle);					//Close output device

	wav.handle = NULL;
}












int CSPCPlayer::Init()
{
	// Initializes the SPC player so you can do stuff with it

	HINSTANCE	hDLL;
	u32			ver,min,opt;

	//==========================================
	// Load SNESAPU.DLL

	hDLL = LoadLibrary("SNESAPU.DLL");			//Try to load SNESAPU
	if (hDLL)
	{
		*(void**)&apu.SNESAPUInfo = (void*)GetProcAddress(hDLL,"SNESAPUInfo");
		apu.SNESAPUInfo(&ver, &min, &opt);

		sprintf(dllVer, "SNESAPU.DLL v%x.%x%c", ver >> 16, (ver >> 8) & 0xFF, ver & 0xFF);
	}
	else
	{
		allegro_message("Could not load SNESAPU.DLL");
		return -1;
	}

	//Get pointers to needed functions ---------
	*(void**)&apu.GetAPUData	= (void*)GetProcAddress(hDLL,"GetAPUData");
	*(void**)&apu.ResetAPU		= (void*)GetProcAddress(hDLL,"ResetAPU");
	*(void**)&apu.FixAPU		= (void*)GetProcAddress(hDLL,"FixAPU");
//	*(void**)&apu.SaveAPU		= (void*)GetProcAddress(hDLL,"SaveAPU");
//	*(void**)&apu.RestoreAPU	= (void*)GetProcAddress(hDLL,"RestoreAPU");
	*(void**)&apu.SetAPUOpt		= (void*)GetProcAddress(hDLL,"SetAPUOpt");
	*(void**)&apu.SetAPUSmpClk	= (void*)GetProcAddress(hDLL,"SetAPUSmpClk");
//	*(void**)&apu.SetAPULength	= (void*)GetProcAddress(hDLL,"SetAPULength");
//	*(void**)&apu.SetAPURAM		= (void*)GetProcAddress(hDLL,"SetAPURAM");
	*(void**)&apu.EmuAPU		= (void*)GetProcAddress(hDLL,"EmuAPU");
//	*(void**)&apu.SeekAPU		= (void*)GetProcAddress(hDLL,"SeekAPU");
//	*(void**)&apu.SetSPCDbg		= (void*)GetProcAddress(hDLL,"SetSPCDbg");
//	*(void**)&apu.GetSPCRegs	= (void*)GetProcAddress(hDLL,"GetSPCRegs");
//	*(void**)&apu.InPort		= (void*)GetProcAddress(hDLL,"InPort");
//	*(void**)&apu.InPortW		= (void*)GetProcAddress(hDLL,"InPortW");
//	*(void**)&apu.EmuSPC		= (void*)GetProcAddress(hDLL,"EmuSPC");
//	*(void**)&apu.SetDSPDbg		= (void*)GetProcAddress(hDLL,"SetDSPDbg");
//	*(void**)&apu.SetDSPReg		= (void*)GetProcAddress(hDLL,"SetDSPReg");
//	*(void**)&apu.SetDSPPitch	= (void*)GetProcAddress(hDLL,"SetDSPPitch");
//	*(void**)&apu.SetDSPAmp		= (void*)GetProcAddress(hDLL,"SetDSPAmp");
//	*(void**)&apu.SetDSPStereo	= (void*)GetProcAddress(hDLL,"SetDSPStereo");
//	*(void**)&apu.SetDSPEFBCT	= (void*)GetProcAddress(hDLL,"SetDSPEFBCT");
//	*(void**)&apu.EmuDSP		= (void*)GetProcAddress(hDLL,"EmuDSP");
//	*(void**)&apu.PlayWave		= (void*)GetProcAddress(hDLL,"PlayWave");
//	*(void**)&apu.UnpackWave	= (void*)GetProcAddress(hDLL,"UnpackWave");
//	*(void**)&apu.PackWave		= (void*)GetProcAddress(hDLL,"PackWave");
//	*(void**)&apu.VMax2dB		= (void*)GetProcAddress(hDLL,"VMax2dB");

	apu.GetAPUData(&apu.ram, &apu.xram, &apu.outPort, &apu.t64Cnt,
				   &apu.dsp, &apu.voice, &apu.vMMaxL, &apu.vMMaxR);


	WavInit();

}

int CSPCPlayer::Load(char* szFilename)
{
	//***LOAD FILE*** //
	HANDLE fh;
	DWORD l;

	fh = CreateFile(szFilename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);

	if(!fh)
	{
		// error opening file
		return -1;
	}

	ReadFile(fh, &spc, sizeof(spc), &l, NULL);
	CloseHandle(fh);


	//***INIT APU***-//
	apu.ResetAPU(0x20000);					//Reset APU

	// Set APU options (needs to be called because current build
	// of SNESAPU.dll incorrectly initializes emulation speed)
	//apu.SetAPUOpt(-1, 2, 16, 32000, -1, -1);
	apu.SetAPUOpt(MIX_I386, 2, 16, 32000, INT_GAUSS, -1);

	memcpy(apu.ram, spc.ram, 0x10000);		//Copy SPC to APU emulator
	memcpy(apu.xram, spc.xram, 0x40);
	memcpy(apu.dsp, &spc.dsp, 0x80);

	apu.FixAPU(*(u16*)spc.hdr.reg.pc,		//Fix up APU after loading an SPC
				spc.hdr.reg.a, spc.hdr.reg.x, spc.hdr.reg.y,
				spc.hdr.reg.psw,
				spc.hdr.reg.sp);
}

int CSPCPlayer::Play()
{
	if(!is_open)
	{
		WavOpen();
	}
	else
	{
		if(!is_playing)
		{
			waveOutRestart(wav.handle);
			is_playing = true;
		}
		else
		{

		}
	}

	return 1;
}

int CSPCPlayer::Stop()
{
	WavQuit();

	return 1;
}

int CSPCPlayer::Pause()
{
	if(is_playing)
	{
		waveOutPause(wav.handle);
		is_playing = false;
	}

	return 1;
}
