

/*********************************************************************************************
*
*	SPCPlay.h: encapsulates SPC player functionality in a single class,
*	CSPCPlayer.
*
*
*
*********************************************************************************************/

#ifndef __INC_SPCPLAY

#define __INC_SPCPLAY

//********************************************************************************************
// Includes

#include <allegro.h>
#include <windef.h>
#include <winbase.h>
#include <mmsystem.h>
#include <stdlib.h>
#include <stdio.h>

#include "types.h"
#include "snesapu.h"




//**************************************************************************************************
// Variables and Data

typedef struct
{
	u8	pc[2];
	u8	a;
	u8	x;
	u8	y;
	u8	psw;
	u8	sp;
} SPCReg;

//SPC header structure
typedef struct
{
	s8		fTag[33];							//File tag
	s8		tTerm[3];							//Tag terminator
	s8		ver;								//Version #/100
	SPCReg	reg;								//SPC Registers
	u16		__r1;
	s8		song[32];							//Song title
	s8		game[32];							//Game title
	s8		dumper[16];							//Name of dumper
	s8		comment[32];						//Comments
	s8		date[11];							//Date dumped
	s8		songLen[3];							//Song length (# of seconds to play before fading)
	s8		fadeLen[5];							//Fade length (milliseconds)
	s8		artist[32];							//Artist of song
	u8		chnDis;								//Channel Disabled?
	u8		emulator;							//Emulator dumped with
	u8		__r2[45];
} SPCHdr;

typedef struct
{
	SPCHdr	hdr;
	u8		ram[0x10000];
	DSPReg	dsp;
	u8		_r[0x40];
	u8		xram[0x40];
} SPCFile;

typedef struct
{
	HWAVEOUT		handle;						//Handle to output device
	tWAVEFORMATEX	fmt;						//Format of sample data being sent
	WAVEHDR			hdr[2];						//Headers for two buffers
} WavOut;


class CSPCPlayer
{
public:
	int Init();

	int Load(char* szFilename);

	int Play();
	int Stop();
	int Pause();

private:

	SAPUFunc		apu;	//** Ah, so these are the pointers to the SAPU functions
	s8				fn[MAX_PATH]; //** probably filename
	SPCFile			spc;		// instance of wrapper for SPC file?
	s8				dllVer[32];	// DLL version?

	WavOut	wav;
	b8		is_open;						//Output device is open
	b8		is_playing;						//Music is playing (not paused)
	void	*pBuf;							//-> memory allocated for sample data
	u32		bufSize;						//Size of single output buffer

	void	WavInit();
	void	WavQuit();
	b8		WavOpen();
	void	WavClose();

};





#endif










